/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"
#include "UGL_Win32_Widget.h"
#include "UGL_win32_display.h"
#include "UGL_Win32_DC.h"

#include "Widget.h"
#include "Control.h"
#include "Image.h"
#include "Button.h"
#include "Graphics.h"

// Defining this flag will include code to custom draw the push button.  Owner
// drawn push buttons are required to support background colors on buttons
#define UGL_CUSTOM_DRAWN_PUSHBUTTON

//========================================================================================
#define UI_CREATE_PUSHBUTTON	1
#define UI_CREATE_RADIOBUTTON	2
#define UI_CREATE_CHECKBUTTON	3
#define UI_CREATE_TOGGLEBUTTON	4

static UGL_Int _Button_New(UGL_Int type, UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError);
static UGL_Int _Button_ExStyle(UGL_Int ugl_style, UGL_Int type);
static UGL_Int _Button_Style(UGL_Int ugl_style, UGL_Int type);
LRESULT _Button_CommandChild(PUGL_Control ugl_control, WPARAM wParam, HWND hWnd);
LRESULT CALLBACK _Button_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
void Button_DrawImage(PUGL_Control);
static void _Button_GetMinimumSize(UGL_Int handle, POINT *point, UGL_Error uglError);
void _Button_DrawFocusRect(HWND hWnd, UINT message);
//========================================================================================

UGL_Int Checkbox_New(UGL_Int parentHandle, UGL_Int style, UGL_Error uglError) {
	return _Button_New(UI_CREATE_CHECKBUTTON, parentHandle, style, uglError);
}
//========================================================================================

void Button_SetAlignment(UGL_Int handle, UGL_Int alignment, UGL_Error uglError) {
	PUGL_Control ugl_button = UGL_CONTROL(handle);

	int bits = GetWindowLong(ugl_button->hWnd, GWL_STYLE);
	int style = bits;
	
	style &= ~(BS_LEFT | BS_CENTER | BS_RIGHT);
	
	if (IS_FLAG_SET(alignment, UGL_STYLE_LEAD)) {
		style |= BS_LEFT;
	} else if (IS_FLAG_SET(alignment, UGL_STYLE_TRAIL)) {
		style |= BS_RIGHT;
	} else if (IS_FLAG_SET(alignment, UGL_STYLE_CENTER)) {
		style |= BS_CENTER;
	}

	SetWindowLong(ugl_button->hWnd, GWL_STYLE, style);
	_Control_Redraw(ugl_button);
}
//========================================================================================

void Button_SetImage(UGL_Int handle, UGL_Int imageHandle, UGL_Error uglError) {
#ifdef _WIN32_WCE
	PUGL_Control ugl_control = UGL_CONTROL(handle);
	PUGL_Button ugl_button = UGL_BUTTON(handle);
	PUGL_Image ugl_image = UGL_IMAGE(imageHandle);
	ugl_button->ugl_image = ugl_image;
	if (ugl_image != NULL) {
		SetWindowText(ugl_control->hWnd, NULL);
	}
#else

	PUGL_Control ugl_control = UGL_CONTROL(handle);
	PUGL_Image ugl_image = UGL_IMAGE(imageHandle);
	PUGL_Button ugl_button = UGL_BUTTON(handle);
	int newBits, oldBits;
	HANDLE hBitmap;

	newBits = GetWindowLong(ugl_control->hWnd, GWL_STYLE);
	oldBits = newBits;

	newBits &= ~(BS_BITMAP | BS_ICON);
	if (ugl_image != NULL) {
		newBits |= BS_BITMAP;
	}

	if (newBits != oldBits) {
		SetWindowLong(ugl_control->hWnd, GWL_STYLE, newBits);
	}

	hBitmap = ugl_image != NULL ? ugl_image->image_handle : NULL;
	SendMessage(ugl_control->hWnd, BM_SETIMAGE, IMAGE_BITMAP, (LPARAM)hBitmap);

	ugl_button->ugl_image = ugl_image;

#endif
}

//========================================================================================

void Button_SetLabel(UGL_Int handle, UGL_String string, UGL_Error uglError) {
	PUGL_Control ugl_button = UGL_CONTROL(handle);

#ifndef _WIN32_WCE
	int newBits = GetWindowLong(ugl_button->hWnd, GWL_STYLE);
	int oldBits = newBits;

	newBits &= ~(BS_BITMAP | BS_ICON);

	if (newBits != oldBits) {
		SetWindowLong(ugl_button->hWnd, GWL_STYLE, newBits);
	}
#endif

	SetWindowText(ugl_button->hWnd, string);
}
//========================================================================================

UGL_Int RadioButton_New(UGL_Int parentHandle, UGL_Int style, UGL_Error uglError) {
	return _Button_New(UI_CREATE_RADIOBUTTON, parentHandle, style, uglError);
}
//========================================================================================

void Button_SetSelection(UGL_Int handle, UGL_Boolean selected, UGL_Error uglError) {
	PUGL_Control ugl_button = UGL_CONTROL(handle);
	int flags = selected ? BST_CHECKED : BST_UNCHECKED;
	
	/*
	* Feature in Windows. When BM_SETCHECK is used
	* to set the checked state of a radio or check
	* button, it sets the WM_TABSTOP style.  This
	* is undocumented and unwanted.  The fix is
	* to save and restore the window style bits.
	*/
	int bits = GetWindowLong(ugl_button->hWnd, GWL_STYLE);
	SendMessage(ugl_button->hWnd, BM_SETCHECK, flags, 0);
	SetWindowLong(ugl_button->hWnd, GWL_STYLE, bits);    
}
//========================================================================================

UGL_Boolean Button_GetSelection(UGL_Int handle, UGL_Error uglError)
{
	PUGL_Control ugl_button = UGL_CONTROL(handle);
	LRESULT state = SendMessage(ugl_button->hWnd, BM_GETSTATE, 0, 0);
	if ((state & BST_PUSHED) == BST_PUSHED) {
		return UGL_TRUE;
	} else {
		return UGL_FALSE;
	}
}
//========================================================================================

UGL_Int PushButton_New(UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	return _Button_New(UI_CREATE_PUSHBUTTON, parentHandle, ugl_style, uglError);
}
//========================================================================================

UGL_Int ToggleButton_New(UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	return _Button_New(UI_CREATE_TOGGLEBUTTON, parentHandle, ugl_style, uglError);
}
//========================================================================================


//========================================================================================
UGL_Int _Button_New(UGL_Int type, UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	int style = _Button_Style(ugl_style, type) | BS_MULTILINE;
	int extended_style = _Button_ExStyle(ugl_style, type);

	PUGL_Control ugl_button = _Control_New(_T("BUTTON"), sizeof(UGL_Button), _Button_Proc, 
						parentHandle, ugl_style, style, extended_style, uglError);

	switch (type) {
		case UI_CREATE_RADIOBUTTON: {
			((PUGL_Button)ugl_button)->ignoreWMCommand = FALSE;
			ugl_button->inheritsBackground = TRUE;
		} break;
		case UI_CREATE_CHECKBUTTON: {
			ugl_button->inheritsBackground = TRUE;
		} break;
	}
	
	return (UGL_Int)ugl_button;
}
//========================================================================================



//========================================================================================
UGL_Int _Button_Style(UGL_Int ugl_style, UGL_Int type) {
	int bits = _Control_Style(ugl_style);

#ifndef _WIN32_WCE
	if (IS_FLAG_SET(ugl_style, UGL_STYLE_FLAT))			bits |= BS_FLAT;
#endif
	
	if (IS_FLAG_SET(ugl_style, UGL_STYLE_LEAD))			bits |= BS_LEFT;
	else if (IS_FLAG_SET(ugl_style, UGL_STYLE_CENTER))	bits |= BS_CENTER;
	else if (IS_FLAG_SET(ugl_style, UGL_STYLE_TRAIL))	bits |= BS_RIGHT;


	//if ((style & SWT.ARROW) != 0) return bits | OS.BS_OWNERDRAW;
	switch (type) {
#ifdef UGL_CUSTOM_DRAWN_PUSHBUTTON
		case UI_CREATE_PUSHBUTTON:		bits |= (BS_OWNERDRAW | WS_TABSTOP); break;
#else
		case UI_CREATE_PUSHBUTTON:		bits |= (BS_PUSHBUTTON | WS_TABSTOP); break;
#endif
		case UI_CREATE_CHECKBUTTON:		bits |= (BS_CHECKBOX | WS_TABSTOP); break;
		case UI_CREATE_RADIOBUTTON:		bits |= BS_RADIOBUTTON; break;
		case UI_CREATE_TOGGLEBUTTON:	bits |= (BS_PUSHLIKE | BS_CHECKBOX); break;
		default:						bits |= (BS_PUSHBUTTON | WS_TABSTOP); break;
	}

	return bits;
}

//========================================================================================
#ifdef UGL_CUSTOM_DRAWN_PUSHBUTTON
void _Button_OwnerDraw(PUGL_Control button, LPDRAWITEMSTRUCT drawItemStruct) {
	HDC hdc = drawItemStruct->hDC;
	LPRECT rect = &drawItemStruct->rcItem;
	int textLength;
	int oldBkMode;
	LPTSTR text;
	HBRUSH backgroundBrush;
	COLORREF foreground, oldForeground;
	HWND hWnd = button->hWnd;
	BOOL disabled = (drawItemStruct->itemState & ODS_DISABLED) == ODS_DISABLED;
	BOOL pushed = (drawItemStruct->itemState & ODS_SELECTED) == ODS_SELECTED;
	BOOL focused = (drawItemStruct->itemState & ODS_FOCUS) == ODS_FOCUS;
	PUGL_Button ugl_button = (PUGL_Button) button;
	PUGL_Image image = ugl_button->ugl_image;
	BITMAP bitmap;


	UINT flags = (pushed ? DFCS_PUSHED : 0);
	int bits = GetWindowLong(hWnd, GWL_STYLE);
	int align; //the horizontal alignment of the text to be drawn
	/*
	 * Check the button style and set the drawing style accordingly.
	 * The logic here is a little strange, but it is because BS_CENTER
	 * also seems to set all the bits of BS_RIGHT, so if we just make
	 * BS_CENTER the default, without checking it first, we end up with
	 * right-justified text.
	 * This flag is also used for manually determining the position of the
	 * button text in the case of a disabled button.
	 */
	if (IS_FLAG_SET(bits, BS_CENTER)) align = DT_CENTER;
	else if (IS_FLAG_SET(bits, BS_RIGHT)) align = DT_RIGHT;
	else if (IS_FLAG_SET(bits, BS_LEFT)) align = DT_LEFT;
	else align = DT_CENTER;

	if (disabled) flags |= DFCS_INACTIVE;

	// LRN - To match the original control behavior, don't draw frame on image buttons
	if (image == NULL) {
#ifndef _WIN32_WCE
		if (IS_FLAG_SET(bits, BS_FLAT)) flags |= DFCS_FLAT;
		DrawFrameControl(hdc, rect, DFC_BUTTON, flags | DFCS_BUTTONPUSH | DFCS_ADJUSTRECT);
#else
		DrawFrameControl(hdc, rect, DFC_BUTTON, flags | DFCS_BUTTONPUSH);
		rect->top += 1;
		rect->left += 1;
		rect->bottom -= 1;
		rect->right -= 1;
#endif
	}

	// Update the background color of the button
#ifdef _WIN32_WCE
	if (pushed) {
		backgroundBrush = GetStockObject(BLACK_BRUSH);
	} else 
#endif		
	if (button->background_color != NULL) {
		backgroundBrush = Color_GetBrush(button->background_color);
	} else {
		backgroundBrush = GetSysColorBrush(COLOR_BTNFACE);
	}
	FillRect(hdc, rect, backgroundBrush);

	// Update the foreground color of the button
#ifdef _WIN32_WCE
	if (pushed) {
		foreground = RGB(255, 255, 255);
	} else
#endif
	if (button->foreground_color != NULL) {
		foreground = button->foreground_color->color;
	} else {
		foreground = GetSysColor(COLOR_BTNTEXT);
	}

	// Draw the focus rectangle if the button has focus.
	// NOTE:  Drawing the focus rectangle after the fillbackground
	// happens is important so that the focus rectangle will still 
	// be visible with a non-default background color
	if (focused && !pushed) {
		DrawFocusRect(hdc, rect);
	}

	textLength = GetWindowTextLength(hWnd)+1;
	text = calloc(sizeof(TCHAR), textLength);
	GetWindowText(hWnd, text, textLength);

	oldBkMode = SetBkMode(hdc, TRANSPARENT);

#ifndef _WIN32_WCE
	if (!disabled) {
#endif
		if (image != NULL) {
			PUGL_DC ugl_dc;
			int imageWidth, imageHeight;
			GetObject(ugl_button->ugl_image->image_handle, sizeof(BITMAP), &bitmap);
			imageWidth = bitmap.bmWidth;
			imageHeight = bitmap.bmHeight;
			ugl_dc = _internal_new_DC(hdc, NULL,rect, NULL);
			if (align == DT_LEFT) {
				Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_button->ugl_image, 2, ((rect->bottom-rect->top)- imageHeight)/2, NULL);
			} else if (align == DT_RIGHT) {
				Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_button->ugl_image, ((rect->right-rect->left)-imageWidth)-2, ((rect->bottom-rect->top)- imageHeight)/2, NULL);
			} else if (align == DT_CENTER) {
				Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_button->ugl_image, ((rect->right-rect->left)-imageWidth)/2, ((rect->bottom-rect->top)- imageHeight)/2, NULL);
			} else {
				Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_button->ugl_image, ((rect->right-rect->left)-imageWidth)/2, ((rect->bottom-rect->top)- imageHeight)/2, NULL);
			}
		} else {
#ifdef _WIN32_WCE
			if (!disabled) {
#endif
				oldForeground = SetTextColor(hdc, foreground);
#ifdef _WIN32_WCE
			} else {
				oldForeground = SetTextColor(hdc, GetSysColor(COLOR_GRAYTEXT));
			}
#endif
			DrawText(hdc, text, -1, rect, align | DT_VCENTER | DT_SINGLELINE);
			SetTextColor(hdc, oldForeground);
		}
#ifndef _WIN32_WCE
	} else {
		int buttonWidth, buttonHeight;
		int x,y,width,height;

		// Cache to size of the button because the next call to DrawText will modify the
		// values in the rect structure
		buttonWidth = rect->right - rect->left;
		buttonHeight = rect->bottom - rect->top;

		// Caclulate the size required to draw the string.  
		DrawText(hdc, text, -1, rect, DT_CALCRECT);

		// Caclulate the correct position of the string to center it in the button
		width = rect->right - rect->left;
		height = rect->bottom - rect->top;

		x = rect->left;
		y = rect->top + (buttonHeight-height)/2;

		if (align == DT_RIGHT){
			x = x + (buttonWidth - width) - 1;
		}
		else if (align == DT_LEFT){
			x = x + 1; //plus 1 and minus 1 (above) just to keep things spacious
		}
		else { //center is the default and all that PPro uses
			x = x + (buttonWidth-width)/2;
		}

		DrawState(hdc, NULL, NULL, (LPARAM)text, 0, x,y, width, height, DST_TEXT | DSS_DISABLED);
	}
#endif

	free(text);

	SetBkMode(hdc, oldBkMode);
}
#endif

//========================================================================================
LRESULT CALLBACK _Button_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {

	switch (message) {
		case UI_COMMAND_CHILD: {					
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			return _Button_CommandChild(ugl_control, wParam, hWnd);
			} break;

		case UI_GET_MINIMUM_SIZE: {
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			UGL_Int handle = (UGL_Int)ugl_control;
			POINT* point = (POINT*)wParam;
			_Button_GetMinimumSize(handle, point, (UGL_Error)lParam);
			return 0;
			} break;
		case UI_GETBACKGROUNDRGB: {
			LONG style = GetWindowLong(hWnd, GWL_STYLE);
			
			// BS_OWNERDRAW style is combination of both
			// BS_CHECKBOX and BS_RADIOBUTTON so we must
			// check that the button is not owner drawn before
			// checking if it is a checkbox or radio button
			if ((style & BS_OWNERDRAW) != BS_OWNERDRAW) {
				if (((style & BS_CHECKBOX) == BS_CHECKBOX) ||
					((style & BS_RADIOBUTTON) == BS_RADIOBUTTON)) {
						// The background for checkboxes and radio buttons is the same as
						// the parents background
						return SendMessage(GetParent(hWnd), UI_GETBACKGROUNDRGB, 0, 0);
				} 
			}

			return (UGL_Int)GetSysColor(COLOR_BTNFACE);
		} break;
		case UI_SET_FOCUS:{
			HWND result;
			
		   /*
			* This extra logic exists because of a bug (or at least undesireable behavior) in Windows. 
			* Ultimately, ForceFocus() results in a WM_Command message. Mouse clicks also result
			* in this message. We are supposed to be able to differentiate where the message
			* comes from based on the HIWORD of the WPARAM, but that value is the same for
			* 'tabbing' and for mouse clicks.
			* That's a problem because it means that by tabbing between RadioButtons, we also
			* toggle which one is selected. The fix is this workaround, which basically
			* amounts to setting a flag on the PUGL_Button struct.
			*/
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			PUGL_Button ugl_button = (PUGL_Button)ugl_control;
			int style = GetWindowLong(ugl_control->hWnd, GWL_STYLE);
			if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
				ugl_button->ignoreWMCommand = TRUE;
			}
			result = SetFocus((HWND)wParam);
			if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
				ugl_button->ignoreWMCommand = FALSE;
			}
			return (result != NULL);
		} break;

#ifdef UGL_CUSTOM_DRAWN_PUSHBUTTON
		case UI_DRAWITEM: {
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			LPDRAWITEMSTRUCT drawItemStruct = (LPDRAWITEMSTRUCT)lParam;
			LPRECT rect = &drawItemStruct->rcItem;
			_Button_OwnerDraw(ugl_control, drawItemStruct);
			PaintCallback(UGL_WIDGET(ugl_control)->callback_target, rect->left, rect->top, (rect->right - rect->left), (rect->bottom - rect->top));
			return TRUE;
		} break;

//fix for bug 158718 space doesn't generate button "click"
		case WM_KEYDOWN:{
		
			if(wParam == VK_SPACE){
				int state;
				BOOL selected;
				int style = GetWindowLong(hWnd, GWL_STYLE);
			
				PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
				UGL_Int handle = (UGL_Int)ugl_control;
				PUGL_Control ugl_button = UGL_CONTROL(handle);

				if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
					/*
					* The explanation for this is given in detail 
					* in the case statement for UI_SET_FOCUS in _Button_Proc
					*/
					if (((PUGL_Button)ugl_button)->ignoreWMCommand) {
						return 0;
					}
				}
				state = SendMessage(hWnd, BM_GETCHECK, 0, 0);
				selected = (state & BST_CHECKED) != 0;
				
				Button_SetSelection(handle, !selected, NULL);
				ButtonCallback(UGL_WIDGET(ugl_button)->callback_target);
				}
		} break;
		case WM_LBUTTONUP:{

		int state;
		BOOL selected;
		int style = GetWindowLong(hWnd, GWL_STYLE);
	
		PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
		UGL_Int handle = (UGL_Int)ugl_control;
		PUGL_Control ugl_button = UGL_CONTROL(handle);

		if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
			/*
			 * The explanation for this is given in detail 
			 * in the case statement for UI_SET_FOCUS in _Button_Proc
			 */
			if (((PUGL_Button)ugl_button)->ignoreWMCommand) {
				return 0;
			}
		}
		state = SendMessage(hWnd, BM_GETCHECK, 0, 0);
		selected = (state & BST_CHECKED) != 0;
		
		Button_SetSelection(handle, !selected, NULL);
		ButtonCallback(UGL_WIDGET(ugl_button)->callback_target);
		} break;
#else

		case WM_KEYDOWN:{
		//fix for bug 158718 space doesn't generate button "click"
#if defined (WIN32_PLATFORM_WFSP)
			if(wParam == VK_RETURN) {
#else
			if(wParam == VK_SPACE || wParam == VK_RETURN) {
#endif
				int state;
				BOOL selected;
				int style = GetWindowLong(hWnd, GWL_STYLE);
			
				PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
				UGL_Int handle = (UGL_Int)ugl_control;
				PUGL_Button ugl_button = (PUGL_Button) ugl_control;

				if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
					/*
					* The explanation for this is given in detail 
					* in the case statement for UI_SET_FOCUS in _Button_Proc
					*/
					if (ugl_button->ignoreWMCommand) {
						return 0;
					}
				}
				state = SendMessage(hWnd, BM_GETCHECK, 0, 0);
				selected = (state & BST_CHECKED) != 0;
				
				Button_SetSelection(handle, !selected, NULL);
				
				ButtonCallback(UGL_WIDGET(ugl_button)->callback_target);
					_Control_Proc(hWnd, message, wParam, lParam);
					if(ugl_button->ugl_image != NULL){
						Button_DrawImage(ugl_control);
					}
				return TRUE;
			}
		} break;

		case WM_SYSKEYUP:
		case WM_KEYUP:{
#if defined (WIN32_PLATFORM_WFSP)
			if(wParam == VK_RETURN) {
#else
			if(wParam == VK_SPACE || wParam == VK_RETURN) {
#endif
				PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
				UGL_Int handle = (UGL_Int)ugl_control;

				LRESULT result = _Control_Proc(hWnd, message, wParam, lParam);

				_Button_DrawFocusRect(hWnd, message);
				return result;
			}
		} break;

		case WM_LBUTTONUP:{
		int state;
		BOOL selected;
		int style = GetWindowLong(hWnd, GWL_STYLE);
	
		PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
		UGL_Int handle = (UGL_Int)ugl_control;
		PUGL_Button ugl_button = (PUGL_Button) ugl_control;

		if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
			/*
			 * The explanation for this is given in detail 
			 * in the case statement for UI_SET_FOCUS in _Button_Proc
			 */
			if (ugl_button->ignoreWMCommand) {
				return 0;
			}
		}
		state = SendMessage(hWnd, BM_GETCHECK, 0, 0);
		selected = (state & BST_CHECKED) != 0;
		
		Button_SetSelection(handle, !selected, NULL);
		
		ButtonCallback(UGL_WIDGET(ugl_button)->callback_target);
			_Control_Proc(hWnd, message, wParam, lParam);
			if(ugl_button->ugl_image != NULL){
				Button_DrawImage(ugl_control);
			}
			_Button_DrawFocusRect(hWnd, message);
			return TRUE;
		}

		case WM_PAINT: {
			
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			PUGL_Button ugl_button = (PUGL_Button) ugl_control;
			_Control_Proc(hWnd, message, wParam, lParam);
			if(ugl_button->ugl_image != NULL){
				Button_DrawImage(ugl_control);
			}
			if(ugl_control->hWnd == GetFocus()) {
				_Button_DrawFocusRect(hWnd, message);
			}
				return TRUE;
		}
		case WM_KILLFOCUS: {
			_Button_DrawFocusRect(hWnd, message);
		} break;
		case WM_SETFOCUS: {
			_Button_DrawFocusRect(hWnd, message);
		} break;
#endif
	}
	return _Control_Proc(hWnd, message, wParam, lParam);
}
//========================================================================================

LRESULT _Button_CommandChild(PUGL_Control ugl_control, WPARAM wParam, HWND hWnd){
	UGL_Int handle = (UGL_Int)ugl_control;
	PUGL_Control ugl_button = UGL_CONTROL(handle);
	LRESULT result = 0;
	int code = wParam >> 16;

	switch (code) {
#ifndef _WIN32_WCE
		case BN_DOUBLECLICKED: 
#endif
		case BN_CLICKED: {

/*

	// this is being marked out and moved to the button proc's WM_LBUTTONUP event, this is for fixing the bug JYJC6QK9ZP where
	// the button does not generate a clicked event when the button is re-layoutted to other location while being clicked
	// the button now uses the WM_LBUTTONUP to detect the button click.

			//if ((style & (SWT.CHECK | SWT.RADIO | SWT.TOGGLE)) == 0) return false;
			int state;
			BOOL selected;
			int style = GetWindowLong(ugl_button->hWnd, GWL_STYLE);
			fprintf(pFile, "bn clicked 00\n");
			if (IS_FLAG_SET(style, BS_RADIOBUTTON)) {
			*/
				/*
				 * The explanation for this is given in detail 
				 * in the case statement for UI_SET_FOCUS in _Button_Proc
				 */
			/*
				if (((PUGL_Button)ugl_button)->ignoreWMCommand) {
					return 0;
				}
			}
			state = SendMessage(hWnd, BM_GETCHECK, 0, 0);
			selected = (state & BST_CHECKED) != 0;
			
			Button_SetSelection(handle, !selected, NULL);
			ButtonCallback(UGL_WIDGET(ugl_button)->callback_target);
*/
		} break;
		default: break;
	}
	return result;
}

//========================================================================================

UGL_Int _Button_ExStyle(UGL_Int ugl_style, UGL_Int type) {
	int bits = 0; 

#ifndef _WIN32_WCE	
	bits |= WS_EX_NOINHERITLAYOUT;
#endif
	if (IS_FLAG_SET(ugl_style, UGL_STYLE_BORDER)) bits |= WS_EX_CLIENTEDGE;
	//else bits &= ~WS_EX_CLIENTEDGE;

	return bits;
}
//========================================================================================

void _Button_GetMinimumSize(UGL_Int handle, POINT *point, UGL_Error uglError) {
	PUGL_Control ugl_control = UGL_CONTROL(handle);
	HWND hWnd = ugl_control->hWnd;
	PUGL_Button ugl_button = UGL_BUTTON(handle);
	int border = _Control_GetBorderWidth(hWnd);
	int width = 0, height = 0;
	int extra = 0;
	int bits = GetWindowLong(hWnd, GWL_STYLE);

	if (ugl_button->ugl_image == NULL) {
		HGDIOBJ oldFont = 0;
		HDC hDC = GetDC(hWnd);
		HGDIOBJ newFont = (HGDIOBJ)SendMessage (hWnd, WM_GETFONT, 0, 0);
		TEXTMETRIC lptm;
		int length = 0;
		LPTSTR window_text = NULL;

		if (newFont != 0) oldFont = SelectObject(hDC, newFont);
		GetTextMetrics(hDC, &lptm);
		length = GetWindowTextLength(hWnd);
		if (length == 0) {
			height += lptm.tmHeight;
		} else {
			RECT rect;
			int flags = DT_CALCRECT | DT_SINGLELINE;

			SetRect((LPRECT)&rect, 0, 0, 1, 1);
			extra = MAX(8, lptm.tmAveCharWidth);
			window_text = (LPTSTR)calloc(sizeof(TCHAR), length + 2);
			GetWindowText(hWnd, window_text, length + 1);
			DrawText(hDC, window_text, length, &rect, flags);
			free(window_text);
			width += rect.right - rect.left;
			height += rect.bottom - rect.top;
		}
		if (newFont != 0) SelectObject (hDC, oldFont);
		ReleaseDC (hWnd, hDC);
	} else {
		int x, y;
		Image_GetBounds((int)ugl_button->ugl_image, &x, &y, &width, &height, uglError);
		extra = 8;
	}


	if ((bits & (BS_CHECKBOX | BS_RADIOBUTTON)) != 0) {
		/*
		 * TODO, should not recalculate checkWidth and checkHeight 
		 * repeatedly, but should cache the value somewhere.
		 */
		int checkWidth = 0, checkHeight = 0;
//		HBITMAP hBitmap = LoadBitmap(NULL, OBM_CHECKBOXES);
		//if (hBitmap == 0) {
#ifndef _WIN32_WCE
			checkWidth = GetSystemMetrics(SM_CXVSCROLL);
			checkHeight = GetSystemMetrics (SM_CYVSCROLL);
#else
			checkWidth = GetSystemMetrics(SM_CXSMICON);
			checkHeight = GetSystemMetrics(SM_CYSMICON);
#endif // if defined _WIN32_WCE
		//} else {
		//	BITMAP bitmap;
		//	GetObject(hBitmap, sizeof(BITMAP), &bitmap);
		//	DeleteObject(hBitmap);
		//	checkWidth = bitmap.bmWidth / 4;
		//	checkHeight =  bitmap.bmHeight / 3;
		//}

		width += checkWidth + extra;
		height = MAX(height+4, (checkHeight + 3));
	} else {
		/**
		 * BS_PUSHBUTTON's value is 0x00, thus, bit
		 * checking for the flag always fails.
		 * However, an else clause is sufficient
		 * because there are only 4 styles, and this
		 * case covers the other two.
		 */
	
		width += 12;  height += 10;
	}
	
	width += border * 2; height += border * 2;
	
	
	point->x = width;
	point->y = height;
	
}

//========================================================================================
// fix for Label.setAlignment, paint to image ourselves to adjust the position depending on the image size
void Button_DrawImage(PUGL_Control handle) {
	
#ifdef _WIN32_WCE
	PUGL_Control ugl_control = (PUGL_Control)handle;
	PUGL_Button ugl_button = (PUGL_Button)ugl_control;
	PUGL_Image ugl_image = ugl_button->ugl_image;
	HWND hWnd = ugl_control->hWnd;
	HDC hdc = GetDC(hWnd);
	HDC srcHDC = CreateCompatibleDC(hdc);
	HBRUSH newBrush;
	BITMAP bitmap;
	int bits = GetWindowLong(hWnd, GWL_STYLE);
	HBRUSH backgroundBrush;

	if(ugl_image != NULL){
		HGDIOBJ oldBrush;
		RECT rect;

		int rop2 = 0;
		int dwRop;
		newBrush = CreatePatternBrush(ugl_button->ugl_image->image_handle);
		oldBrush = SelectObject(hdc, newBrush);
		rop2 = SetROP2(hdc, SRCCOPY);
		SetROP2(hdc, rop2);
		dwRop = SRCCOPY;	
		GetClipBox(hdc, &rect);

		// Paint the background for the label
			if (ugl_control->background_color != NULL) {
				backgroundBrush = Color_GetBrush(ugl_control->background_color);
			} else {
				backgroundBrush = GetSysColorBrush(COLOR_BTNFACE);
			}
		if (((bits & BS_CHECKBOX) == BS_CHECKBOX) ||
		((bits & BS_RADIOBUTTON) == BS_RADIOBUTTON)) {
 
		} else {
			FillRect(hdc, &rect, backgroundBrush);
		}

		// check alignment and paint the image in accordance to the alignment
		SelectObject(srcHDC, ugl_image->image_handle);
		GetObject(ugl_button->ugl_image->image_handle, sizeof(BITMAP), &bitmap);			
			
		if (IS_FLAG_SET(bits, BS_CENTER)) {	
			BitBlt(hdc, ((rect.right-rect.left)-bitmap.bmWidth)/2, ((rect.bottom-rect.top)-bitmap.bmHeight)/2, bitmap.bmWidth, bitmap.bmHeight, srcHDC, 0, 0, dwRop); 
		} else if (IS_FLAG_SET(bits, BS_RIGHT)) {
			BitBlt(hdc, rect.right-bitmap.bmWidth-2, ((rect.bottom-rect.top)-bitmap.bmHeight)/2, bitmap.bmWidth, bitmap.bmHeight, srcHDC, 0, 0, dwRop); 
		} else {
			if (((bits & BS_CHECKBOX) == BS_CHECKBOX) || ((bits & BS_RADIOBUTTON) == BS_RADIOBUTTON)) 
			{
				BitBlt(hdc, rect.left+20, ((rect.bottom-rect.top)-bitmap.bmHeight)/2, bitmap.bmWidth, bitmap.bmHeight, srcHDC, 0, 0, dwRop); 
			}
			else
			{
				BitBlt(hdc, rect.left+2, ((rect.bottom-rect.top)-bitmap.bmHeight)/2, bitmap.bmWidth, bitmap.bmHeight, srcHDC, 0, 0, dwRop); 
			}

		}
			SelectObject(hdc, oldBrush);
			DeleteDC(srcHDC);
	}
	ReleaseDC(hWnd, hdc);
#endif
}

//========================================================================================

void _Button_DrawFocusRect(HWND hWnd, UINT message) {
#ifdef _WIN32_WCE
	RECT rect;
	HDC hdc = GetDC(hWnd);
//	HWND parent = GetParent(hWnd);

//	if(parent == NULL) // Shell
//		return;

	if (GetClipBox(hdc, &rect) != NULLREGION) {
		DrawFocusRect(hdc, &rect);
	}

#endif
}
