/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

package org.eclipse.swt.widgets;


import org.eclipse.swt.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.events.*;

import com.ibm.ugl.eswt.OS;

/**
 * Instances of this class represent a selectable user interface object that
 * issues notification when pressed and released. 
 * <dl>
 * <dt><b>Styles:</b></dt>
 * <dd>CHECK, PUSH, RADIO, TOGGLE, FLAT</dd>
 * <dd>LEFT, RIGHT, CENTER</dd>
 * <dt><b>Events:</b></dt>
 * <dd>Selection</dd>
 * </dl>
 * <p>
 * Note: Only one of the styles CHECK, PUSH, RADIO, and TOGGLE 
 * may be specified.
 * </p><p>
 * Note: Only one of the styles LEFT, RIGHT, and CENTER may be specified.
 * </p><p>
 * IMPORTANT: This class is intended to be subclassed <em>only</em>
 * within the SWT implementation.
 * </p>
 */

public class Button extends Control {
	Image image;
	int alignment = -1;
	boolean selection = false;
	String text;
	boolean isDefault = false;
	
/**
 * Constructs a new instance of this class given its parent
 * and a style value describing its behavior and appearance.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>SWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>SWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p>
 *
 * @param parent a composite control which will be the parent of the new instance (cannot be null)
 * @param style the style of control to construct
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 *
 * @see SWT#CHECK
 * @see SWT#PUSH
 * @see SWT#RADIO
 * @see SWT#TOGGLE
 * @see SWT#FLAT
 * @see SWT#LEFT
 * @see SWT#RIGHT
 * @see SWT#CENTER
 * @see Widget#checkSubclass
 * @see Widget#getStyle
 */
public Button(Composite parent, int style) {
	super(parent, checkStyle(style));
	OS.Button_SetAlignment(internal_handle, getAlignmentStyle(internal_style));
}
/**
 * Adds the listener to the collection of listeners who will
 * be notified when the control is selected, by sending
 * it one of the messages defined in the <code>SelectionListener</code>
 * interface.
 * <p>
 * <code>widgetSelected</code> is called when the control is selected.
 * <code>widgetDefaultSelected</code> is not called.
 * </p>
 *
 * @param listener the listener which should be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see SelectionListener
 * @see #removeSelectionListener
 * @see SelectionEvent
 */
public void addSelectionListener(SelectionListener listener) {
	checkWidget();
	if (listener == null) error(SWT.ERROR_NULL_ARGUMENT);
	TypedListener typedListener = new TypedListener(listener);
	addListener(SWT.Selection, typedListener);
	addListener(SWT.DefaultSelection, typedListener);
}
static int checkStyle(int style) {
	style = checkBits(style, SWT.PUSH, SWT.CHECK, SWT.RADIO, SWT.TOGGLE, 0, 0);
	if ((style & SWT.PUSH) != 0) {
		return checkBits(style, SWT.CENTER, SWT.LEFT, SWT.RIGHT, 0, 0, 0);
	}
	if ((style & (SWT.CHECK | SWT.RADIO | SWT.TOGGLE)) != 0) {
		return checkBits(style, SWT.LEFT, SWT.RIGHT, SWT.CENTER, 0, 0, 0);
	}
	return style;
}
private int getAlignmentStyle(int style) {
	int nativeStyle = 0;
	
	if ((style & SWT.LEAD) == SWT.LEAD) {
		nativeStyle = OS.STYLE_LEAD;
	} else if ((style & SWT.CENTER) == SWT.CENTER) {
		nativeStyle = OS.STYLE_CENTER;
	} else if ((style & SWT.TRAIL) == SWT.TRAIL) {
		nativeStyle = OS.STYLE_TRAIL;
	}
	return nativeStyle;
}
private void buttonCallback() {
    if ((internal_style & SWT.PUSH) == 0) {
		if ((internal_style & SWT.RADIO) != 0 && (internal_parent.getStyle() & SWT.NO_RADIO_GROUP) == 0) {
		    selectRadio();
		}
		else {
		    this.selection = !this.selection;
		}
    }
    postEvent(SWT.Selection);
}

void click() {
	internal_sendEvent(SWT.Selection);		
}

protected void internal_createHandle(int index) {
	/*
	 * Feature in QT. Buttons don't support alignment, so 
	 * we only need to check the flags for button type.
	 */			
	/*
	 * Another feature in QT - widgets are initially set to
	 * be their preferred sizes.Fix is to explictily set the 
	 * initial widget size to (0, 0). 
	 */
	 
	/* CHECK button */
	if ((internal_style & SWT.CHECK) != 0) {
		internal_handle = OS.Checkbox_New(internal_parent.internal_handle, internal_getNativeStyle());
		if (internal_handle == 0) error(SWT.ERROR_NO_HANDLES);
		return;
	}

	/* RADIO button */
	if ((internal_style & SWT.RADIO) != 0) {
		internal_handle = OS.RadioButton_New(internal_parent.internal_handle, internal_getNativeStyle());
		if (internal_handle == 0) error(SWT.ERROR_NO_HANDLES);
		return;
	}

	/* PUSH button */
	if ((internal_style & SWT.PUSH) != 0) {
		internal_handle = OS.PushButton_New(internal_parent.internal_handle, internal_getNativeStyle());
		if (internal_handle == 0) error(SWT.ERROR_NO_HANDLES);
		return;
	}

	/* TOGGLE button */
	if ((internal_style & SWT.TOGGLE) != 0) {
		internal_handle = OS.ToggleButton_New(internal_parent.internal_handle, internal_getNativeStyle());
		if (internal_handle == 0) error(SWT.ERROR_NO_HANDLES);
		return;
	}
	OS.Button_SetAlignment(internal_handle, getAlignmentStyle(internal_style));
}
protected int internal_getNativeStyle() {
	int nativeStyle = super.internal_getNativeStyle();
	if ((internal_style & SWT.FLAT) == SWT.FLAT) {
		nativeStyle |= OS.STYLE_FLAT;
	}
	return nativeStyle;
}
/**
 * Returns a value which describes the position of the
 * text or image in the receiver. The value will be one of
 * <code>LEFT</code>, <code>RIGHT</code> or <code>CENTER</code>.
 *
 * @return the alignment 
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public int getAlignment() {
	checkWidget();
	if (alignment != -1) return alignment;
	if ((internal_style & SWT.LEFT) != 0) return SWT.LEFT;
	if ((internal_style & SWT.CENTER) != 0) return SWT.CENTER;
	return SWT.RIGHT;
}

/**
 * Returns the receiver's image if it has one, or null
 * if it does not.
 *
 * @return the receiver's image
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public Image getImage() {
	checkWidget();
	return image;
}

/**
 * Returns <code>true</code> if the receiver is selected,
 * and false otherwise.
 * <p>
 * When the receiver is of type <code>CHECK</code> or <code>RADIO</code>,
 * it is selected when it is checked. When it is of type <code>TOGGLE</code>,
 * it is selected when it is pushed in. If the receiver is of any other type,
 * this method returns false.
 *
 * @return the selection state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public boolean getSelection() {
	checkWidget();
	if ((internal_style & SWT.CHECK) != 0 || (internal_style & SWT.RADIO) != 0 || (internal_style & SWT.TOGGLE) != 0) {
	    return selection;
    } else {
		return OS.Button_GetSelection(internal_handle);
    }
}
/**
 * Returns the receiver's text, which will be an empty
 * string if it has never been set.
 *
 * @return the receiver's text
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public String getText() {
	checkWidget();
	return text;
}

void releaseWidget() {
	// Unset the default button before the widget is 
	// disposed
	if (isDefault) {
		getShell().setDefaultButton(null);
	}	
	
	super.releaseWidget();
	image = null;
}

/**
 * Removes the listener from the collection of listeners who will
 * be notified when the control is selected.
 *
 * @param listener the listener which should no longer be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see SelectionListener
 * @see #addSelectionListener
 */
public void removeSelectionListener(SelectionListener listener) {
	checkWidget();
	if (listener == null) error(SWT.ERROR_NULL_ARGUMENT);
	if (eventTable == null) return;
	eventTable.unhook(SWT.Selection, listener);
	eventTable.unhook(SWT.DefaultSelection, listener);
}
void selectRadio() {
	/*
	* This code is intentionally commented.  When two groups
	* of radio buttons with the same parent are separated by
	* another control, the correct behavior should be that
	* the two groups act independently.  This is consistent
	* with radio tool and menu items.  The commented code
	* implements this behavior.
	*/
//	int index = 0;
//	Control [] children = parent.getChildren ();
//	while (index < children.length && children [index] != this) index++;
//	int i = index - 1;
//	while (i >= 0 && children [i].setRadioSelection (false)) --i;
//	int j = index + 1;
//	while (j < children.length && children [j].setRadioSelection (false)) j++;
//	setSelection (true);
    
	Control[] children = internal_parent.getChildren();
	for (int i = 0; i < children.length; i++) {
		Control child = children[i];
		if (this != child) child.setRadioSelection(false);
	}
	setSelection(true);
}
boolean setRadioSelection(boolean value) {
	if ((internal_style & SWT.RADIO) == 0) return false;
	if (getSelection() != value) {
		setSelection(value);
		postEvent(SWT.Selection);
	}
	return true;
}
/**
 * Controls how text and images will be displayed
 * in the receiver. The argument should be one of
 * <code>LEFT</code>, <code>RIGHT</code> or <code>CENTER</code>.
 *
 * @param alignment the new alignment 
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setAlignment(int alignment) {
	checkWidget();
	if ((alignment & (SWT.LEFT | SWT.RIGHT | SWT.CENTER)) == 0) return;
	if (this.alignment == alignment) return;
	this.alignment = alignment;
	OS.Button_SetAlignment(internal_handle, getAlignmentStyle(alignment));
}
void setDefault(boolean value) {
	isDefault = value;
}

/**
 * Sets the receiver's image to the argument, which may be
 * <code>null</code> indicating that no image should be displayed.
 *
 * @param image the image to display on the receiver (may be <code>null</code>)
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_INVALID_ARGUMENT - if the image has been disposed</li>
 * </ul> 
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setImage(Image image) {
	checkWidget();
	this.image = image;
	
	if (image == null) {
		OS.Button_SetImage(internal_handle, 0);
	} else {
		if (image.isDisposed()) error(SWT.ERROR_INVALID_ARGUMENT);
		OS.Button_SetImage(internal_handle, image.internal_handle);
	}
}
/**
 * Sets the selection state of the receiver, if it is of type <code>CHECK</code>, 
 * <code>RADIO</code>, or <code>TOGGLE</code>.
 *
 * <p>
 * When the receiver is of type <code>CHECK</code> or <code>RADIO</code>,
 * it is selected when it is checked. When it is of type <code>TOGGLE</code>,
 * it is selected when it is pushed in.
 *
 * @param selected the new selection state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setSelection(boolean selected) {
	checkWidget();
	if ((internal_style & SWT.CHECK) != 0 || (internal_style & SWT.RADIO) != 0 || (internal_style & SWT.TOGGLE) != 0) {
		selection = selected;
		OS.Button_SetSelection(internal_handle, selected);
	}
}
/**
 * Sets the receiver's text.
 * <p>
 * This method sets the button label.  The label may include
 * the mnemonic character but must not contain line delimiters.
 * </p>
 * <p>
 * Mnemonics are indicated by an '&amp' that causes the next
 * character to be the mnemonic.  When the user presses a
 * key sequence that matches the mnemonic, a selection
 * event occurs. On most platforms, the mnemonic appears
 * underlined but may be emphasised in a platform specific
 * manner.  The mnemonic indicator character '&amp' can be
 * escaped by doubling it in the string, causing a single
 *'&amp' to be displayed.
 * </p>
 * 
 * @param string the new text
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the text is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setText(String string) {
	checkWidget();
	if (string == null) error (SWT.ERROR_NULL_ARGUMENT);
	text = string;
	OS.Button_SetLabel(internal_handle, string);
	OS.Button_SetImage(internal_handle,0);
}

public boolean allowTraverseByArrowKey(Event event) {
	switch (event.keyCode) {
		case SWT.ARROW_UP :
		case SWT.ARROW_LEFT :
		case SWT.ARROW_DOWN :
		case SWT.ARROW_RIGHT :
		return true;
	}
	return false;
}

protected boolean traverse(Event event) {
	if (isDisposed()) return false;
	if(allowTraverseByArrowKey(event)) {
		return traverseByArrowKey(event);
	}
	return super.traverse(event);
}

}
