/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"
#include "UGL_Win32_Widget.h"
#include "UGL_win32_display.h"
#include "UGL_Win32_DC.h"

#include "Widget.h"
#include "Control.h"
#include "Choice.h"
#include "Combo.h"
#include "VerifyHelpers.h"

//========================================================================================
#define CBID_EDIT		1001
#define CBID_LIST		1000
static UGL_Int _Combo_ExStyle(UGL_Int ugl_style);
static UGL_Int _Combo_Style(UGL_Int ugl_style);
void _Dispose_Combo(UGL_Int handle, UGL_Error uglError);
int _Combo_CTLCOLOR(PUGL_Control parent_control, HDC hdc, PUGL_Control color_control);
LRESULT _Combo_CommandChild(PUGL_Control ugl_control, WPARAM wParam, HWND hWnd);
void _Combo_SetBounds(PUGL_Control ugl_control, LPRECT bounds);

static LRESULT CALLBACK _Combo_Proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void _Combo_GetMinimumSize(UGL_Int handle, POINT *point, UGL_Error uglError);
LRESULT CALLBACK _Combo_Text_Proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK _Combo_List_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);

UGL_Boolean _Combo_SendKeyEvent(PUGL_Control control, UGL_Int type, UGL_Char character, UGL_Int keycode, UGL_Int modifiers);

UGL_Boolean _Combo_Text_ProcessWmChar(PUGL_Control ugl_control, TCHAR ch);
void _Combo_Text_CombineNewString(TCHAR *curStr, TCHAR *newStr, int pos1, int pos2, TCHAR ch);
UGL_Boolean _Combo_Text_FindAndSelectItem(HWND hwnd, LPCSTR str);
void _Combo_Text_RestoreText(PUGL_Control ugl_control);
void _Combo_Text_SaveText(PUGL_Control ugl_control);
//========================================================================================

void Combo_ReplaceText(UGL_Int handle, UGL_String text, int start, int end, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;
	HWND textHWnd = GetDlgItem(ugl_combo->hWnd, CBID_EDIT);
	
	SendMessage(textHWnd, EM_SETSEL, start, end);
	SendMessage(textHWnd, EM_SCROLLCARET, 0, 0);

	UGL_COMBO(handle)->ignoreChange = TRUE;
	SendMessage(textHWnd, EM_REPLACESEL, (WPARAM)TRUE, (LPARAM)text);
	UGL_COMBO(handle)->ignoreChange = FALSE;
}

//========================================================================================

UGL_String Combo_GetText(UGL_Int handle, UGL_Boolean* shouldFreeResult, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;
	UGL_String text = NULL;
	int length = GetWindowTextLength(ugl_combo->hWnd);
	
	if (length == 0) {
		*shouldFreeResult = FALSE;
		return _T("");
	}

	text = (UGL_String)calloc(sizeof(TCHAR), length+1);
	if (text == NULL) {
		Win32Error_SetError(uglError, UGL_ERROR_OUT_OF_MEMORY, _T("Unable to allocate memory for text"));
		return (UGL_String)NULL;
	}

	GetWindowText(ugl_combo->hWnd, (LPTSTR)text, length + 1);
	
	*shouldFreeResult = TRUE;
	return text;
}
//========================================================================================

UGL_Int Combo_GetDefaultTextLimit(UGL_Error uglError) {
	OSVERSIONINFO info;
		
	GetVersionEx((LPOSVERSIONINFO)&info);
	return (info.dwPlatformId == VER_PLATFORM_WIN32_NT) ? 0x7FFFFFFF : 0x7FFF;
}
//========================================================================================

UGL_Int Combo_GetTextHeight(UGL_Int handle, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;

	int result = SendMessage(ugl_combo->hWnd, CB_GETITEMHEIGHT, -1, 0);
	if (result == CB_ERR) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get item height"));
		return 0;
	}
	
	// why the +6? SWT does it, no other reason...
	return result + 6;
}
//========================================================================================

void Combo_SetTextLimit(UGL_Int handle, UGL_Int limit, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;

	if (limit == 0) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Limit can not be 0 for combo"));
		return;
	}

	SendMessage(ugl_combo->hWnd, CB_LIMITTEXT, limit, 0);
}
//========================================================================================

UGL_Int Combo_GetItemHeight(UGL_Int handle, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;

	int result = SendMessage(ugl_combo->hWnd, CB_GETITEMHEIGHT, 0, 0);
	if (result == CB_ERR) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get item height from combo"));
		return 0;
	} else {
		return result;
	}
}
//========================================================================================

void Combo_DeselectItem(UGL_Int handle, UGL_Int index, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;
	int selection = SendMessage(ugl_combo->hWnd, CB_GETCURSEL, 0, 0);

	if (index != selection) return;
	SendMessage(ugl_combo->hWnd, CB_SETCURSEL, -1, 0);
}
//========================================================================================

void Combo_DeselectAllItems(UGL_Int handle, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;
	
	SendMessage(ugl_combo->hWnd, CB_SETCURSEL, -1, 0);
}
//========================================================================================

void Combo_SetText(UGL_Int handle, UGL_String string, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;

	// this is desired behavior?
	if (((PUGL_Combo)ugl_combo)->is_read_only) {
		//int index = indexOf (string);
		//if (index != -1) select (index);
		return;
	}

	SetWindowText(ugl_combo->hWnd, string);
}
//========================================================================================

void Combo_SetOrientation(UGL_Int handle, UGL_Int orientation, UGL_Error uglError) {
	PUGL_Combo ugl_combo = UGL_COMBO(handle);

#ifdef _WIN32_WCE
	return;
#else
	//if ((OS.WIN32_MAJOR << 16 | OS.WIN32_MINOR) < (4 << 16 | 10)) return;
	int flags = UGL_STYLE_LEFT_TO_RIGHT | UGL_STYLE_RIGHT_TO_LEFT;
	int bits = 0;
	HWND hwndList, hwndText, hwnd;

	hwnd = UGL_CONTROL(ugl_combo)->hWnd;
	if ((orientation & flags) == 0 || (orientation & flags) == flags) return;
	UGL_WIDGET(ugl_combo)->ugl_style &= ~flags;
	UGL_WIDGET(ugl_combo)->ugl_style |= orientation & flags;
	
	bits = GetWindowLong(hwnd, GWL_EXSTYLE);
	if (IS_FLAG_SET(orientation, UGL_STYLE_RIGHT_TO_LEFT)) {
		//style |= SWT.MIRRORED;
		bits |= WS_EX_LAYOUTRTL;
	} else {
		//style &= ~SWT.MIRRORED;
		bits &= ~WS_EX_LAYOUTRTL;
	}

	SetWindowLong(hwnd, GWL_EXSTYLE, bits);

	hwndList = GetDlgItem(hwnd, CBID_LIST);
	hwndText = GetDlgItem(hwnd, CBID_EDIT);

	if (hwndText != NULL) {
		int bits_exstyle = GetWindowLong(hwndText, GWL_EXSTYLE);
		int bits_style = GetWindowLong(hwndText, GWL_STYLE);
		if (IS_FLAG_SET(orientation, UGL_STYLE_RIGHT_TO_LEFT)) {
			bits_exstyle |= WS_EX_RIGHT | WS_EX_RTLREADING;
			bits_style |= ES_RIGHT;
		} else {
			bits_exstyle &= ~(WS_EX_RIGHT | WS_EX_RTLREADING);
			bits_style &= ~ES_RIGHT;
		}
		SetWindowLong(hwndText, GWL_EXSTYLE, bits_exstyle);
		SetWindowLong(hwndText, GWL_STYLE, bits_style);
		
		/*
		* Bug in Windows.  For some reason, the single line text field
		* portion of the combo box does not redraw to reflect the new
		* style bits.  The fix is to force the widget to be resized by
		* temporarily shrinking and then growing the width and height.
		*/

		InvalidateRect(hwndText, NULL, TRUE);
		//RECT rect = new RECT ();
		//OS.GetWindowRect (hwndText, rect);
		//int width = rect.right - rect.left, height = rect.bottom - rect.top;
		//OS.GetWindowRect (handle, rect);
		//int widthCombo = rect.right - rect.left, heightCombo = rect.bottom - rect.top;
		//int uFlags = OS.SWP_NOMOVE | OS.SWP_NOZORDER | OS.SWP_NOACTIVATE;
		//SetWindowPos (hwndText, 0, 0, 0, width - 1, height - 1, uFlags);
		//SetWindowPos (handle, 0, 0, 0, widthCombo - 1, heightCombo - 1, uFlags);
		//SetWindowPos (hwndText, 0, 0, 0, width, height, uFlags);
		//SetWindowPos (handle, 0, 0, 0, widthCombo, heightCombo, uFlags);
		//OS.InvalidateRect (handle, null, true);
	}	
	if (hwndList != NULL) {
		int bits_exstyle = GetWindowLong(hwndList, GWL_EXSTYLE);		
		if (IS_FLAG_SET(orientation, UGL_STYLE_RIGHT_TO_LEFT)) {
			bits_exstyle |= WS_EX_LAYOUTRTL;
		} else {
			bits_exstyle &= ~WS_EX_LAYOUTRTL;
		}
		SetWindowLong(hwndList, GWL_EXSTYLE, bits_exstyle);
	}
#endif // defined _WIN32_WCE
}
//========================================================================================

void Combo_SetTextSelection(UGL_Int handle, UGL_Int start, UGL_Int end, UGL_Error uglError) {
	PUGL_Combo ugl_combo = UGL_COMBO(handle);

	int bits = start | (end << 16);
	SendMessage(UGL_CONTROL(ugl_combo)->hWnd, CB_SETEDITSEL, 0, bits);
}
//========================================================================================

UGL_Int Combo_New(UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	return _Combo_New(parentHandle, ugl_style, uglError);
}

UGL_Int _Combo_New(UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	int style = _Combo_Style(ugl_style);
	int extended_style = _Combo_ExStyle(ugl_style);
	HWND textHWnd, listHWnd;

	PUGL_Control ugl_control = _Control_New(_T("COMBOBOX"), sizeof(UGL_Combo), _Combo_Proc, 
						parentHandle, ugl_style, style, extended_style, uglError);

// TODO check return conditions!
	UGL_COMBO(ugl_control)->is_read_only = (IS_FLAG_SET(ugl_style, UGL_STYLE_READ_ONLY));

	UGL_WIDGET(ugl_control)->widget_dispose = _Dispose_Combo;
	ugl_control->sendKeyEvent = _Combo_SendKeyEvent;
	
	textHWnd = GetDlgItem(ugl_control->hWnd, CBID_EDIT);
	listHWnd = GetDlgItem(ugl_control->hWnd, CBID_LIST);

	if (textHWnd != NULL) {
		SetWindowLong(textHWnd, GWL_USERDATA, (LONG)ugl_control);
		UGL_COMBO(ugl_control)->textProc = (WNDPROC)GetWindowLong(textHWnd, GWL_WNDPROC);
		SetWindowLong(textHWnd, GWL_WNDPROC, (LONG)(_Combo_Text_Proc));
	}

	if (listHWnd != NULL) {
		/*
		 * listHWnd is almost always NULL
		 */
		SetWindowLong(listHWnd, GWL_USERDATA, (LONG)ugl_control);
		UGL_COMBO(ugl_control)->listProc = (WNDPROC)GetWindowLong(listHWnd, GWL_WNDPROC);
	}

	/*
	 * We default to 7 for PPro to mimic the JDK. SWT prefers 5, but it isn't spec'd.
	 */
	UGL_COMBO(ugl_control)->visibleItems = 7;

	return (UGL_Int)ugl_control;
}
//========================================================================================

void Combo_GetTextSelection(UGL_Int handle, UGL_Int* returnX, UGL_Int* returnY, UGL_Error uglError) {
	PUGL_Control ugl_combo = (PUGL_Control)handle;

	//if ((style & SWT.DROP_DOWN) != 0 && (style & SWT.READ_ONLY) != 0) {
	if (((PUGL_Combo)ugl_combo)->is_read_only) {
		*returnX = 0;
		*returnY = GetWindowTextLength(ugl_combo->hWnd);
	}

	SendMessage(ugl_combo->hWnd, CB_GETEDITSEL, (WPARAM)returnX, (LPARAM)returnY);
}
//========================================================================================

void Combo_SetItem(UGL_Int handle, UGL_String string, UGL_Int index, UGL_Error uglError) {
	Choice_RemoveItem(handle, index, uglError);
	Choice_AddItem(handle, string, index, uglError);
}
//========================================================================================

UGL_Int Combo_GetVisibleItemCount(UGL_Int handle, UGL_Error uglError) {
	PUGL_Combo ugl_control = UGL_COMBO(handle);
	return ugl_control->visibleItems;

}
//========================================================================================

void Combo_SetVisibleItemCount(UGL_Int handle, UGL_Int count, UGL_Error uglError) {
	PUGL_Combo ugl_combo = UGL_COMBO(handle);
	PUGL_Control ugl_control = (PUGL_Control)ugl_combo;
	int numItems = SendMessage(ugl_control->hWnd, CB_GETCOUNT, 0, 0);
	if (count < 0) return; //error
	ugl_combo->visibleItems = count; //don't care if it's greater than numItems and don't have to repaint
	
}
//========================================================================================

UGL_Int _Combo_Style(UGL_Int ugl_style) {
	int bits = _Control_Style(ugl_style);
	
	bits |= CBS_AUTOHSCROLL | CBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP;

	//if (IS_FLAG_SET(ugl_style, UGL_STYLE_SIMPLE)) return bits | CBS_SIMPLE;
	if (IS_FLAG_SET(ugl_style, UGL_STYLE_READ_ONLY)) return bits | CBS_DROPDOWNLIST;
	return bits | CBS_DROPDOWN;
}
//========================================================================================

UGL_Int _Combo_ExStyle(UGL_Int ugl_style) {
	int bits = 0; 
	return bits;
}
//=======================================================================================

LRESULT CALLBACK _Combo_Proc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	PUGL_Combo ugl_combo = (PUGL_Combo)GetWindowLong(hwnd, GWL_USERDATA);
	switch (message) {
		case WM_CTLCOLOREDIT:
		case WM_CTLCOLORLISTBOX: {
			HWND color_widget_hWnd = (HWND)lParam;
			PUGL_Control color_widget = (PUGL_Control)GetWindowLong(color_widget_hWnd, GWL_USERDATA);
			PUGL_Control parent_control = (PUGL_Control)GetWindowLong(hwnd, GWL_USERDATA);
			if (color_widget == NULL) {
				return _Combo_CTLCOLOR(parent_control, (HDC)wParam, color_widget);
			}
		} break;
		case UI_GET_MINIMUM_SIZE: {
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hwnd, GWL_USERDATA);
			UGL_Int handle = (UGL_Int)ugl_control;
			POINT* point = (POINT*)wParam;
			_Combo_GetMinimumSize(handle, point, (UGL_Error)lParam);
			return 0;
		} break;
		case UI_COMMAND_CHILD: {
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hwnd, GWL_USERDATA);
			return _Combo_CommandChild(ugl_control, wParam, hwnd);
		} break;
		case UI_SET_BOUNDS: {
			/*
			 * What's going on here is that when we call SetWindowPos() on a Combo, 
			 * Windows takes those bounds and constrains the whole Combo, including the
			 * drop-down list to them. So if we pass along the size that we want for the
			 * 'minimized' Combo, then it will be useless because the drop-down list will
			 * never be visible because it will go outside of the bounds we're giving to 
			 * Windows.
			 * The fix is to return the minimum size of an un-expanded Combo box in 
			 * _Combo_GetMinimumSize(), but to always override the height passed to 
			 * SetBounds and set it to the size necessary to show the elements of the list.
			 * This way, the list is expandable, but the layouts to think that the Combo
			 * is the size of the text plus the list.
			 */
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hwnd, GWL_USERDATA);
			UGL_Int handle = (UGL_Int)ugl_control;
			LPRECT bounds = (LPRECT)wParam;
			_Combo_SetBounds(ugl_control, bounds);

		} break; //we want to let this pass through to ultimately go through _Control_SetBounds
		case WM_KILLFOCUS: {
			/*
			 * When _Control_ForceFocus() is called on a Combo, the native call to 
			 * setFocus() sends a WM_SETFOCUS message to the combo, and it immediately
			 * bounces the focus to the text component, which causes a WM_KILLFOCUS message
			 * to be sent to the Combo and a WM_SETFOCUS to go to the text component.
			 * This results in three focus callbacks reported by the combo in java: 
			 * (focus in, focus out, focus in). 
			 * So here and in WM_SETFOCUS, we suppress the callbacks for the first two messages
			 * using the fact that the WPARAM parameter is the HWND of the window gaining or 
			 * losing focus (gaining on a WM_KILLFOCUS, losing on a WM_SETFOCUS)
			 */
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hwnd, GWL_USERDATA);
			HWND lastFocusWind = (HWND)wParam;
			PUGL_Control lastFocusControl = (PUGL_Control)GetWindowLong(lastFocusWind, GWL_USERDATA);
			UGL_Boolean ignoreFocusEvent = (ugl_control == lastFocusControl);
			if (ignoreFocusEvent) {
				if (ugl_control->widgetProc != NULL) return CallWindowProc(ugl_control->widgetProc, hwnd, message, wParam, lParam);
				else return DefWindowProc(hwnd, message, wParam, lParam);
			}
		} break;
		case WM_SETFOCUS: {
			PUGL_Control ugl_control = (PUGL_Control)ugl_combo;
			HWND lastFocusWind = (HWND)wParam;
			int selection, count, style;
			PUGL_Control lastFocusControl = (PUGL_Control)GetWindowLong(lastFocusWind, GWL_USERDATA);
			UGL_Boolean ignoreFocusEvent = (ugl_control == lastFocusControl);
			if (ignoreFocusEvent) {
				if (ugl_control->widgetProc != NULL) return CallWindowProc(ugl_control->widgetProc, hwnd, message, wParam, lParam);
				else return DefWindowProc(hwnd, message, wParam, lParam);
			}
			
			//Fot read-only combo, while combo is focused, to select the first item if it existes and nothing was selected before.
			style = GetWindowLong(ugl_control->hWnd, GWL_STYLE);
			if (IS_FLAG_SET(style, CBS_DROPDOWNLIST)) {
				count = SendMessage(ugl_control->hWnd, CB_GETCOUNT, 0, 0);
				if (count > 1) {
					selection = SendMessage(ugl_control->hWnd, CB_GETCURSEL, 0, 0);
					if (selection < 0)
			         SendMessage(ugl_control->hWnd, CB_SETCURSEL, 0, 0);
				}
			}

		} break;

	}
	return _Control_Proc(hwnd, message, wParam, lParam);
}

//========================================================================================

void _Combo_SetBounds(PUGL_Control ugl_control, LPRECT bounds){
	/*
	 * The height should be the height of an item times the user-requested number of
	 * visible items. If that number exceeds the actual number of items in Combo, we 
	 * still make the Combo this size and Windows fills up the difference with white space.
	 */
	HWND hwnd = ugl_control->hWnd;
	UGL_Int height = SendMessage(hwnd, CB_GETITEMHEIGHT, 1, 0);
	int visibleNum = ((PUGL_Combo)ugl_control)->visibleItems;
	height = height * visibleNum;
	bounds->bottom = bounds->bottom + height + 2;
}

UGL_Boolean _Combo_SendKeyEvent(PUGL_Control control, UGL_Int type, UGL_Char character, UGL_Int keycode, UGL_Int modifiers) {
	return VerifyHelper_SendKeyEvent(control, GetDlgItem(control->hWnd, CBID_EDIT), type, character, keycode, modifiers);
}

//========================================================================================
LRESULT CALLBACK _Combo_List_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	/*
	 * I don't think we ever get a handle on the List Proc, but if we did, we would point it here
	 */
	PUGL_Combo ugl_combo = (PUGL_Combo)GetWindowLong(hWnd, GWL_USERDATA);
	if (ugl_combo != NULL) return ugl_combo->listProc(hWnd, message, wParam, lParam);
	return 0;
}
//========================================================================================
LRESULT CALLBACK _Combo_Text_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	/*
	 * This is set in _Combo_New()
	 *
	 * Because Combos are complex widgets, some messages go to the HWND of the combo,
	 * and some go to the HWND of the text component of the combo.
	 * This proc handles some of the messages that go to the text component.
	 */ 
	PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA); //this is the parent (the Combo)
	int result = 0;
	switch (message) {
		case WM_KEYDOWN: {
			if(SendMessage(ugl_control->hWnd, CB_GETDROPPEDSTATE, 0, 0) && wParam == VK_RETURN) {
				return 0;
			}
			result = _Control_ProcessWmKeyDown(ugl_control->hWnd, wParam, lParam);
			if(wParam == VK_DELETE) {
				_Combo_Text_ProcessWmChar(ugl_control, VK_DELETE);
				return 0;
			}
		} break;

		case WM_KEYUP: {
			if(SendMessage(ugl_control->hWnd, CB_GETDROPPEDSTATE, 0, 0) && wParam == VK_RETURN) {
				return 0;
			}
			result = _Control_ProcessWmKeyUp(ugl_control->hWnd, wParam, lParam);
		} break;
		
		case WM_CHAR: {
			if(wParam != VK_RETURN) {
				result = _Control_ProcessWmChar(ugl_control->hWnd, wParam, lParam);
			}
			switch (wParam){
				case VK_TAB: {
					return 0; 
				} break;
				case VK_RETURN: {
					if(SendMessage(ugl_control->hWnd, CB_GETDROPPEDSTATE, 0, 0)) {
						_Combo_Text_ProcessWmChar(ugl_control, VK_RETURN);
					} else {
						MessageBeep(MB_OK);
						result = _Control_ProcessWmChar(ugl_control->hWnd, wParam, lParam);
						ComboCallback(ugl_control->widget.callback_target, UGL_EVENT_COMBO_DEFAULTSELECTION);
					}
					return 0; 
				} break;
				case VK_ESCAPE: {
					int style = GetWindowLong(hWnd, GWL_STYLE);
					SendMessage(ugl_control->hWnd, CB_SHOWDROPDOWN, FALSE, 0);
					if ((style & UGL_STYLE_DROP_DOWN) != 0) {
						if (SendMessage(hWnd, CB_GETDROPPEDSTATE, wParam, lParam) == 0){
							return 0;
						};
						
					}
				} break;
				default : {
					_Combo_Text_ProcessWmChar(ugl_control, wParam);
					return 0;
				} break;
			}
		} break;

		case WM_CLEAR:
		case WM_CUT: {
			if (VerifyHelper_HandleCut(ugl_control, hWnd, ((PUGL_Combo)ugl_control)->textProc, message) == FALSE) return 0;
			break;
		}

		case WM_PASTE: {
			if (VerifyHelper_HandlePaste(ugl_control, hWnd) == FALSE) return 0;
			break;
		}

		case WM_SETFOCUS: {
			/*
			 * When _Control_ForceFocus() is called on a Combo, the native call to 
			 * setFocus() sends a WM_SETFOCUS message to the combo, and it immediately
			 * bounces the focus to the text component, which causes a WM_KILLFOCUS message
			 * to be sent to the Combo and a WM_SETFOCUS to go to the text component.
			 * This results in three focus callbacks reported by the combo in java: 
			 * (focus in, focus out, focus in). 
			 */
			HWND lastFocusWind = (HWND)wParam;
			UGL_Boolean ignoreFocusEvent = (lastFocusWind == ugl_control->hWnd);

			_Combo_Text_RestoreText(ugl_control);

			if (!ignoreFocusEvent) {
                FocusCallback(UGL_WIDGET(ugl_control)->callback_target, UGL_EVENT_FOCUS_IN);
				//only send the event if the combo didn't already do it
			}
		} break;

		case WM_KILLFOCUS: {
			HWND lastFocusWind = (HWND)wParam;
			UGL_Boolean ignoreFocusEvent = (lastFocusWind == ugl_control->hWnd);

			_Combo_Text_SaveText(ugl_control);

            if (!ignoreFocusEvent) {
                FocusCallback(UGL_WIDGET(ugl_control)->callback_target, UGL_EVENT_FOCUS_OUT);
				//only send the event if the combo didn't already do it
			}
		} break;
	}
	
	if (result != 0) return result;
	else return CallWindowProc(((PUGL_Combo)ugl_control)->textProc, hWnd, message, wParam, lParam);
}
//========================================================================================

LRESULT _Combo_CommandChild(PUGL_Control ugl_control, WPARAM wParam, HWND hWnd){
	UGL_Int handle = (UGL_Int)ugl_control;
	PUGL_Combo ugl_combo = UGL_COMBO(handle);
	LRESULT result = 1;
	HWND textHWnd = GetDlgItem(ugl_control->hWnd, CBID_EDIT);
	HWND listHWnd = GetDlgItem(ugl_control->hWnd, CBID_LIST);
	int code = HIWORD(wParam);
	int current_control = LOWORD(wParam);
	/*
	 * This is weird. For starters, we have a really quarky way of sizing Combos.
	 * We have to return the size of an unexpanded Combo from _Combo_GetMinimumSize()
	 * so that the Java layouts work correctly. But we have to return the size of an
	 * expanded Combo from _Control_GetBounds(), otherwise the OS won't give us enough room
	 * to show an expanded Combo. We do this by causing Control_SetBounds() to go through the 
	 * procs and override UI_SET_BOUNDS in _Combo_Proc to take the bounds set in _Combo_GetMinimumSize()
	 * and increase the height to account for the expanded Combo.
	 * That brings us to the next few lines of code. They are necessary because of a Windows bug. It 
	 * seems that no matter what we set the bounds of the window to be in UI_SET_BOUNDS, and eventually
	 * _Control_SetBounds(), Windows always resets the height according to some other constraint. But we 
	 * have to control this to satisfy the eSWT Get/SetVisibleItemCount() API. So the fix is to resize the
	 * drop-down window after Windows ignores our request, which we do here, right after the user presses 
	 * the drop-down button.
	 */
	int x, y, width, height = 0;
	RECT boundsRect;
	UGL_Error error = NULL;
    Control_GetBounds((UGL_Int)ugl_control, &x, &y, &width, &height, error);
	boundsRect.left = x;
	boundsRect.right = x + width;
	boundsRect.top = y;
	boundsRect.bottom = y + height;
	SendMessage(ugl_control->hWnd, UI_SET_BOUNDS, (WPARAM)&boundsRect, (LPARAM)&error);

	switch (code) {
		case CBN_SELCHANGE: {
			int current_item = SendMessage(hWnd, CB_GETCURSEL, 0, 0);
			
			if (current_item == 0) {
				int count = SendMessage(hWnd, CB_GETCOUNT, 0, 0);
				if (count == 0) current_item = -1;
			}

			if (current_control != CBID_LIST) {
				/*
				 * This is necessary because this message comes through twice per item
				 * selection: once for the list and once for something else.
				 * It doesn't seem that the second one is for the text, because 
				 * the value of current_control is 0, not CBID_TEXT (1001). 
				 * Regardless of why the second call comes through, we're only interested
				 * in the one for the list control. 
				 * Also note that we don't change the result variable, which means that 
				 * processing on this message will end here, but that's OK because the
				 * processing that occurs for the list control is sufficient.
				 */
				break; 
			}

			ChoiceCallback(UGL_WIDGET(ugl_combo)->callback_target, current_item);
			//don't mess with result -- if we return 0 we skip important processings

		} break;

		case CBN_EDITCHANGE: {
		} break;

		case CBN_SELENDOK: {
			//This called whenever the user selects an item from the list 
			//(using the mouse or the keyboard)
			ComboCallback(UGL_WIDGET(ugl_combo)->callback_target, UGL_EVENT_COMBO_TEXT);	
		} break;
		case CBN_SETFOCUS: {
			_Combo_Text_RestoreText(ugl_control);
		} break;
		case CBN_KILLFOCUS: {
			_Combo_Text_RestoreText(ugl_control);
		} break;
		case CBN_DROPDOWN: {
			_Combo_Text_RestoreText(ugl_control);
			return 0;
		} break;
		case CBN_CLOSEUP: {
			_Combo_Text_RestoreText(ugl_control);
			return 0;
		} break;
		case CBN_SELENDCANCEL: {
			if(!UGL_COMBO(ugl_control)->is_read_only &&
				SendMessage(ugl_control->hWnd, CB_GETDROPPEDSTATE, 0, 0)) {
				_Combo_Text_SaveText(ugl_control);
			}
		} break; 
		default:{
				
				} break;
	}
	return result;
	
}

//========================================================================================

int _Combo_CTLCOLOR(PUGL_Control parent_control, HDC hdc, PUGL_Control color_control){
	/*
	 * This function gets called whenever a WM_CTLCOLOREDIT or WM_CTLCOLORLISTBOX messsage
	 * comes through the ComboProc and doesn't map to a UGL_Control. It means that the
	 * current control is the child text or child list of a combo. That means that we need
	 * to get the correct background and foreground color information from the parent, where
	 * it gets stored on a Control_SetBackground() or Control_SetForeGround() call.
	 */
	
	COLORREF foreground;
	COLORREF background;
	HGDIOBJ background_brush;

	/*
		* We don't have any special logic to set the foreground
		* because all the controls use COLOR_WINDOWTEXT.
		*/
	foreground = (parent_control->foreground_color != NULL) ? parent_control->foreground_color->color : GetSysColor(COLOR_WINDOWTEXT);

	if (parent_control->background_color != NULL) {
		background_brush = Color_GetBrush(parent_control->background_color);
		background = parent_control->background_color->color;
	} else {
		background_brush = GetSysColorBrush(COLOR_WINDOW);
		background = GetSysColor(COLOR_WINDOW);

	}

	SetTextColor(hdc, foreground);
	SetBkColor(hdc, background);
	return (int)background_brush;
}

//========================================================================================

void _Combo_GetMinimumSize(UGL_Int handle, POINT *point, UGL_Error uglError) {
	/*
	 * This is a rough mimic of how they do things in SWT.
	 * The logic goes something like this:
	 * For width: find the longest item in the combo, draw that string
	 * to an HDC using the current font, figure out the width of the rect
	 * that surrounds the text you just drew, and add the width of the drop-
	 * down button plus a little buffer just so things aren't cramped.
	 * For height: find the height of the button on the combo, plus a little breathing room. 
	 * Don't worry about how many items are in the drop-down list. We take care of that in
	 * SetBounds(). If we return that size here, then the layouts will always
	 * treat the Combo like its size is its expanded size.
	 * The exceptions are that we only show the first six items in the combo
	 * if there are more than 6, and we scale to the width of the second-
	 * longest item in the combo if the longest item is 15 or more chars longer
	 * than the second-longest.
	 * 
	 * We probably need to do a bit of error handling on the OS calls.
	 * We use a default CBSize of 52, that is hard-coded in SWT. That may 
	 * cause some problems sooner or later.
	 */
	
	PUGL_Control ugl_control = UGL_CONTROL(handle);
	HWND hWnd = ugl_control->hWnd;
	int width = 0;
	int height = 0;
	int num_items = SendMessage(hWnd, CB_GETCOUNT, 0, 0);
	int max_length = 0;
	int current_length = 0;
	int second_longest = 0;
	TEXTMETRIC tm;
	HDC hDC = GetDC(hWnd);
	int working_vertical_count = 0;
	int i = 0;
	int max_viewable_items = 6;
	COMBOBOXINFO pcbi;
	int presetCBSize = 52; //from SWT Default
	RECT text_size_rect;
	HGDIOBJ newFont = 0;
	int flags = DT_CALCRECT | DT_NOPREFIX;
	int first_index = 0;
	int second_index = 0;
	int widthBuffer = 10; //based on testing
	int heightBuffer = 3;

	
	/*
	 * Find and record the number of chars and indexes of the longest and second-
	 * longest strings in the combo
	 */
	for (i = 0; i<num_items; i++){
		current_length = SendMessage(hWnd, CB_GETLBTEXTLEN, i, 0);
		if (current_length > max_length) {
			second_longest = max_length;
			max_length = current_length;
			second_index = first_index;
			first_index = i;
		}
	}

	/*
	 * Now determine whether you want to use the longest or second-longest, and
	 * find the width in pixels, based on drawing that string to the screen in the
	 * current font. 
	 * No need to bother with all the overhead if we know that the combo is empty.
	 */
	if (num_items != 0) {
		UGL_String text;
		int working_horizontal_count = 0;

		if (second_longest != 0 && (max_length - second_longest) > 15) {
			working_horizontal_count = second_longest;
			text = (UGL_String)calloc(sizeof(TCHAR), working_horizontal_count+1);
			SendMessage(hWnd, CB_GETLBTEXT, second_index, (LPARAM)text);
			/*SendMessage(hWnd, CB_SETDROPPEDWIDTH, tm.tmAveCharWidth * max_length * 0.80, 0);*/
		}
		else { 
			working_horizontal_count = max_length;
			text = (UGL_String)calloc(sizeof(TCHAR), working_horizontal_count+1);
			SendMessage(hWnd, CB_GETLBTEXT, first_index, (LPARAM)text);
		}

		newFont = (HGDIOBJ)SendMessage(hWnd, WM_GETFONT, 0, 0);
		SelectObject(hDC, newFont);
		DrawText(hDC, text, working_horizontal_count, (LPRECT)&text_size_rect, flags);
		width = text_size_rect.right - text_size_rect.left;
		width += widthBuffer; //Just so everything isn't really cramped
		free(text);
	}
	else {
		/*
		 * Since the combo is empty, make it the size of a six-character string
		 * in the default font.
         */
		GetTextMetrics(hDC, (LPTEXTMETRIC)&tm);
		width = tm.tmAveCharWidth * 6;
	}

	pcbi.cbSize = presetCBSize;

	/**
	 * WinCE does not have the GetComboBoxInfo method so the message has to be
	 * used directly instead.
	 */
#ifdef _WIN32_WCE
	if (SendMessage(hWnd, CB_GETCOMBOBOXINFO, (WPARAM)0, (LPARAM)&pcbi)) {
#else
	if (GetComboBoxInfo(hWnd, &pcbi)) {
#endif
		width += pcbi.rcItem.left + (pcbi.rcButton.right - pcbi.rcButton.left);
		height = pcbi.rcButton.bottom - pcbi.rcButton.top;
		height += heightBuffer;
	}
	else {
		/*
		 * If the call fails, which it shouldn't, just return the default values
		 * that _Control_GetMinimumSize returns.
		 */
		width = 64;
		height = 64;
	}

	ReleaseDC (hWnd, hDC);

	point->x = width;
	point->y = height;
	
}

//========================================================================================

void _Dispose_Combo(UGL_Int handle, UGL_Error uglError){
	PUGL_Combo ugl_combo = UGL_COMBO(handle);
	if(ugl_combo->text != NULL)
		free(ugl_combo->text);
	_Dispose_Scrollable(handle, uglError);
}

UGL_Boolean _Combo_Text_ProcessWmChar(PUGL_Control ugl_control, TCHAR ch) {
	TCHAR *curStr=NULL, *newStr=NULL;
	UGL_Boolean modify = FALSE;
	DWORD rs;
	int start, end, len;

	rs = (DWORD)SendMessage(ugl_control->hWnd, CB_GETEDITSEL, 0, 0);
	start = LOWORD(rs);
	end = HIWORD(rs);

	len = GetWindowTextLength(ugl_control->hWnd);
	curStr = (TCHAR *)calloc(len + 1, sizeof(TCHAR));
	if(curStr == NULL)
		return FALSE;
	newStr = (TCHAR *)calloc(len + 2, sizeof(TCHAR));
	if(newStr == NULL) {
		free(curStr);
		return FALSE;
	}

	GetWindowText(ugl_control->hWnd, (LPTSTR)curStr, len + 1);
	switch(ch) {
		case VK_RETURN: {
			SendMessage(ugl_control->hWnd, CB_SHOWDROPDOWN, FALSE, 0);
			SetWindowText(ugl_control->hWnd, (LPTSTR)curStr);
			SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(len, len));
		} break;
		case VK_DELETE : {
			if(len == start) {
				break;
			}

			modify = TRUE;
			if(start == end) {
				_Combo_Text_CombineNewString(curStr, newStr, start, end+1, 0);
			} else {
				_Combo_Text_CombineNewString(curStr, newStr, start, end, 0);
			}
			len = _tcslen(newStr);
			if(len == 0) {
				SetWindowText(ugl_control->hWnd, (LPTSTR)newStr);
				break;
			}

			SendMessage(ugl_control->hWnd, CB_SHOWDROPDOWN, TRUE, 0);
			_Combo_Text_FindAndSelectItem(ugl_control->hWnd, (LPCSTR)newStr);
			SetWindowText(ugl_control->hWnd, (LPTSTR)newStr);
			SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(start, start));
		} break;
		case VK_BACK : {
			if(end == 0) {
				break;
			}

			modify = TRUE;
			if(start == end) {
				_Combo_Text_CombineNewString(curStr, newStr, start-1, end, 0);
			} else {
				_Combo_Text_CombineNewString(curStr, newStr, start, end, 0);
			}
			len = _tcslen(newStr);
			if(len == 0) {
				SetWindowText(ugl_control->hWnd, (LPTSTR)newStr);
				break;
			}

			SendMessage(ugl_control->hWnd, CB_SHOWDROPDOWN, TRUE, 0);
			_Combo_Text_FindAndSelectItem(ugl_control->hWnd, (LPCSTR)newStr);
			SetWindowText(ugl_control->hWnd, (LPTSTR)newStr);

			if(start == end) {
				SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(start-1, start-1));
			} else {
				SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(start, start));
			}
		} break;
		default : {
			if(_istcntrl(ch)) {
				break;
			}

			modify = TRUE;
			SendMessage(ugl_control->hWnd, CB_SHOWDROPDOWN, TRUE, 0);
			_Combo_Text_CombineNewString(curStr, newStr, start, end, ch);
			SetWindowText(ugl_control->hWnd, (LPTSTR)newStr);
			if(_Combo_Text_FindAndSelectItem(ugl_control->hWnd, (LPCSTR)newStr)) {
				SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(start+1, -1));
			} else {
				SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(start+1, start+1));
			}
		} break;
	}

	free(curStr);
	free(newStr);

	if(modify) {
		ComboCallback(UGL_WIDGET(ugl_control)->callback_target, UGL_EVENT_COMBO_TEXT);
	}

	return modify;
}

void _Combo_Text_CombineNewString(TCHAR *curStr, TCHAR *newStr, int pos1, int pos2, TCHAR ch) {
	int off = 0;
	int len = _tcslen(curStr);

	_tcsncpy(newStr, curStr, pos1);
	off = pos1;
	if(ch != 0) {
		newStr[off++] = ch;
	}
	_tcsncpy(newStr+off, curStr+pos2, len-pos2);
	off += len-pos2;
	newStr[off] = 0;
}

UGL_Boolean _Combo_Text_FindAndSelectItem(HWND hwnd, LPCSTR str) {
	return FALSE;
	/*
	int index;
	
	index = SendMessage(hwnd, CB_FINDSTRINGEXACT, (WPARAM)-1, (LPARAM)str);
	if(index == CB_ERR) {
		index = SendMessage(hwnd, CB_FINDSTRING, (WPARAM)-1, (LPARAM)str);
	}

	if(index != CB_ERR) {
		SendMessage(hwnd, CB_SETCURSEL, (WPARAM)index, (LPARAM)0);
	}

	return (index != CB_ERR);
*/
}

void _Combo_Text_SaveText(PUGL_Control ugl_control) {
	DWORD rs;
	int len;
	if(UGL_COMBO(ugl_control)->is_read_only) {
		return;
	}

	len = GetWindowTextLength(ugl_control->hWnd);

	if(UGL_COMBO(ugl_control)->text != NULL)
		free(UGL_COMBO(ugl_control)->text);

	UGL_COMBO(ugl_control)->text = (TCHAR *)calloc(len + 1, sizeof(TCHAR));
	if(UGL_COMBO(ugl_control)->text == NULL) {
		return;
	}
	GetWindowText(ugl_control->hWnd, (LPTSTR)(UGL_COMBO(ugl_control)->text), len + 1);

	rs = (DWORD)SendMessage(ugl_control->hWnd, CB_GETEDITSEL, 0, 0);
	UGL_COMBO(ugl_control)->selStart = LOWORD(rs);
	UGL_COMBO(ugl_control)->selEnd = HIWORD(rs);
}
void _Combo_Text_RestoreText(PUGL_Control ugl_control) {
	if(UGL_COMBO(ugl_control)->is_read_only) {
		return;
	}

	if(UGL_COMBO(ugl_control)->text != NULL) {
		SetWindowText(ugl_control->hWnd, (LPTSTR)(UGL_COMBO(ugl_control)->text));

		#if !defined(WIN32_PLATFORM_WFSP)
			SendMessage(ugl_control->hWnd, CB_SETEDITSEL, (WPARAM)0, (LPARAM)MAKELPARAM(UGL_COMBO(ugl_control)->selStart, UGL_COMBO(ugl_control)->selEnd));
		#endif
	}
}
