/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"

#include <math.h>
#include "UGL_Win32_Display.h"
#include "UGL_Win32_DC.h"

#include "Font.h"

/*
 * The recommended macro MulDiv is not defined
 * in WinCE so a quick implementation has to be provided
 */ 
#ifndef UNDER_NT
#define MulDiv(num, numerator, denominator) ((num*numerator)/denominator)
#endif

//========================================================================================
PUGL_Font _internal_new_Font(HFONT fontHandle, UGL_Error uglError);
//========================================================================================

void _Font_FillInFont(LPLOGFONT fontData, int height, int style) {
	HDC hdc = GetDC(NULL);

	if (IS_FLAG_SET(style, UGL_FONT_STYLE_BOLD)) fontData->lfWeight = FW_BOLD;
	else fontData->lfWeight = FW_NORMAL;

	if (IS_FLAG_SET(style, UGL_FONT_STYLE_ITALIC)) fontData->lfItalic = TRUE;
	else fontData->lfItalic = FALSE;
	
	if (IS_FLAG_SET(style, UGL_FONT_STYLE_UNDERLINED)) fontData->lfUnderline = TRUE;	
	else fontData->lfUnderline = FALSE;	

	// We set the charset field so that
	// wildcard searching will work properly
	// out of the box
	fontData->lfCharSet = DEFAULT_CHARSET;

#ifdef _WIN32_WCE
	/**
	 * MulDiv() does not seem to exist in Windows CE, regardless of documentation.
	 * Workaround is to accept the possible rounding error.
	 */
	fontData->lfHeight = (height * GetDeviceCaps(hdc, LOGPIXELSY)) / 72;
#else
	fontData->lfHeight = MulDiv(height, GetDeviceCaps(hdc, LOGPIXELSY), 72);
#endif
	ReleaseDC(NULL, hdc);
}


UGL_Int Font_New(UGL_Int deviceHandle, UGL_String name, UGL_Int height, UGL_Int style, UGL_Error uglError) {
	PUGL_Display ugl_display = UGL_DISPLAY(deviceHandle);
	LOGFONT logFont;
	HFONT fontHandle;
	PUGL_Font ugl_font = NULL;

	memset(&logFont, 0, sizeof(LOGFONT));

	if (lstrcmpi(name, _T("fixed")) == 0) {
		// This here because some older PPro applcations
		// depend on "fixed" resolving to something monospaced
 		// and windows will resolve "fixed" to something
		// that is not visible when displayed.
		// see bugs 108764 and 108003.
		logFont.lfPitchAndFamily = FF_MODERN;
	} else {
		_tcsncpy(logFont.lfFaceName, name, LF_FACESIZE);
	}
	_Font_FillInFont(&logFont, height, style);

	fontHandle = CreateFontIndirect(&logFont);
	if (!fontHandle) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to createfont"));
		return 0;
	}

	ugl_font = _internal_new_Font(fontHandle, uglError);

	return (UGL_Int)ugl_font;
}
//========================================================================================

void Font_Dispose(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	DeleteObject(ugl_font->font_handle);
	memset(ugl_font, 0, sizeof(UGL_Font));
	free(ugl_font);
}
//========================================================================================

UGL_Int Font_GetAdvanceWidth(UGL_Int handle, UGL_Char ch, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;
	HDC hdc = GetDC(NULL);

#ifdef _WIN32_WCE
	SIZE size;;
	SelectObject(hdc, ugl_font->font_handle);

	GetTextExtentPoint32W(hdc, &ch, 1, &size);
	ReleaseDC(NULL, hdc);
	return size.cx;
#else
	int tch = ch;
	int width;

	SelectObject(hdc, ugl_font->font_handle);
	GetCharWidth(hdc, tch, tch, &width);
	ReleaseDC(NULL, hdc);
	return width;
#endif
}
//========================================================================================

UGL_Int Font_GetAscent(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	return ugl_font->font_metrics.tmAscent
		   - ugl_font->font_metrics.tmInternalLeading;
}
//========================================================================================

UGL_Int Font_GetDescent(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	return ugl_font->font_metrics.tmDescent;
}
//========================================================================================

UGL_Int Font_GetHeight(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	return ugl_font->font_metrics.tmHeight;
}
//========================================================================================

UGL_Int Font_GetLeading(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	return ugl_font->font_metrics.tmInternalLeading;
}
//========================================================================================

UGL_Int Font_StringAdvance(UGL_Int handle, UGL_String string, UGL_Error uglError){
	HDC hdc;
	if (string == NULL) return 0;

	hdc = GetDC(NULL);
	if (hdc != NULL) {
		int advance=0;

		PUGL_Font ugl_font = (PUGL_Font)handle;
		HGDIOBJ oldFont = SelectObject(hdc, ugl_font->font_handle);
		if (oldFont != NULL) {
			SIZE size;
			int length = lstrlen(string);
			if (!GetTextExtentPoint32(hdc, string, length, &size)) {
				Win32Error_SetError(uglError, GetLastError(), _T("Failed to calculate text extent"));
			} else {
				advance = size.cx;
			}

			SelectObject(hdc, oldFont);
		} else {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Failed to select font"));
		}

		ReleaseDC(NULL, hdc);
		return advance;
	}

	Win32Error_SetError(uglError, GetLastError(), _T("Failed to get screen DC"));
	return 0;
}

//========================================================================================

void Font_StringExtent(UGL_Int handle, UGL_String string, UGL_Int flags, UGL_Int* returnWidth, UGL_Int* returnHeight, UGL_Error uglError) {
	HDC hdc;
	PUGL_Font ugl_font = (PUGL_Font)handle;
	HGDIOBJ oldFont;

	int uFormat = 0;
	RECT rect;

	if (string == NULL || string[0] == '\0') {
		// If the string is null or the length of the string is 0 then 
		// a shortcut can be taken to avoid doing more work than needed.

		// In addition to being quicker, this shortcut is required because
		// DrawText does not properly handle the 0 length string case properly.
		// The return value for DrawText is the height of the string or 0 if an 
		// error occurred, and since the height of an empty string is 0 it is not possible 
		// to tell whether the method failed or completed successfully.  

		// One solution would be to check GetLastError() for ERROR_SUCCESS after calling
		// the method. That does not work either becacuse the error value is not reset 
		// to SUCCESS when a 0 length string is passed in.  Calling SetLastError(ERROR_SUCCESS)
		// before DrawText and checking GetLastError() afterwards will fix the problem, but all
		// that extra work is too much for a case that can be avoided so easily.

		*returnWidth = 0;
		*returnHeight = 0;
		return;
	} 

	hdc = GetDC(NULL);

	SetRect(&rect, 0, 0, 1, 1);

	oldFont = SelectObject(hdc, ugl_font->font_handle);
	
	SET_FLAG(uFormat, DT_LEFT);
	SET_FLAG(uFormat, DT_CALCRECT);

	if (IS_FLAG_SET(flags, UGL_STRING_WRAP)) {
		// String wrapping is the default for the DrawText
		// API so nothing needs to be done.  DT_WORDBREAK is 
		// not correct to be set here since that will break 
		// on spaces as well as line breaks.
	} else {
		SET_FLAG(uFormat, DT_SINGLELINE);
	}

	if (IS_FLAG_SET(flags, UGL_STRING_EXPAND_TAB)) SET_FLAG(uFormat, DT_EXPANDTABS);
	if (!IS_FLAG_SET(flags, UGL_STRING_MNEMONIC)) SET_FLAG(uFormat, DT_NOPREFIX);

	if (!DrawText(hdc, string, -1, &rect, uFormat)) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to calculcate string extent"));
	}
	SelectObject(hdc, oldFont);
	ReleaseDC(NULL, hdc);

	*returnWidth = rect.right;
	*returnHeight = rect.bottom;

}
//========================================================================================



//========================================================================================

UGL_Int Font_GetAverageCharWidth(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	return ugl_font->font_metrics.tmAveCharWidth;
}

//========================================================================================

UGL_Int Font_GetStyle(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;
	LPTEXTMETRIC tm = &ugl_font->font_metrics;
	int ugl_style = 0;

	if (tm->tmItalic) ugl_style |= UGL_FONT_STYLE_ITALIC;
	if (tm->tmWeight == FW_BOLD) ugl_style |= UGL_FONT_STYLE_BOLD;

	return ugl_style;
}
//========================================================================================

UGL_String Font_GetName(UGL_Int handle, UGL_Boolean* shouldFreeResult, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;
	LPTSTR font_name=NULL;
	HDC hdc = GetDC(NULL);
	if (hdc != NULL) {
		HGDIOBJ selectedFont;
		int length = 0;
		
		selectedFont = SelectObject(hdc, ugl_font->font_handle);
		
		length = GetTextFace(hdc, 0, NULL);
		if (length != 0) {
			font_name = (LPTSTR)calloc(sizeof(TCHAR), length);
			
			if (font_name != NULL) {
				if (GetTextFace(hdc, length, font_name) == 0) {
					Win32Error_SetError(uglError, GetLastError(), _T("Unable to copy font face name into buffer"));
					free(font_name);
					font_name = NULL;
				} else {
					*shouldFreeResult = TRUE;
				}
			} else {
				Win32Error_SetError(uglError, UGL_ERROR_OUT_OF_MEMORY, _T("Not enough memory to store font face name"));
			}
		} else {
			Win32Error_SetError(uglError, GetLastError(), _T("Unable to get font face length"));
		}

		SelectObject(hdc, selectedFont);
		ReleaseDC(NULL, hdc);
	} else {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get screen dc"));
	}

	return font_name;
}
//========================================================================================

UGL_Int Font_GetSizeInPoints(UGL_Int handle, UGL_Error uglError) {
	PUGL_Font ugl_font = (PUGL_Font)handle;

	HDC hDC = GetDC(NULL);
	if (hDC != NULL) {
		int size=MulDiv(ugl_font->font_metrics.tmHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY));
		ReleaseDC(NULL, hDC);
		return size;
	} else {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get screen dc"));
		return 0;
	}
}
//========================================================================================

PUGL_Font _internal_new_Font(HFONT fontHandle, UGL_Error uglError) {
	PUGL_Font ugl_font = NULL;
	HDC hdc = GetDC(NULL);
	HGDIOBJ selectedFont;
	

	ugl_font = (PUGL_Font)calloc(sizeof(UGL_Font), 1);
	if (ugl_font == NULL) {
		Win32Error_SetError(uglError, UGL_ERROR_OUT_OF_MEMORY, _T("Not enough memory to allocate font"));
		return 0;
	}
	ugl_font->font_handle = fontHandle;

	selectedFont = SelectObject(hdc, fontHandle);
	// caches text metric information here while we have a valid dc
	GetTextMetrics(hdc,&(ugl_font->font_metrics));
	
	SelectObject(hdc, selectedFont);
	ReleaseDC(NULL, hdc);

	return ugl_font;
}
//========================================================================================
