/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"
#include "Graphics.h"
#include "UGL_Win32_DC.h"

#include "Color.h"
#include "Image.h"
#ifndef _WIN32_WCE
#include "ImageLoader_GDI+.h"
#include "ImageLoader_DirectX.h"
#else
// #include "ImageLoader_External.h"
#include "ImageLoader_Imaging.h"
#endif

#ifdef _WIN32_WCE
#include <winuser.h>
#endif

#include <math.h>

static HANDLE _createBMI(int width, int height, int depth, int paletteSize);
static UINT _Image_GetScanlineLength(UINT width, UINT depth);

//========================================================================================

UINT _Image_GetScanlineLength(UINT width, UINT depth) {
	UINT bitsPerLine = width * depth;
	UINT bitAlignment = sizeof(LONG) * 8;

	UINT bitStride = bitAlignment * (bitsPerLine / bitAlignment);	// The image buffer is always padded to LONG boundaries
	if ((bitsPerLine % bitAlignment) != 0) bitStride += bitAlignment; // Add a bit more for the leftover values
	return bitStride / 8;
}

//========================================================================================

void Image_Dispose(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image image = UGL_IMAGE(handle);
	HANDLE imageHandle = image->image_handle;

	DeleteCriticalSection(&(image->dcSection));

	free(image); 

	DeleteObject(imageHandle);
	
}
//========================================================================================

void _Image_GetSize(PUGL_Image ugl_image, UGL_Int* returnWidth, UGL_Int* returnHeight, UGL_Error uglError) {
	BITMAP bitmap;

	GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);
	*returnWidth = bitmap.bmWidth;
	*returnHeight = bitmap.bmHeight;
}
//========================================================================================

void Image_GetBounds(UGL_Int handle, UGL_Int* returnX, UGL_Int* returnY, UGL_Int* returnWidth, UGL_Int* returnHeight, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	BITMAP bitmap;

	GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);
	*returnX = 0;
	*returnY = 0;
	*returnWidth = bitmap.bmWidth;
	*returnHeight = bitmap.bmHeight;
}
//========================================================================================
UGL_Int Image_GetDepth(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	BITMAP bitmap;

	GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);
	return bitmap.bmBitsPixel;
}
//========================================================================================

void Image_GetRGB(UGL_Int handle, UGL_Int startX, UGL_Int startY, UGL_Int width, UGL_Int height, UGL_Int offset, UGL_Int scanSize, UGL_IntArray rgbs, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);

	BITMAP bitmap;
	LPBYTE pvBits;
	int xPos, yPos;

	int size = GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);

	pvBits = (LPBYTE)bitmap.bmBits;
	for (yPos = startY; yPos < (startY + height); yPos++) {
		for (xPos = startX; xPos < (startX + width); xPos++) {
	   		int rgbIndex = offset + (xPos - startX) + ((yPos - startY) * scanSize);
			int bitsIndex = (xPos*3) + (yPos * bitmap.bmWidthBytes);

			rgbs[rgbIndex] = pvBits[bitsIndex] & 0x0000FF;
			rgbs[rgbIndex] |= (pvBits[bitsIndex+1] << 8) & 0x00FF00;
			rgbs[rgbIndex] |= (pvBits[bitsIndex+2] << 16);
			rgbs[rgbIndex] |= 0xFF000000;
		}
	}

	/*
	HDC imageHDC;
	HGDIOBJ oldBitmap;
	int xPos, yPos;

	if (ugl_image->selected_DC == NULL) {
		HDC hdc = GetDC(NULL);
		imageHDC = CreateCompatibleDC(hdc);
		ReleaseDC(NULL, hdc);
		oldBitmap = SelectObject(imageHDC, ugl_image->image_handle);
	} else {
		imageHDC = ugl_image->selected_DC->hdc;
	}

	for (yPos = 0; yPos < height; yPos++) {
		for (xPos = 0; xPos < width; xPos++) {
			int x = startX + xPos;
			int y = startY + yPos;	    	        
	   		int index = offset + yPos * scanSize + xPos;
			COLORREF pixelRGB = GetPixel(imageHDC, x, y);

			rgbs[index] = RGB_SWAP(pixelRGB);
		}
	}

	if (ugl_image->selected_DC == NULL) {
		SelectObject(imageHDC, oldBitmap);
		DeleteDC(imageHDC);
	} 
	*/
}
//========================================================================================

UGL_Int Image_GetSize(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	BITMAP bitmap;

	GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);

	return _Image_GetScanlineLength(bitmap.bmWidth, bitmap.bmBitsPixel) * bitmap.bmHeight;
}


PUGL_Image _Image_New(HANDLE handle, BOOL is_direct, UCHAR transparent_pixel, INT32 transparent_color, UGL_Error error) {
	PUGL_Image image = (PUGL_Image)calloc(sizeof(UGL_Image),1);
	if (image == NULL) {
		DeleteObject(handle);
		UGL_SetError(error, UGL_ERROR_OUT_OF_MEMORY);
	} else {
		image->image_handle = handle;
		image->is_direct = is_direct;
		image->transparent_pixel = transparent_pixel;
		image->transparent_color = transparent_color;
		InitializeCriticalSection(&(image->dcSection));
	}

	return image;
}

//========================================================================================

UGL_Int Image_NewFromFileData(UGL_Int deviceHandle, UGL_ByteArray data, UGL_Int length, UGL_Error uglError) {
	BOOL direct = TRUE;
	BOOL hasAlpha = FALSE;
	int transparentPixel = -1;
	int transparentColor = -1;
	
	HBITMAP imageHandle;

#ifndef _WIN32_WCE
	imageHandle = GDIpImageLoader_LoadFromFileData(data, length, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
	if ((UGL_GetError(uglError) != 0) && (UGL_GetError(uglError) != UGL_ERROR_INVALID_IMAGE)) {
		BOOL direct = TRUE;
		BOOL hasAlpha = FALSE;
		int transparentPixel = -1;
		int transparentColor = -1;
		
		UGL_ClearError(uglError);
		imageHandle = DXImageLoader_LoadFromFileData(data, length, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
		if (UGL_GetError(uglError) != 0) {
			return NULL_HANDLE;
		}
	} 
#else
	imageHandle = ImagingImageLoader_LoadFromFileData(data, length, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
	if (imageHandle == NULL_HANDLE) return NULL_HANDLE;

//	imageHandle = ExternalImageLoader_LoadFromFileData(data, length, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
//	if (imageHandle == NULL_HANDLE) return NULL_HANDLE;
#endif

	return (UGL_Int)_Image_New(imageHandle, direct, transparentPixel, transparentColor, uglError);
}
//========================================================================================

/**
 * Creates a new, empty image with the given dimensions.
 * <p>
 * Image will be filled with the provided fill color which
 * is in 0xAARRGGBB format.
 * </p>
 *
 * @ingroup Image
 *
 * @param deviceHandle the device/display handle
 * @param width width of the new image
 * @param height height of the new image
 * @param fillColor value with which to fill the image in 0xAARRGGBB format.
 * @param error the error status of the call
 *
 * @return handle to the new image
 */
UGL_Int Image_NewFromSize(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_Int fillColor, UGL_Error error) {
	PUGL_Image ugl_image = NULL;
	BYTE* pBits = NULL;
	HANDLE hDib;
	BITMAPINFO bmi;
	BITMAPINFO *pbmi;
	BYTE* lastPixel;
	UINT scanlineLength;

	/*
	 * Assumes 16-bit format data will be in 5-5-5, and converts
	 * green component proportionally (value / 2 ^ 5).
	 */


	/*
	* Note on WinCE.  CreateDIBSection requires the biCompression
	* field of the BITMAPINFOHEADER to be set to BI_BITFIELDS for
	* 16 and 32 bit direct images (see MSDN for CreateDIBSection).
	* In this case, the color mask can be set to any value.  For
	* consistency, it is set to the same mask used by non WinCE
	* platforms in BI_RGB mode.
	*/

	/*
	 * Windows supports 16-bit mask of 5-5-5 (0x7C00, 0x3E0, 0x1F),
	 * 16-bit 5-6-5 (0xF800, 0x07E0, 0x001F) (only using BI_BITFIELDS), 
	 * 24-bit mask of (0xFF0000, 0xFF00, 0xFF) and 32-bit mask
	 * (0x00FF0000, 0x0000FF00, 0x000000FF) as documented in 
	 * MSDN BITMAPINFOHEADER.  Make sure the image is 
	 * Windows-supported.
	 */

	memset(&bmi, 0, sizeof(BITMAPINFOHEADER));
	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = width;
	bmi.bmiHeader.biHeight = -height;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = 24;
	bmi.bmiHeader.biClrUsed = 0;
	bmi.bmiHeader.biCompression = BI_RGB;
	pbmi = &bmi;

	hDib = CreateDIBSection(NULL, pbmi, DIB_RGB_COLORS, (void **)&pBits, NULL, 0x0);
	if (hDib == 0) {
		Win32Error_SetError(error, GetLastError(), _T("Unable to create device-independent image handle"));
		if (pbmi != (&bmi)) free(pbmi);
		return NULL_HANDLE;
	}

	if (pbmi != (&bmi)) free(pbmi);

	scanlineLength = _Image_GetScanlineLength(width, 24);
	lastPixel = pBits + (scanlineLength * height);
	while (pBits < lastPixel) {
		int x;
		for (x=0;x<width;x++) {
			pBits[x*3] = (fillColor >> 0) & 0xFF;
			pBits[x*3+1] = (fillColor >> 8) & 0xFF;
			pBits[x*3+2] = (fillColor >> 16) & 0xFF;
		}
		pBits = (BYTE*)pBits + scanlineLength;
	}

	return (UGL_Int)_Image_New(hDib, TRUE, -1, -1, error);
}
//========================================================================================

UGL_Int Image_NewFromIntData(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_IntArray imageData, UGL_Int depth, UGL_Int transparentColor, UGL_Int globalAlpha, UGL_ByteArray alphaData, UGL_Error uglError) {
	PUGL_Image ugl_image = NULL;
	int pBits[1] = {0};
	HANDLE hDib;
	BITMAPINFO bmi;
	BITMAPINFO *pbmi;

	/*
	 * Assumes 16-bit format data will be in 5-5-5, and converts
	 * green component proportionally (value / 2 ^ 5).
	 */


	/*
	* Note on WinCE.  CreateDIBSection requires the biCompression
	* field of the BITMAPINFOHEADER to be set to BI_BITFIELDS for
	* 16 and 32 bit direct images (see MSDN for CreateDIBSection).
	* In this case, the color mask can be set to any value.  For
	* consistency, it is set to the same mask used by non WinCE
	* platforms in BI_RGB mode.
	*/

	/*
	 * Windows supports 16-bit mask of 5-5-5 (0x7C00, 0x3E0, 0x1F),
	 * 16-bit 5-6-5 (0xF800, 0x07E0, 0x001F) (only using BI_BITFIELDS), 
	 * 24-bit mask of (0xFF0000, 0xFF00, 0xFF) and 32-bit mask
	 * (0x00FF0000, 0x0000FF00, 0x000000FF) as documented in 
	 * MSDN BITMAPINFOHEADER.  Make sure the image is 
	 * Windows-supported.
	 */

	memset(&bmi, 0, sizeof(BITMAPINFOHEADER));
	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = width;
	bmi.bmiHeader.biHeight = -height;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = depth;
	if ((alphaData != NULL) || (globalAlpha != -1)) {
		// Windows only supports aplha blending for 32-bpp
		bmi.bmiHeader.biBitCount = 32;
	}

	bmi.bmiHeader.biClrUsed = 0;

#if (_WIN32_WCE == 300)
	if ((depth == 16 || depth == 32)) {
		// investigate use of BI_BITFIELDS in WIN32 for 5-6-5 16-bit color,
		// using masks in above comment.

		bmi.bmiHeader.biCompression = BI_BITFIELDS;

		pbmi = (BITMAPINFO *)calloc(sizeof(BITMAPINFOHEADER) + 12, 1);
		MoveMemory((void *)pbmi, (void *)&bmi.bmiHeader, sizeof(BITMAPINFOHEADER));
		{
			LPBYTE pbytebmi = (LPBYTE)pbmi;
			int offset = sizeof(BITMAPINFOHEADER);
			int redMask = (depth == 32 ? 0xFF0000 : 0xF800);
			int greenMask = (depth == 32 ? 0xFF00 : 0x07E0);
			int blueMask = (depth == 32 ? 0xFF : 0x001F);
		
			pbytebmi[offset] = (BYTE)((redMask & 0xFF) >> 0);
			pbytebmi[offset + 1] = (BYTE)((redMask & 0xFF00) >> 8);
			pbytebmi[offset + 2] = (BYTE)((redMask & 0xFF0000) >> 16);
			pbytebmi[offset + 3] = (BYTE)((redMask & 0xFF000000) >> 24);
			pbytebmi[offset + 4] = (BYTE)((greenMask & 0xFF) >> 0);
			pbytebmi[offset + 5] = (BYTE)((greenMask & 0xFF00) >> 8);
			pbytebmi[offset + 6] = (BYTE)((greenMask & 0xFF0000) >> 16);
			pbytebmi[offset + 7] = (BYTE)((greenMask & 0xFF000000) >> 24);
			pbytebmi[offset + 8] = (BYTE)((blueMask & 0xFF) >> 0);
			pbytebmi[offset + 9] = (BYTE)((blueMask & 0xFF00) >> 8);
			pbytebmi[offset + 10] = (BYTE)((blueMask & 0xFF0000) >> 16);
			pbytebmi[offset + 11] = (BYTE)((blueMask & 0xFF000000) >> 24);
		}
	}	
#else
	bmi.bmiHeader.biCompression = BI_RGB;
	pbmi = &bmi;
#endif

	hDib = CreateDIBSection(NULL, (CONST BITMAPINFO *)pbmi, DIB_RGB_COLORS, (void **)&pBits, NULL, 0x0);
	if (hDib == 0) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to create device-independent image handle"));
		if (pbmi != (&bmi)) free(pbmi);
		return NULL_HANDLE;
	}


//	if (depth == 32) {
//		MoveMemory((void *)pBits[0], imageData, (depth * width * height) / 8);
//	} else {
	{
		int data_int_length = width * height;
		int buffer_index, data_index;
		LPBYTE image_buffer = (LPBYTE)pBits[0];

		for (data_index = 0, buffer_index = 0; data_index < data_int_length; data_index++) {
			int pixel = imageData[data_index];

			int red;
			int green;
			int blue;
			if (depth == 16) {
				// 16-bpp arrives as 5-6-5 (0xF800, 0x07E0, 0x001F)
				red = (pixel & 0xF800) >> 11;
				green = (pixel & 0x07E0) >> 5;
				blue = (pixel & 0x001F);
				if (alphaData != NULL) {
					// with alpha, 16-bpp 5-6-5 must be converted to 32-bpp 8-8-8
					red = ((red * 0x0FF) / 0x01F) << 16;
					green = ((green * 0x0FF) / 0x03F) << 8;
					blue = (blue * 0x0FF) / 0x01F;
				} else {
					// No alpha, we only have to find max green in 16 bit 5-5-5
					green = (green * 0x01F) / 0x03F;
				}
			} else {
				// 24-bpp and 32-bpp arrive as 8-8-8
				red = (pixel & 0x00FF0000);  
				green = (pixel & 0x0000FF00);  
				blue = (pixel & 0x000000FF);  
			}

			if (bmi.bmiHeader.biBitCount == 32) {
				// 32-bpp (with or without alpha), 16-bpp with alpha or, 24-bpp with alpha
				unsigned long *target_pixel = (unsigned long *)&image_buffer[buffer_index];
				*target_pixel = 0;
				if (alphaData != NULL) {
					*target_pixel = (alphaData[data_index] << 24);
				} else 
				if (globalAlpha != -1) {
					*target_pixel = (globalAlpha << 24) & 0xFF000000;
				} else 
				if (depth == 32) {
					*target_pixel = (pixel & 0xFF000000);
				}
				*target_pixel |= red;  // red
				*target_pixel |= green;  //green
				*target_pixel |= blue;  //blue
//				printf("target_pixel= %d", *target_pixel);
				buffer_index += sizeof(unsigned long);
			} else if (depth == 24) {			
				image_buffer[buffer_index++] = blue;  //blue
				image_buffer[buffer_index++] = green >> 8;  //green
				image_buffer[buffer_index++] = red >> 16;  //red
			} else if (depth == 16) {
				unsigned short *target_pixel = (unsigned short *)&image_buffer[buffer_index];
				// 16-bit comes in as 5-6-5 (0xF800, 0x07E0, 0x001F)
//				int red = (pixel & 0xF800) >> 11;
//				int green = (pixel & 0x07E0) >> 5;
//				int blue = (pixel & 0x001F);
//				// we only have to find max green in 16 bit 5-5-5
//				green = (green * 0x01F) / 0x03F;

//				unsigned short *target_pixel = (unsigned short *)&image_buffer[buffer_index];
				*target_pixel = ((red << 10)) | ((green  << 5)) | (blue);
				buffer_index += sizeof(unsigned short);
			}
			if (((data_index + 1) % width) == 0){
				// at the end of one row of pixel data. Each scan
				// line must be padded with zeroes to end on a LONG boundary
				int remainder = ((buffer_index) % sizeof(long));
				if (remainder != 0) {
					buffer_index += sizeof(long) - remainder;
				}
			}

		}
	} 


	if (pbmi != (&bmi)) free(pbmi);
	return (UGL_Int)_Image_New(hDib, TRUE, 0, transparentColor, uglError);
}
//========================================================================================

UGL_Int Image_NewFromSystemResource(UGL_Int deviceHandle, UGL_Int identifier, UGL_Error uglError) {
	UGL_Int handle;
	PUGL_Image ugl_image;
	LPCWSTR lpName;
	HICON hIcon;
	HDC hDC;
	HBITMAP hBitmap;
	HBITMAP hOldBitmap;

#ifndef _WIN32_WCE
	switch (identifier)
	{
	default:
	case UGL_IMAGE_ERROR:
		lpName = IDI_ERROR;
		break;

	case UGL_IMAGE_INFORMATION:
		lpName = IDI_INFORMATION;
		break;

	case UGL_IMAGE_QUESTION:
		lpName = IDI_QUESTION;
		break;

	case UGL_IMAGE_WARNING:
		lpName = IDI_WARNING;
		break;
	}

	hIcon = LoadIcon(NULL, lpName);
#else
    HMODULE hGWES;

	switch (identifier)
	{
	default:
	case UGL_IMAGE_ERROR:
		lpName = MAKEINTRESOURCE(MB_ICONERROR);
		break;

	case UGL_IMAGE_INFORMATION:
		lpName = MAKEINTRESOURCE(MB_ICONINFORMATION);
		break;

	case UGL_IMAGE_QUESTION:
		lpName = MAKEINTRESOURCE(MB_ICONQUESTION);
		break;

	case UGL_IMAGE_WARNING:
		lpName = MAKEINTRESOURCE(MB_ICONWARNING);
		break;
	}

	hGWES = LoadLibraryEx(L"gwes.exe", NULL, LOAD_LIBRARY_AS_DATAFILE);
	hIcon = LoadIcon(hGWES, lpName);
#endif

	if (hIcon == NULL)
	{
		UGL_SetError(uglError, UGL_ERROR_INVALID_IMAGE);
		return NULL_HANDLE;
	}
	else
	{
		// Though it defines an ICON constant as well as BITMAP, and tracks the type,
		// the Image class only implements BITMAP support, so we must convert.
		handle = Image_NewFromSize(deviceHandle, GetSystemMetrics(SM_CXICON), GetSystemMetrics(SM_CYICON), 0x00FFFFFF, uglError);
		ugl_image = UGL_IMAGE(handle);
		hDC = CreateCompatibleDC(NULL);
		hOldBitmap = (HBITMAP)SelectObject(hDC, ugl_image->image_handle);
		DrawIcon(hDC, 0, 0, hIcon);
		SelectObject(hDC, hOldBitmap);
		DeleteDC(hDC);
		return handle;
	}
}
//========================================================================================

UGL_Int Image_NewGraphics(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	PUGL_DC ugl_dc = NULL;
	HBITMAP dummyBitmap;

	HDC imageHDC = CreateCompatibleDC(NULL);
	RECT minClip;

	dummyBitmap = CreateCompatibleBitmap(imageHDC, 0, 0);
	ugl_image->null_bitmap = (HBITMAP)SelectObject(imageHDC, dummyBitmap);
	DeleteObject(dummyBitmap);

	{
		int x,y,width,height;
		Image_GetBounds(handle, &x, &y, &width, &height, uglError);

		minClip.top = y;
		minClip.left = x;
		minClip.right = x+width;
		minClip.bottom = y+height;
	}

	ugl_dc = _internal_new_DC(imageHDC, ugl_image, &minClip, uglError);

	return (UGL_Int)ugl_dc;
}

//========================================================================================

HANDLE _createBMI(int width, int height, int depth, int paletteSize) {
	BITMAPINFO *pbmi;

	pbmi = (BITMAPINFO *)calloc(sizeof(BITMAPINFOHEADER) + paletteSize * sizeof(RGBQUAD), 1);
	pbmi->bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	pbmi->bmiHeader.biWidth = width;
	pbmi->bmiHeader.biHeight = -height;
	pbmi->bmiHeader.biPlanes = 1;
	pbmi->bmiHeader.biBitCount = depth;
	pbmi->bmiHeader.biCompression = BI_RGB;
	pbmi->bmiHeader.biClrUsed = paletteSize;

	return pbmi;
}

//========================================================================================


//========================================================================================

void Image_CopyArea(UGL_Int handle, UGL_Int destImageHandle, UGL_Int x, UGL_Int y, UGL_Int width, UGL_Int height, UGL_Error uglError)  {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	PUGL_Image ugl_dest_image = UGL_IMAGE(destImageHandle);
	PUGL_DC ugl_dest_dc = (PUGL_DC)Image_NewGraphics(destImageHandle, uglError);

	Graphics_DrawImageRegion((UGL_Int)ugl_dest_dc, (UGL_Int)ugl_image, x, y, width, height, 0, 0, width, height, uglError);
	Graphics_Dispose((UGL_Int)ugl_dest_dc, uglError);
}
//========================================================================================

UGL_Int Image_Flip(UGL_Int image_handle, UGL_Boolean hFlip, UGL_Boolean vFlip, UGL_Int x, UGL_Int y, UGL_Int width, UGL_Int height, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(image_handle);
	PUGL_Image ugl_subimage = NULL;

	HDC srcHdc, subHdc;
	HBITMAP oldSubBitmap, oldSrcBitmap;

	int destX, destY, destWidth, destHeight;

	// create a memory DC
	srcHdc = CreateCompatibleDC(NULL);
	// select the image into the DC
	oldSrcBitmap = SelectObject(srcHdc, ugl_image->image_handle);

	// create the subimage the size of the area to flip
	ugl_subimage = UGL_IMAGE(Image_NewFromSize(NULL_HANDLE, width, height, 0xFF000000, uglError));
	// save all of the transparency and alpha information from the original image
	ugl_subimage->transparent_color = ugl_image->transparent_color;
	ugl_subimage->transparent_pixel = ugl_image->transparent_pixel;

	// create another memory DC
	subHdc = CreateCompatibleDC(NULL);
	// select the new subimage bitmap into it
	oldSubBitmap = SelectObject(subHdc, ugl_subimage->image_handle);

	destWidth = width;
	destX = 0;
	if (hFlip) {
		// indicate image rotates around x-axis
		destWidth = -width;
		// rotate the x coordinate also
		destX = width - 1;
	}

	destHeight = height;
	destY = 0;
	if (vFlip) {
		// indicate image rotates around y-axis
		destHeight = -height;
		// rotate the y coordinate also
		destY = height -1 ;
	}

	// copy described portion of the original image into the new subimage 
	// rotating as defined by the destination.
	StretchBlt(subHdc, destX, destY, destWidth, destHeight, srcHdc, x, y, width, height, SRCCOPY);

	// free resources
	SelectObject(srcHdc, oldSrcBitmap);
	DeleteDC(srcHdc);
	SelectObject(subHdc, oldSubBitmap);
	DeleteDC(subHdc);

	return (UGL_Int)ugl_subimage;
}
//========================================================================================

void Image_GetDirectPaletteMasks(UGL_Int handle, UGL_Int* returnRedMask, UGL_Int* returnGreenMask, UGL_Int* returnBlueMask, UGL_Int* returnAlphaMask, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	int depth = Image_GetDepth(handle, uglError);

	/*
	 * Windows supports 16-bit mask of 5-5-5 (0x7C00, 0x3E0, 0x1F),
	 * 16-bit 5-6-5 (0xF800, 0x07E0, 0x001F) (only using BI_BITFIELDS), 
	 * 24-bit mask of (0xFF0000, 0xFF00, 0xFF) and 32-bit mask
	 * (0x00FF0000, 0x0000FF00, 0x000000FF) as documented in 
	 * MSDN BITMAPINFOHEADER.  Make sure the image is 
	 * Windows-supported.
	 */

	switch (depth) {
		case 32: {
			*returnRedMask = 0xFF0000;
			*returnGreenMask = 0xFF00;
			*returnBlueMask = 0xFF;
			*returnAlphaMask = 0xFF000000;
				 } break;
		case 24: {
			*returnRedMask = 0xFF0000;
			*returnGreenMask = 0xFF00;
			*returnBlueMask = 0xFF;
			*returnAlphaMask = 0;
				 } break;
		case 16: {
			// 5-5-5 for now, must make 5-6-5 eventually
			*returnRedMask = 0x7C00;
			*returnGreenMask = 0x3E0;
			*returnBlueMask = 0x1F;
			*returnAlphaMask = 0;

			// 5-6-5
			//*returnRedMask = 0xF800;
			//*returnGreenMask = 0x07E0;
			//*returnBlueMask = 0x001F;
				 } break;

		default: {
			*returnRedMask = 0;
			*returnGreenMask = 0;
			*returnBlueMask = 0;
			*returnAlphaMask = 0;


				 } break;
	}

}
//========================================================================================

void Image_GetIndexedPalette(UGL_Int handle, UGL_ByteArray red, UGL_ByteArray green, UGL_ByteArray blue, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	HDC imageHDC;
	HGDIOBJ oldBitmap;
	int depth = 0;
	int paletteSize, index = 0;
	LPRGBQUAD palette = NULL;
	DIBSECTION ds;


	if (Image_IsDirect(handle, uglError))	{
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Indexed palette not available for direct color image"));
		return;
	}

	// determine size of palette so we don't try to get more colors than we are using
	GetObject(ugl_image->image_handle, sizeof(DIBSECTION), &ds);
	paletteSize = ds.dsBmih.biClrUsed;
	if (paletteSize == 0) {
		// if biClrUsed equals 0, maximum palette for bit count is used
		paletteSize = 1 << ds.dsBm.bmBitsPixel;
	}

	palette = (LPRGBQUAD)calloc(sizeof(RGBQUAD), paletteSize);
	
	EnterCriticalSection(&(ugl_image->dcSection));
	if (ugl_image->selected_DC == NULL) {
		imageHDC = CreateCompatibleDC(NULL);
		oldBitmap = SelectObject(imageHDC, ugl_image->image_handle);
	} else {
		imageHDC = ugl_image->selected_DC;
	}

	if (!GetDIBColorTable(imageHDC, 0, paletteSize, palette)) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to get DIB color table"));
	} else {
		for (index = 0; index < paletteSize; index++) {
			LPRGBQUAD palette_entry = (LPRGBQUAD)&palette[index];

			red[index] = palette_entry->rgbRed;
			green[index] = palette_entry->rgbGreen;
			blue[index] = palette_entry->rgbBlue;
		}
	}

	free(palette);
	if (ugl_image->selected_DC == NULL) {
		SelectObject(imageHDC, oldBitmap);
		DeleteDC(imageHDC);
	}

	LeaveCriticalSection(&(ugl_image->dcSection));
}
//========================================================================================

UGL_Int Image_NewFromHandle(UGL_Int deviceHandle, UGL_Int handle, UGL_Error uglError){
 
	PUGL_Image ugl_image = NULL;
	PUGL_Image originalImage = UGL_IMAGE(handle);
	HBITMAP hBitmap = originalImage->image_handle;
	HBITMAP hBitmapNew, hbmSrcOld;
	DIBSECTION ds;
	HDC hdcSrc;

	int paletteSize, index;
	BITMAPINFO *pBmi;
	LPBYTE *pBits;

	// there exists a DC that has this image selected, temporarily disable it
	EnterCriticalSection(&(originalImage->dcSection));
	if (originalImage->selected_DC != NULL) {
		SelectObject(originalImage->selected_DC, originalImage->null_bitmap);
	}
	// get information about the bitmap
	hdcSrc = CreateCompatibleDC(NULL);
	hbmSrcOld = SelectObject(hdcSrc, hBitmap);
    GetObject(hBitmap, sizeof(DIBSECTION), &ds);

	// create a palette only for images which are not direct
	paletteSize = ds.dsBmih.biClrUsed;
	if (!originalImage->is_direct && paletteSize == 0) {
		// if biClrUsed equals 0, maximum palette for bit count is used
		paletteSize = 1 << ds.dsBm.bmBitsPixel;
	}
	pBmi = _createBMI(ds.dsBm.bmWidth, ds.dsBm.bmHeight, ds.dsBm.bmBitsPixel, paletteSize);

	if (!originalImage->is_direct) {
		// for bitmaps which use indexed palettes we must copy the color
		// table from the original bitmap into the new one. In win32 and
		// newer versions of wince you can get the color table for DIBs with
		// GetDIBColorTable.
		RGBQUAD *pBmiColors = (LPRGBQUAD)(&pBmi->bmiColors);
		LPBYTE pBitsOrig = ds.dsBm.bmBits;
		BYTE pBitsOld = pBitsOrig[0];
		for (index = 0; index < paletteSize; index++) {
			// iterate through the entire color table. Set byte 0 of the 
			// image data to point to the index in the palette, and then 
			// get the color that is pointed to by that index.
			COLORREF rgbColor;
			pBitsOrig[0] = (BYTE) index & 0xff;
			rgbColor = GetPixel(hdcSrc, 0, 0);
			// store the color into the palette for the new bitmap
			pBmiColors[index].rgbBlue = GetBValue(rgbColor);
			pBmiColors[index].rgbGreen = GetGValue(rgbColor);
			pBmiColors[index].rgbRed = GetRValue(rgbColor);
			pBmiColors[index].rgbReserved = 0;
		}
		// restore byte 0 back to its original value
		pBitsOrig[0] = pBitsOld;
	}

	// create the dib to match our original image
	hBitmapNew = CreateDIBSection(NULL, pBmi, DIB_RGB_COLORS, (void **)&pBits, NULL, 0x0);
	if (hBitmapNew == 0) {
		free(pBmi);
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to create device-independent image handle"));
		if (originalImage->selected_DC != NULL) {
			SelectObject(originalImage->selected_DC, originalImage->image_handle);
		}
		LeaveCriticalSection(&(originalImage->dcSection));
		return NULL_HANDLE;
	}

	// copy the image data bits to the new bitmap
	MoveMemory(pBits, ds.dsBm.bmBits, ds.dsBm.bmWidthBytes * ds.dsBm.bmHeight);

	// release the memory
	free(pBmi);
    SelectObject(hdcSrc, hbmSrcOld);
    DeleteDC(hdcSrc);
	if (originalImage->selected_DC != NULL) {
		SelectObject(originalImage->selected_DC, originalImage->image_handle);
	}
	LeaveCriticalSection(&(originalImage->dcSection));

	return (UGL_Int)_Image_New(hBitmapNew, originalImage->is_direct, originalImage->transparent_pixel, originalImage->transparent_color, uglError);
}
//========================================================================================

UGL_Int Image_NewFromFile(UGL_Int deviceHandle, UGL_String fileName, UGL_Error uglError) {
	BOOL direct = TRUE;
	BOOL hasAlpha = FALSE;
	int transparentPixel = -1;
	int transparentColor = -1;
	HBITMAP imageHandle;

#ifndef _WIN32_WCE
	imageHandle = GDIpImageLoader_LoadFromFile(fileName, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
	if ((UGL_GetError(uglError) != 0) && (UGL_GetError(uglError) != UGL_ERROR_INVALID_IMAGE)) {
		BOOL direct = TRUE;
		BOOL hasAlpha = FALSE;
		int transparentPixel = -1;
		int transparentColor = -1;
		
		UGL_ClearError(uglError);
		imageHandle = DXImageLoader_LoadFromFile(fileName, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
		if (UGL_GetError(uglError) != 0) {
			return NULL_HANDLE;
		}
	} 
#else
	imageHandle = ImagingImageLoader_LoadFromFile(fileName, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
	if (imageHandle == NULL_HANDLE) return NULL_HANDLE;

//	imageHandle = ExternalImageLoader_LoadFromFile(fileName, &direct, &hasAlpha, &transparentPixel, &transparentColor, uglError);
//	if (imageHandle == NULL_HANDLE) return NULL_HANDLE;
#endif

	return (UGL_Int)_Image_New(imageHandle, direct, transparentPixel, transparentColor, uglError);
}
//========================================================================================

void Image_SetRGB(UGL_Int handle, UGL_Int x, UGL_Int y, UGL_Int argb, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);

	BITMAP bitmap;
	LPBYTE pvBits;
	int index;

	int size = GetObject(ugl_image->image_handle, sizeof(BITMAP), &bitmap);

	pvBits = (LPBYTE)bitmap.bmBits;
	index = (x*3) + y * bitmap.bmWidthBytes;
	*(pvBits + index) = (argb >> 0) & 0xFF;
	*(pvBits + index+1) = (argb >> 8) & 0xFF;
	*(pvBits + index+2) = (argb >> 16) & 0xFF;

	/*
	HDC imageHDC;
	HGDIOBJ oldBitmap;
	COLORREF color;
	int red, green, blue;

	if (ugl_image->selected_DC == NULL) {
		HDC hdc = GetDC(NULL);
		imageHDC = CreateCompatibleDC(hdc);
		ReleaseDC(NULL, hdc);
		oldBitmap = SelectObject(imageHDC, ugl_image->image_handle);
	} else {
		imageHDC = ugl_image->selected_DC->hdc;
	}
	
	red = (rgb >> 16) & 0xFF;
	green = (rgb >> 8) & 0xFF;
	blue = (rgb) & 0xFF;

	color = RGB(red, green, blue);
	color = RGB_SWAP(color);

#ifdef _WIN32_WCE
	SetPixel(imageHDC, x, y, color);
#else
	SetPixelV(imageHDC, x, y, color);
#endif

	if (ugl_image->selected_DC == NULL) {
		SelectObject(imageHDC, oldBitmap);
		DeleteDC(imageHDC);
	}
	*/
}
//========================================================================================

void Image_GetPixelsHighColor(UGL_Int handle, UGL_IntArray pixels, UGL_Int length, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	HBITMAP hBitmap = ugl_image->image_handle;
    BITMAP bitmapObject;
	int depth = 0;

	LPBYTE buffer = NULL;

	/* Get the HDC for the device */

	if (!Image_IsDirect(handle, uglError)) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Pixels High color invalid on indexed color image"));
		return;
	}


//	if (ugl_image->selected_DC != NULL) {
//		SelectObject(ugl_image->selected_DC->hdc, ugl_image->null_bitmap);
//	}
 
    GetObject(hBitmap, sizeof(BITMAP), &bitmapObject);
	depth = bitmapObject.bmPlanes * bitmapObject.bmBitsPixel;


//	if (depth == 32) {
		// for 32-bpp image the pixel data can be copied directly into the allocated pixels
		// as this image is not padded in Windows
//		buffer = (LPBYTE)pixels;
//	} else {
		// allocate space for the pixel data
//		buffer = (LPBYTE)calloc(sizeof(BYTE), pbmi->bmiHeader.biSizeImage);
		//if (buffer == NULL) {
			//Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to calloc memory"));
			//free(pbmi);
			//return;
		//} 
//	}

	// get the pixel data
	buffer = bitmapObject.bmBits;

	{
		// For 24-bpp and 16-bpp images covert the windows pixel data.
		int output_index = 0;
		int buffer_index = 0;

		for (output_index = 0; output_index < length; output_index++) {
			UGL_Int *pixel = &pixels[output_index];
			if (depth == 32) {
				unsigned long *ppixel_32bit = (unsigned long *)&buffer[buffer_index];
				*pixel = *ppixel_32bit;
//				printf("target_pixel= %d", *pixel);
				buffer_index += sizeof(unsigned long);
			}else if (depth == 24) {
				// For 24-bpp, each 3-byte triplet in the bitmap array represents the 
				//relative intensities of blue, green, and red, respectively, for a pixel. 
				*pixel = (buffer[buffer_index++]);  //blue
				*pixel |= (buffer[buffer_index++]) << 8;  //green
				*pixel |= (buffer[buffer_index++]) << 16;  //red
			} else {
				// For 16-bpp, each WORD in the bitmap array represents a single pixel. 
				// The relative intensities of red, green, and blue are represented 
				// with five bits for each color component. The value for blue is in the 
				// least significant five bits, followed by five bits each for green and red. 
				// The most significant bit is not used.
				unsigned short *ppixel_16bit = (unsigned short *)&buffer[buffer_index];
				int red = ((*ppixel_16bit) & 0x7C00) << 1;
				int green = ((*ppixel_16bit) & 0x3E0) >> 5;
				int blue = (*ppixel_16bit) & 0x1F;
				// we only have to convert max green in 16 bit 5-5-5 to the max
				// green in 5-6-5.
				green = ((green * 0x03F) / 0x01F) << 5;

				*pixel = red | green | blue;
				buffer_index += sizeof(unsigned short);
			}

			if (((output_index + 1) % bitmapObject.bmWidth) == 0){
				// at the end of one row of pixel data. Each scan
				// line is padded with zeroes to end on a LONG boundary
				int remainder = ((buffer_index) % sizeof(long));
				if (remainder != 0) {
					buffer_index += sizeof(long) - remainder;
				}
			}
		}
	}

//	if (ugl_image->selected_DC != NULL) {
//		SelectObject(ugl_image->selected_DC->hdc, ugl_image->image_handle);
//	} 
}
//========================================================================================

void Image_GetPixelsLowColor(UGL_Int handle, UGL_ByteArray pixels, UGL_Int length, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	HBITMAP hBitmap = ugl_image->image_handle;
    BITMAP bitmapObject;

	LPBYTE buffer = NULL;

	if (Image_IsDirect(handle, uglError)) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Pixels Low color invalid on direct color image"));
		return;
	}

//	if (ugl_image->selected_DC != NULL) {
//		SelectObject(ugl_image->selected_DC->hdc, ugl_image->null_bitmap);
//	}
 
    GetObject(hBitmap, sizeof(BITMAP), &bitmapObject);

	// get the pixel data from the bitmap
	buffer = bitmapObject.bmBits;

	// destination, source, length
	MoveMemory(pixels, &buffer[0], length);

//	if (ugl_image->selected_DC != NULL) {
//		SelectObject(ugl_image->selected_DC->hdc, ugl_image->image_handle);
//	}
}
//========================================================================================

UGL_Int Image_GetTransparentColor(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);

	if (ugl_image->is_direct)
		return ugl_image->transparent_color;
	// when the image in indexed, the transparent color indicates whether one
	// of the pixels is transparent
	if (ugl_image->transparent_color == -1)
		return -1;
	// return the pixel which is transparent for indexed images
	return ugl_image->transparent_pixel;
}
//========================================================================================

UGL_Boolean Image_IsDirect(UGL_Int handle, UGL_Error uglError) {
	PUGL_Image ugl_image = UGL_IMAGE(handle);
	return ugl_image->is_direct;
}
//========================================================================================
UGL_Int Image_NewFromByteData(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_ByteArray imageData, UGL_Int bytesPerLine, UGL_Int depth, UGL_ByteArray indexedReds, UGL_ByteArray indexedGreens, UGL_ByteArray indexedBlues, UGL_Int paletteSize, UGL_Int transparentPixel, UGL_Error uglError) {
	PUGL_Image ugl_image = NULL;
	void *pBits[1];
	LPBYTE buffer;
	HBITMAP hDib;
	BITMAPINFO bmi;
	BITMAPINFO *pbmi;
	int imageData_length = bytesPerLine * height;
	int offset, index, bufferIndex, transparentColor;
	LPRGBQUAD pcolortable;
	UINT windowsScanline;

	/*
	* Note on WinCE.  CreateDIBSection requires the biCompression
	* field of the BITMAPINFOHEADER to be set to BI_BITFIELDS for
	* 16 and 32 bit direct images (see MSDN for CreateDIBSection).
	* In this case, the color mask can be set to any value.  For
	* consistency, it is set to the same mask used by non WinCE
	* platforms in BI_RGB mode.
	*/

	/*
	 * BUG in Windows 98:
	 * A monochrome DIBSection will display as solid black
	 * on Windows 98 machines, even though it contains the
	 * correct data. The fix is to convert 1-bit ImageData
	 * into 4-bit ImageData before creating the image.
	 */

	/* Windows does not support 2-bit images. Convert to 4-bit image. */

	/*
	* Note on WinCE.  CreateDIBSection requires the biCompression
	* field of the BITMAPINFOHEADER to be set to BI_BITFIELDS for
	* 16 and 32 bit direct images (see MSDN for CreateDIBSection).
	* In this case, the color mask can be set to any value.  For
	* consistency, it is set to the same mask used by non WinCE
	* platforms in BI_RGB mode.
	*/
	memset(&bmi, 0,sizeof(BITMAPINFOHEADER));
	bmi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmi.bmiHeader.biWidth = width;
	bmi.bmiHeader.biHeight = -height;
	bmi.bmiHeader.biPlanes = 1;
	bmi.bmiHeader.biBitCount = depth;
	bmi.bmiHeader.biCompression = BI_RGB;
	// don't create a palette any larger than necessary
	bmi.bmiHeader.biClrUsed = paletteSize;
	pbmi = (BITMAPINFO *)calloc(sizeof(BITMAPINFOHEADER) + (paletteSize * sizeof(RGBQUAD)), 1);
	if (pbmi == NULL) {
		Win32Error_SetError(uglError, UGL_ERROR_OUT_OF_MEMORY, _T("Not enough memory to create image"));
		return 0;
	}
	MoveMemory((void *)pbmi, (void *)&bmi.bmiHeader, sizeof(BITMAPINFOHEADER));

	offset = sizeof(BITMAPINFOHEADER);
	pcolortable = (LPRGBQUAD)(&pbmi->bmiColors);

	for (index = 0; index < paletteSize; index++) {
		pcolortable[index].rgbRed = indexedReds[index];
		pcolortable[index].rgbBlue = indexedBlues[index];
		pcolortable[index].rgbGreen = indexedGreens[index];
		pcolortable[index].rgbReserved = 0;
	}

	hDib = CreateDIBSection(NULL, (CONST BITMAPINFO *)pbmi, DIB_RGB_COLORS, (void **)&pBits, NULL, 0x0);
	if (hDib == 0) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to create device-independent image handle"));
		free(pbmi);
		return NULL_HANDLE;
	}

	// since the bytesPerLine might not match the scanline pad required by Windows, copy
	// the pixels to our bitmap a row at at time. At the end of each row of pixel data,
	// pad with zeroes to end on a LONG boundary.
	buffer = (LPBYTE)pBits[0];
	bufferIndex = 0;
	windowsScanline = _Image_GetScanlineLength(width,bmi.bmiHeader.biBitCount);
	for (index = 0; index < imageData_length; index += bytesPerLine) {
		memcpy((void *) &buffer[bufferIndex], (void *) &imageData[index], bytesPerLine);
		bufferIndex += windowsScanline;
	}

	free(pbmi);

	if (transparentPixel != -1) {
		transparentColor = indexedBlues[transparentPixel] << 16 |
						   indexedGreens[transparentPixel] << 8 |
						   indexedReds[transparentPixel];
	} else {
		transparentColor = -1;
	}

	return (UGL_Int)_Image_New(hDib, FALSE, transparentPixel, transparentColor, uglError);
}
//========================================================================================
