/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"
#include "UGL_Win32_Widget.h"
#include "UGL_win32_display.h"
#include "UGL_Win32_DC.h"

#include "Widget.h"
#include "Control.h"
#include "Image.h"
#include "Label.h"
#include "Graphics.h"

//========================================================================================
static UGL_Int _Label_ExStyle(UGL_Int ugl_style);
static UGL_Int _Label_Style(UGL_Int ugl_style);
LRESULT CALLBACK _Label_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
static void _Label_GetMinimumSize(UGL_Int handle, int prefWidth, POINT *point, UGL_Error uglError);
static int _Label_ApplyAlignmentStyle(PUGL_Label label, int oldStyle);
void _Label_OwnerDraw(PUGL_Control handle, LPDRAWITEMSTRUCT draw_struct); 
void Label_DrawImage(PUGL_Control);
//========================================================================================

#ifdef _WIN32_WCE
/**
 * Filters and passes incomming messages to the real windowProc.
 */
LRESULT CALLBACK _Label_FilterProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam);
#endif

UGL_Int Label_New(UGL_Int parentHandle, UGL_Int ugl_style, UGL_Error uglError) {
	int style = _Label_Style(ugl_style);
	int extended_style = _Label_ExStyle(ugl_style);

	PUGL_Control ugl_label = _Control_New(_T("STATIC"), sizeof(UGL_Label), _Label_Proc, 
						parentHandle, ugl_style, style, extended_style, uglError); 
	
	ugl_label->inheritsBackground = TRUE;

#ifdef _WIN32_WCE
	((PUGL_Label)ugl_label)->realWidgetProc = ugl_label->widgetProc;
	ugl_label->widgetProc = _Label_FilterProc;
#endif

	return (UGL_Int)ugl_label;
}

#ifdef _WIN32_WCE
/**
 * Filters and passes incomming messages to the real windowProc.
 */
LRESULT CALLBACK _Label_FilterProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	// Filter certian messages from the native windowProc because the native window proc does 
	// something that we don't want.
	switch (message) {
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE: {
			// Filter mouse events from going to the native windowProc because
			// the default behavior is to pass the message along to the parent.
			// Passing the message to the parent results in multiple windows getting
			// events for the same mouse messages.
			return 0;
		}


	}

	{
		UGL_Label* label = (UGL_Label*)GetWindowLong(hWnd, GWL_USERDATA);
		return CallWindowProc(label->realWidgetProc, hWnd, message, wParam, lParam);
	}
}
#endif


//========================================================================================
LRESULT CALLBACK _Label_Proc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	
//	PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
//	//FocusCallback(UGL_WIDGET(ugl_control)->callback_target, message);
	switch (message) {
		case UI_GET_MINIMUM_SIZE: {
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			UGL_Int handle = (UGL_Int)ugl_control;
			POINT* point = (POINT*)wParam;
			_Label_GetMinimumSize(handle, -1, point, (UGL_Error)lParam);
			return 0;
		} break;
		
		case WM_LBUTTONDOWN: {
			// STATIC controls don't receive focus, so we need to fake it out
			// by attempting to retrieve focus whenever the left mouse
			// button is pressed
			int returnVal;
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);

			returnVal = SendMessage(ugl_control->hWnd, UI_SET_FOCUS, (WPARAM)ugl_control->hWnd, (LPARAM)0);
			return returnVal != 0;
		} break;

		case WM_SIZE: {
			LRESULT result = _Control_Proc(hWnd, message, wParam, lParam);
			BOOL needsRepaint = FALSE;

			// Windows does not always repaint the control correctly on resize.  
			// When an alignment other than SS_LEFTNOWORDWRAP is used it does not update 
			// correctly
			LONG style = GetWindowLong(hWnd, GWL_STYLE);
			if ((style | (SS_LEFT | SS_CENTER | SS_RIGHT)) != 0) {
				needsRepaint = TRUE;
			}

			if (needsRepaint) {
				InvalidateRect(hWnd, NULL, TRUE);
			}
			return result;
		}

		case WM_SETFONT: {
			// Windows will drop/forget about the font
			// when an image is displayed, so caching the
			// current font will allow us to restore it when
			// the image is removed
			PUGL_Label label = (PUGL_Label)GetWindowLong(hWnd, GWL_USERDATA);
			label->font = (HFONT)wParam;
			break;
		}

		case UI_DRAWITEM: {
			
			
			PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
			LPDRAWITEMSTRUCT drawItemStruct = (LPDRAWITEMSTRUCT)lParam;	
			_Label_OwnerDraw(ugl_control, drawItemStruct);
			return TRUE;
			
			
		}		

#ifdef _WIN32_WCE


		// fix for Label.setAlignment, paint to image ourselves to adjust the position depending on the image size
		case WM_PAINT: {
			
				PUGL_Control ugl_control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
				LONG bits = GetWindowLong(hWnd, GWL_STYLE);
				_Control_Proc(hWnd, message, wParam, lParam);
					if ((bits & SS_BITMAP) == SS_BITMAP) {
						Label_DrawImage(ugl_control);
					} 

					return TRUE;
						}
		
		case WM_ERASEBKGND: {
			LONG bits = GetWindowLong(hWnd, GWL_STYLE);
			
			if ((bits & SS_BITMAP) == SS_BITMAP) {
				/**
				 * On PocketPC, the background of a label with the SS_BITMAP
				 * style flag set will not be cleared automaticly by the OS.
				 */
				HDC hdc = (HDC)wParam;

				HBRUSH newBrush;
				HGDIOBJ oldBrush;
				RECT rect;

				int rop2 = 0;
				int dwRop;

				PUGL_Control control = (PUGL_Control)GetWindowLong(hWnd, GWL_USERDATA);
				PUGL_Label label_control = (PUGL_Label)control;

				if (control->background_color != NULL) {
					newBrush = Color_GetBrush(control->background_color);
				} else{
					newBrush = GetSysColorBrush(COLOR_STATIC);
				}					
				oldBrush = SelectObject(hdc, newBrush);
				rop2 = SetROP2(hdc, R2_COPYPEN);
				SetROP2(hdc, rop2);
				dwRop = rop2 == R2_XORPEN ? PATINVERT : PATCOPY;	

				GetClipBox(hdc, &rect);				
				PatBlt(hdc, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, dwRop); 
				SelectObject(hdc, oldBrush);
				
				return 1;
			}
			break;
		} 
#endif		

	}

	

	return _Control_Proc(hWnd, message, wParam, lParam);
}
//=======================================================================================

void Label_SetImage(UGL_Int handle, UGL_Int imageHandle, UGL_Error uglError) {
	
	PUGL_Control ugl_control = (PUGL_Control)handle;
	PUGL_Image ugl_image = UGL_IMAGE(imageHandle);
	PUGL_Label ugl_label = (PUGL_Label)ugl_control;
	int newBits;
	int oldBits;
	int flags = SWP_NOZORDER | SWP_DRAWFRAME | SWP_NOACTIVATE | SWP_NOMOVE;
	RECT rect;
	HWND hWnd = ugl_control->hWnd;
	
	GetWindowRect(hWnd, &rect);

	newBits = GetWindowLong (hWnd, GWL_STYLE);
	oldBits = newBits;
	newBits &= ~(SS_BITMAP | SS_ICON );
	if (ugl_image != NULL) {
#ifdef _WIN32_WCE
		newBits |= SS_BITMAP ;
#else	
		newBits |= SS_OWNERDRAW;
#endif
	} else {
		// Because SS_BITMAP conflicts with the native
		// alignment styles when the image is set to NULL
		// the alignment styles need to be restored
		newBits = _Label_ApplyAlignmentStyle(ugl_label, newBits);
	}

	if (newBits != oldBits) {
		SetWindowLong(hWnd, GWL_STYLE, newBits);
	}

	/*
	* Feature in Windows.  When STM_SETIMAGE is used to set the
	* image for a static control, Windows either streches the image
	* to fit the control or shrinks the control to fit the image.
	* While not stricly wrong, neither of these is desirable.
	* The fix is to stop Windows from stretching the image by
	* using SS_REALSIZEIMAGE and SS_CENTERIMAGE, allow Windows
	* to shrink the control, and then restore the control to the
	* original size.
	*/


	if (imageHandle == 0) {
		// When an image is displayed in a label, the current font
		// will be dropped.  The fix is to save the font and to restore 
		// it when the image is removed.
		SendMessage(hWnd, WM_SETFONT, (WPARAM)ugl_label->font, FALSE);
	}

	ugl_label->ugl_image = ugl_image;

	// Always repaint the control to either display the new image, or
	// to restore the text
	InvalidateRect(hWnd, NULL, TRUE);
	
}
//========================================================================================

void Label_SetText(UGL_Int handle, UGL_String string, UGL_Error uglError) {
	
	PUGL_Control label = (PUGL_Control)handle;

	/**
	 * If the style of the Label contains the SS_BITMAP flag
	 * then the SetWindowText call will silently be ignored.
	 * To avoid the problem, unsetting SS_BITMAP and restoring
	 * it afterwards does the trick.
	 */
	LONG style = GetWindowLong(label->hWnd, GWL_STYLE);
	
	if ((style & SS_BITMAP) == SS_BITMAP) {
		SetWindowLong(label->hWnd, GWL_STYLE, style & ~SS_BITMAP);
	}
	SetWindowText(label->hWnd, string);

	if ((style & SS_BITMAP) == SS_BITMAP) {
		SetWindowLong(label->hWnd, GWL_STYLE, style);
	}
}
//========================================================================================

/**
 * Returns the new native style with the native alignment updated
 */
int _Label_ApplyAlignmentStyle(PUGL_Label label, int oldStyle) {
	
	int nativeAlignment;
	int alignment = label->alignment;

	if (IS_FLAG_SET(alignment, UGL_STYLE_LEAD)) {
		if (IS_FLAG_SET(label->control.widget.ugl_style, UGL_STYLE_WRAP)) {
			nativeAlignment = SS_LEFT;
		} else {
			nativeAlignment = SS_LEFTNOWORDWRAP;
		}
	} else if (IS_FLAG_SET(alignment, UGL_STYLE_TRAIL)) {
		nativeAlignment = SS_RIGHT;
	} else if (IS_FLAG_SET(alignment, UGL_STYLE_CENTER)) {
		nativeAlignment = SS_CENTER;
	}

	if (IS_FLAG_SET(label->control.widget.ugl_style, UGL_STYLE_MIDDLE)) nativeAlignment |= SS_CENTERIMAGE;

	oldStyle &= ~(SS_LEFT | SS_LEFTNOWORDWRAP | SS_CENTER | SS_RIGHT | SS_CENTERIMAGE);
	return oldStyle | nativeAlignment;
}

//========================================================================================

void Label_SetAlignment(UGL_Int handle, UGL_Int alignment, UGL_Error uglError) {
	
	PUGL_Label ugl_label = (PUGL_Label)handle;

	int bits = GetWindowLong(ugl_label->control.hWnd, GWL_STYLE);
	int style = bits;

	ugl_label->alignment = alignment;

	/*
	* Feature in Windows.  The windows label does not align
	* the bitmap or icon.  Any attempt to set alignment bits
	* such as SS_CENTER cause the label to display text.  The
	* fix is to disallow alignment.
	*
	* NOTE: SS_BITMAP and SS_ICON are not single bit
	* masks so it is necessary to test for all of the
	* bits in these masks.
	*/
	if (IS_FLAG_SET(bits, SS_BITMAP)) return;
	if (IS_FLAG_SET(bits, SS_ICON)) return;
	style &= ~(SS_LEFT | SS_LEFTNOWORDWRAP | SS_CENTER | SS_RIGHT);
	style = _Label_ApplyAlignmentStyle(ugl_label, style);
	
	if (style != bits) {
		SetWindowLong(ugl_label->control.hWnd, GWL_STYLE, style);
		_Control_Redraw(&(ugl_label->control));
	}
}

/**
 * Computes the minimum size of the label based on the width passed in.  This
 * method is used for labels instead of Control_GetMinimumSize because the
 * WRAP flag makes the width important when calculating the height.
 * 
 * @param handle the label
 * @param width the desired width, this value will be -1 if no
 *              desired width is provided.
 */
void Label_GetMinimumSize(UGL_Int handle, UGL_Int width, UGL_Int* returnX, UGL_Int* returnY, UGL_Error uglError) {
	
	POINT p;
	
	_Label_GetMinimumSize(handle, width, &p, uglError);
	*returnX = p.x;
	*returnY = p.y;
}


//========================================================================================

void _Label_GetMinimumSize(UGL_Int handle, int prefWidth, POINT *point, UGL_Error error) {
	
	PUGL_Control control = UGL_CONTROL(handle);
	PUGL_Label ugl_label = (PUGL_Label)control;
	HWND hWnd = control->hWnd;

	/*
	* NOTE: SS_BITMAP and SS_ICON are not single bit
	* masks so it is necessary to test for all of the
	* bits in these masks.
	*/
	int bits = GetWindowLong(hWnd, GWL_STYLE);
	BOOL isBitmap = FALSE;
	BOOL isIcon = FALSE;


	int width = 0, height = 0, widthBuffer = 0;

	int border = _Control_GetBorderWidth(hWnd);

#ifndef _WIN32_WCE
	isIcon = (bits & SS_OWNERDRAW) == SS_OWNERDRAW;//(bits & SS_ICON) == SS_ICON;
#endif
	isBitmap = (bits & SS_BITMAP) == SS_BITMAP;
	if (isBitmap || isIcon) {
		if (ugl_label->ugl_image != NULL) {
			int image_x, image_y, image_width, image_height;

			Image_GetBounds((UGL_Int)ugl_label->ugl_image, &image_x, &image_y, &image_width, &image_height, error);				
			width = image_width;
			height = image_height;
		}
	} else {
		HDC hDC = GetDC(hWnd);
		HGDIOBJ newFont = (HGDIOBJ)SendMessage(hWnd, WM_GETFONT, 0, 0);
		HGDIOBJ oldFont = SelectObject(hDC, newFont);
		RECT rect;
		int length;
		
		length = GetWindowTextLength(hWnd);
		if (length > 0) {
			LPTSTR window_text;

			int flags = DT_CALCRECT | DT_EXPANDTABS;
		
#ifndef _WIN32_WCE
			flags |= DT_EDITCONTROL;
#endif

			SetRect(&rect, 0, 0, 1, 0);

			if ((prefWidth != -1) && (!(IS_FLAG_SET(bits, SS_LEFTNOWORDWRAP)))) {
				rect.right = prefWidth;
				flags |= DT_WORDBREAK;
			}

			window_text = (LPTSTR)calloc(sizeof(TCHAR), length + 2);
			GetWindowText(hWnd, window_text, length + 1);
			DrawText(hDC, window_text, length, &rect, flags);
			free(window_text);

			width = rect.right - rect.left;
			height = rect.bottom - rect.top;
		} else {
			TEXTMETRIC tm;
			GetTextMetrics(hDC, &tm);
			height = tm.tmHeight;
		}
		if (newFont != 0) SelectObject(hDC, oldFont);
		ReleaseDC(hWnd, hDC);
	}
	width += border * 2; height += border * 2;
	width += widthBuffer;
	
#ifdef _WIN32_WCE
	/* 
	* Feature in WinCE PPC.  Text labels have a trim
	* of one pixel wide on the right and left side.
	* The fix is to increase the size.
	*/		
	if (!isBitmap && !isIcon) width += 2;		
#endif
	
	point->x = width;
	point->y = height;
}
//========================================================================================

UGL_Int _Label_Style(UGL_Int ugl_style) {
	
	//int bits = _Control_Style(ugl_style) & (~WS_TABSTOP);
	//bits |= SS_NOTIFY;
	int bits = _Control_Style(ugl_style);
	bits |= SS_NOTIFY | WS_TABSTOP;


	if (IS_FLAG_SET(ugl_style, UGL_STYLE_WRAP)) bits |= SS_LEFT;
	else bits |= SS_LEFTNOWORDWRAP;
	
	if (IS_FLAG_SET(ugl_style, UGL_STYLE_MIDDLE)) bits |= SS_CENTERIMAGE;
	return bits;
}
//========================================================================================

UGL_Int _Label_ExStyle(UGL_Int ugl_style) {
	
	int bits = 0; 

#ifndef _WIN32_WCE
	bits |= WS_EX_NOINHERITLAYOUT;
#endif

	bits &= ~WS_EX_CLIENTEDGE;

	if (IS_FLAG_SET(ugl_style, UGL_STYLE_BORDER)) bits |= WS_EX_STATICEDGE;

	return bits;
}
//========================================================================================

// fix for Label.setAlignment, paint to image ourselves to adjust the position depending on the image size
void Label_DrawImage(PUGL_Control handle) {
	
#ifdef _WIN32_WCE
	
	PUGL_Control ugl_control = (PUGL_Control)handle;
	PUGL_Label ugl_label = (PUGL_Label)ugl_control;
	PUGL_Image ugl_image = ugl_label->ugl_image;
	HWND hWnd = ugl_control->hWnd;
	HDC hdc = GetDC(hWnd);
	HBRUSH newBrush;
	BITMAP bitmap;
	int alignment;
	HBRUSH backgroundBrush;


	

	if(ugl_image != NULL){
		HGDIOBJ oldBrush;
		RECT rect;
		PUGL_DC ugl_dc;
		int rop2 = 0;
		int dwRop;
		newBrush = CreatePatternBrush(ugl_label->ugl_image->image_handle);
		oldBrush = SelectObject(hdc, newBrush);
		rop2 = SetROP2(hdc, SRCCOPY);
		SetROP2(hdc, rop2);
		dwRop = SRCCOPY;	
		GetClipBox(hdc, &rect);	

		// Paint the background for the label
		if (ugl_control->background_color != NULL) {
			backgroundBrush = Color_GetBrush(ugl_control->background_color);
		} else {
			backgroundBrush = GetSysColorBrush(COLOR_BTNFACE);
		}
		FillRect(hdc, &rect, backgroundBrush);

		GetObject(ugl_label->ugl_image->image_handle, sizeof(BITMAP), &bitmap);		
		ugl_dc = _internal_new_DC(hdc, NULL,&(rect), NULL);

		// check alignment and paint the image in accordance to the alignment
		alignment = ugl_label->alignment;
			
		// fix for bug 156893 the original bitblt draws from the top of the image even when the label containing the image is partly scrolled
		// out of visible area.  Using Graphics_DrawImage fixes this problem.
		if (IS_FLAG_SET(alignment, UGL_STYLE_LEAD)) {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, 0, 0, NULL);
		} else if (IS_FLAG_SET(alignment, UGL_STYLE_TRAIL)) {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, ((rect.right-rect.left)-bitmap.bmWidth), 0, NULL);
		} else if (IS_FLAG_SET(alignment, UGL_STYLE_CENTER)) {				
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, ((rect.right-rect.left)-bitmap.bmWidth)/2, 0, NULL);
		} else {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, 0, 0, NULL);
		}
		SelectObject(hdc, oldBrush);
		DeleteObject(newBrush);
	}

	ReleaseDC(hWnd, hdc);
#endif
}

//========================================================================================

void _Label_OwnerDraw(PUGL_Control handle, LPDRAWITEMSTRUCT draw_struct) {

	int imageWidth;
	BITMAP bitmap;
	LPRECT rect = &draw_struct->rcItem;
	int alignment;
	PUGL_Label ugl_label = UGL_LABEL(handle);
	PUGL_Control ugl_control = UGL_CONTROL(handle);
	PUGL_Image image = ugl_label->ugl_image;
	HBRUSH backgroundBrush;
	HDC hdc = draw_struct->hDC;
		
	if (image != NULL) {
		
		HRGN clip = CreateRectRgn(0, 0, 0, 0);
		PUGL_DC ugl_dc;
		RECT shiftedClip;
		
		int res = GetClipRgn(draw_struct->hDC, clip);

		// Update the background color of the button
		if (ugl_control->background_color != NULL) {
		backgroundBrush = Color_GetBrush(ugl_control->background_color);
	} else {
		backgroundBrush = GetSysColorBrush(COLOR_BTNFACE);
	}
	FillRect(hdc, rect, backgroundBrush);


		GetObject(ugl_label->ugl_image->image_handle, sizeof(BITMAP), &bitmap);
		imageWidth = bitmap.bmWidth;
		alignment = ugl_label->alignment;
		ugl_dc = _internal_new_DC(draw_struct->hDC, NULL,&(shiftedClip), NULL);
		
		/*
		* Bug in Windows.  When a bitmap is included in the
		* menu bar, the HDC seems to already include the left
		* coordinate.  The fix is to ignore this value when
		* the item is in a menu bar.
		*/
		
	if (IS_FLAG_SET(alignment, UGL_STYLE_LEAD)) {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, 0, 0, NULL);
		} else if (IS_FLAG_SET(alignment, UGL_STYLE_TRAIL)) {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, ((rect->right-rect->left)-imageWidth), 0, NULL);
		} else if (IS_FLAG_SET(alignment, UGL_STYLE_CENTER)) {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, ((rect->right-rect->left)-imageWidth)/2, 0, NULL);
		} else {
			Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_label->ugl_image, 0, 0, NULL);
		}

		SelectClipRgn(draw_struct->hDC, res == 1 ? clip : NULL );
		DeleteObject(clip);
	}
	
}

//==============================================================================================================
