/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "ugl_win32.h"
#include "UGL_Win32_Widget.h"
#include "UGL_win32_display.h"
#include "UGL_Win32_Menu.h"
#include "UGL_Win32_DC.h"
#include "KeyUtils.h"
#include "Platform.h"

#ifdef _WIN32_WCE
#include <commctrl.h>
#include <sipapi.h>

#if (_WIN32_WCE >= 420)
#include <aygshell.h>
#endif
#endif /* _WIN32_WCE */

// for softkey
#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
#include "i18n.h"
#include "Resource.h"

#define MENU_MSG_BUFF_SIZE	100
#endif

#include <assert.h>

#include "Widget.h"
#include "Control.h"
#include "Image.h"
#include "MenuItem.h"
#include "Menu.h"
#include "Display.h"


#define MENU_ITEM_TEXT_LENGTH 128

//========================================================================================
static void _CascadeMenuItem_SetMenu(PUGL_MenuItem ugl_cascade, PUGL_Menu ugl_dropdown, UGL_Error uglError);
static void _LabeledMenuItem_SetText(PUGL_MenuItem ugl_menuItem, UGL_String text, UGL_String accelText, UGL_Error uglError);
static void _AbstractMenuItem_SetEnabled(PUGL_MenuItem ugl_menuItem, UGL_Boolean enabled, UGL_Error uglError);
static PUGL_MenuItem _MenuItem_New_WinCe(PUGL_Menu ugl_Menu, PUGL_MenuItem ugl_menuItem, int index, int type);
static PUGL_MenuItem _MenuItem_New(UGL_Int menuHandle, int type, BOOLEAN isCascade, int index, UGL_Error uglError);
static LRESULT _wmDrawChild(UGL_Int handle, WPARAM wParam, LPARAM lParam);
static LRESULT _wmMeasureChild(UGL_Int handle, WPARAM wParam, LPARAM lParam);
static LRESULT _wmCommandChild(UGL_Int handle, WPARAM wParam, LPARAM lParam);
static void _Dispose(PUGL_Menu ugl_menu, int id, UGL_Error uglError);
void _Dispose_MenuItem(UGL_Int handle, UGL_Error uglError);
void _Update_Enabled(PUGL_Display display, HWND hwndCB);
//========================================================================================
#ifdef _WIN32_WCE
static HMENU getMenuForCheck(HMENU parent_menu);
#endif // #ifdef _WIN32_WCE

#ifdef _WIN32_WCE
static HMENU getMenuForCheck(PUGL_Menu parent_menu)
{
	HMENU hMenu = parent_menu->hMenu;
	// if the parent menu is the menubar, we need to find the submenu
	// of the popup so that the check will actually work
	if (parent_menu->hwndCB != 0)
	{
		HMENU hMenuNew = SendMessage (parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
		if (hMenuNew != 0)
		{
			hMenu = hMenuNew;
		}
	}

	return hMenu;
}
#endif // #ifdef _WIN32_WCE

void Menu_Redraw(PUGL_Menu menu) {
#if _WIN32_WCE && !WIN32_PLATFORM_PSPC && !WIN32_PLATFORM_WFSP
	
	if (menu->hwndCB != NULL) {
		if (!CommandBar_DrawMenuBar(menu->hwndCB, 1)) {
			return; // Something went wrong
		}
	} else {
		if (menu->cascadeItem != NULL) {
			Menu_Redraw(menu->cascadeItem->parent_menu);
		}
	}
	
#endif

#ifndef _WIN32_WCE
	DrawMenuBar(menu->parent->hWnd);
#endif
}

/**
 * Returns the position of the menu item with the 
 * specified id in the parent menu.
 */
UINT FindMenu(HMENU parent, UINT id) {
	UINT pos;
	MENUITEMINFO info;
	info.cbSize = sizeof(MENUITEMINFO);
	info.fMask = MIIM_DATA;

	for (pos=0;;pos++) {
		if (!GetMenuItemInfo(parent, pos, TRUE, &info)) {
			return -1;
		}


		// When a submenu is inserted the id of the cascade menu item is
		// set to the address of the submenu. Therefore we cannot use the 
		// wId field.
		if (info.dwItemData == id) return pos;
	}
}

/**
 * Updates the cascade item to have the new menu item specified.  
 * This method is used instead of SetMenuItemInfo because SetMenuItemInfo
 * will delete the previous menu automaticly.  This behavior is not desired 
 * and this method provides the behavior we want without the previous menu
 * being deleted.
 */
void _CascadeMenuItem_SetMenu(PUGL_MenuItem ugl_cascade, PUGL_Menu ugl_dropdown, UGL_Error uglError) {
	HMENU parentMenu = ugl_cascade->parent_menu->hMenu;
	int id = ugl_cascade->id;
	MENUITEMINFO info;
	int flags;
	HMENU dropdown;
	TCHAR oldText[MENU_ITEM_TEXT_LENGTH];
	

	// Calculate the position of the menu item with the 
	// id that we are looking for by walking through the 
	// list untill we get a match or and error.  This code is not very
	// pretty but it is required that we use the position instead of the id
	// when accessing the menu item below.  See the comments below for more
	// details.
	int pos = FindMenu(parentMenu, id);
	if (pos == -1) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to find the item"));
		return;
	}

	// The text from the old menu item needs be retrieved so that it can be used in the new item
	memset(&info, 0, sizeof(MENUITEMINFO));
	info.cbSize = sizeof(MENUITEMINFO);
#ifndef _WIN32_WCE
	info.fMask = MIIM_STRING; // MSDN says that the dwTypeData is only used when MIIM_STRING is used
#else
	info.fMask = MIIM_TYPE;
#endif
	info.dwTypeData = oldText;
	info.cch = MENU_ITEM_TEXT_LENGTH;
	if (!GetMenuItemInfo(parentMenu, pos, TRUE, &info)) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get the menu item text"));
		return;		
	}

	// Insert the new menu item at the same location as the old item.  
	// NOTE:  The insert call must use MF_BYPOSITION since windows will
	// not properly find menu items by id if they have a submenu set.  
	flags = MF_BYPOSITION;
	if (ugl_dropdown != 0) {
		flags |= MF_POPUP;
		dropdown = ugl_dropdown->hMenu;
	} else {
		dropdown = 0;
	}
	if (!InsertMenu(parentMenu, pos, flags, (UINT)dropdown, oldText)) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to insert new cascade menu item"));
		return;
	}

	// When a submenu is inserted the id of the cascade menu item is 
	// set to the address of the submenu. Set our internal item id 
	// (global item index) in the data field.
	memset(&info, 0, sizeof(MENUITEMINFO));
	info.cbSize = sizeof(MENUITEMINFO);
	info.fMask = MIIM_DATA;
	info.dwItemData = id;

//fix for Bug#:107883 make sure the image retains when menuitem cascades
#if (_WIN32_WCE < 500)
	#if !(defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP))
		info.fMask |= MIIM_BITMAP;
		info.hbmpItem = HBMMENU_CALLBACK;
	#endif
#endif

	
	SetMenuItemInfo(parentMenu, pos, TRUE, &info);

	// If the new item was inserted correctly the old item can be removed
	// NOTE:  The delete call must use MF_BYPOSITION since windows will
	// not properly find menu items by id if they have a submenu set.  

	// NOTE:  Remove menu should be used because DeleteMenu will destroy
	// the submenu which is not what we want.
	if (!RemoveMenu(parentMenu, pos+1, MF_BYPOSITION)) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to delete old menu item"));
	}
}

void CascadeMenuItem_SetMenu(UGL_Int cascadeMenuItemHandle, UGL_Int dropDownMenuHandle, UGL_Error uglError) {

	PUGL_MenuItem ugl_cascade = UGL_MENUITEM(cascadeMenuItemHandle);
	PUGL_Menu ugl_dropdown = UGL_MENU(dropDownMenuHandle);
	UGL_Display *display = display = UGL_WIDGET(ugl_cascade)->display;


	#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	{
		TBBUTTONINFO info; 
		PUGL_MenuItem tempItem;
		BOOL found;
		int i;
		found = FALSE;
		memset(&info, 0, sizeof(TBBUTTONINFO));
		// check for softkey support
		if (Platform_GetMajorVersion() < 5) {
			HWND hwndCB = ugl_cascade->parent_menu->hwndCB;
			int id = ugl_cascade->id;
			if (hwndCB != 0) {
				HWND oldMenu = (HWND)SendMessage(hwndCB, SHCMBM_SETSUBMENU, id, (LPARAM) ugl_dropdown->hMenu);
			} else {
				_CascadeMenuItem_SetMenu(ugl_cascade, ugl_dropdown, uglError);
			}
		} else {
			HWND hwndCB = ugl_cascade->parent_menu->hwndMB;
			HWND realhwndCB = ugl_cascade->parent_menu->hwndCB;
			HMENU hMenu = (HMENU) SendMessage (ugl_cascade->parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
			_TCHAR msgBuff[MENU_MSG_BUFF_SIZE];
			// if main menu
			if (hwndCB != 0) {
				// if the first item on the button is being set a  submenu, find a item that doesn't have a submenu to replace it
				if(ugl_cascade->index == 0) {
					// if found, set the item's index to 0 and increment every other items by 1.
					for(i = 1; i < display->menuItemsLength; i++){
						tempItem = Display_FindMenuItem(display, realhwndCB, i);
						if(tempItem != NULL && !tempItem->hasSubMenu){
							// move the first item that doesn't have submenu into the first spot
							RemoveMenu (hMenu, tempItem->index-1, MF_BYPOSITION);
							InsertMenu (hMenu, 0, MF_BYPOSITION | MF_POPUP, (UINT) ugl_dropdown->hMenu, ugl_cascade->text);
							Display_DecMenuItemIndex(display, realhwndCB, tempItem->index);
							Display_IncMenuItemIndex(display, realhwndCB, 0);
							tempItem->index = 0;
							info.cbSize = sizeof(TBBUTTONINFO);
							info.dwMask = TBIF_TEXT;
							info.pszText = (LPWSTR) tempItem->text;
							if(!SendMessage (ugl_cascade->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
								Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
							}
							LoadResourceString(display->hInstance, IDS_U_MENU, msgBuff, MENU_MSG_BUFF_SIZE);
							info.pszText = msgBuff;
							SendMessage (ugl_cascade->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);
							found = TRUE;

							//ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
							//ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);

							if (display->activeShell == ugl_cascade->parent_menu->parent) {
								ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
								ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
							} /*else {
								ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_HIDE);
								ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
							}*/


							break;
						}
					}
					// if not found, simply set the left softkey to be blank and add one item with submenu on the right menu.
					if(!found){
						info.cbSize = sizeof(TBBUTTONINFO);
						info.dwMask = TBIF_TEXT;
						info.pszText = (LPWSTR) _T("");
						if(!SendMessage (ugl_cascade->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
						}
						Display_IncMenuItemIndex(display, realhwndCB, 0);
						InsertMenu (hMenu, 0, MF_BYPOSITION | MF_POPUP, (UINT) ugl_dropdown->hMenu, ugl_cascade->text);
						LoadResourceString(display->hInstance, IDS_U_MENU, msgBuff, MENU_MSG_BUFF_SIZE);
						info.pszText = msgBuff;
						SendMessage (ugl_cascade->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);

						if (display->activeShell == ugl_cascade->parent_menu->parent) {
							ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
							ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
						} /*else {
							ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_HIDE);
							ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
						}*/
						//ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
						//ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
					}
					ugl_cascade->hasSubMenu=TRUE;
				} else {
					// if adding submenu to menuitem, directly add it.
					ugl_cascade->hasSubMenu=TRUE;
					RemoveMenu (hMenu, ugl_cascade->index-1, MF_BYPOSITION);
					InsertMenu (hMenu, ugl_cascade->index-1, MF_BYPOSITION | MF_POPUP, (UINT) ugl_dropdown->hMenu, ugl_cascade->text);

					// if index = 1, it's likely the right softkey is not set to IDS_U_MENU, load the resource and set it //Uriel
					if(ugl_cascade->index == 1) {
						info.cbSize = sizeof(TBBUTTONINFO);
						info.dwMask = TBIF_TEXT;
						LoadResourceString(display->hInstance, IDS_U_MENU, msgBuff, MENU_MSG_BUFF_SIZE);
						info.pszText = msgBuff;
						SendMessage (ugl_cascade->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);
					}

				}
				// when adding submenu to main menu, always show MB menu bar
				//ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
				//ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
				if (display->activeShell == ugl_cascade->parent_menu->parent) {
					ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
					ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
				} /*else {
					ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_HIDE);
					ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
				}*/


			} else {
				// submenu
				_CascadeMenuItem_SetMenu(ugl_cascade, ugl_dropdown, uglError);
				if (display->activeShell == ugl_cascade->parent_menu->parent) {
					ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
					ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
				}/* else {
					ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_HIDE);
					ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);
				}*/

				//ShowWindow(ugl_cascade->parent_menu->hwndMB, SW_SHOW);
				//ShowWindow(ugl_cascade->parent_menu->hwndCB, SW_HIDE);

			}
		}
	}
#else 
	_CascadeMenuItem_SetMenu(ugl_cascade, ugl_dropdown, uglError);
#endif

	{
		if (ugl_cascade->dropDownMenu != NULL) {
#if defined(DEBUG) && !defined(_WIN32_WCE)
			if (IsMenu(ugl_cascade->dropDownMenu->hMenu) == FALSE) {
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Previous menu was diposed when removing it from menuitem"));
			}
#endif
			ugl_cascade->dropDownMenu->cascadeItem = NULL;
		}
		if (ugl_dropdown != NULL) ugl_dropdown->cascadeItem = ugl_cascade;
		ugl_cascade->dropDownMenu = ugl_dropdown;
	}

	Menu_Redraw(ugl_cascade->parent_menu);
}
//========================================================================================

UGL_Int StateMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Int style, UGL_Error uglError) {
	int type = (IS_FLAG_SET(style, UGL_STYLE_RADIO) ? MFT_RADIOCHECK : MFT_STRING);
	
	PUGL_MenuItem ugl_menuItem = _MenuItem_New(menuHandle, type, FALSE, index, uglError);
	UGL_WIDGET(ugl_menuItem)->ugl_style = style | UGL_STYLE_CHECK;
	return (UGL_Int)ugl_menuItem;
}
//========================================================================================

void LabeledMenuItem_SetFont(UGL_Int menuItemHandle, UGL_Int fontHandle, UGL_Error uglError) {
	//NO OP
}
//========================================================================================

UGL_Int CascadeMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error uglError) {
	int type = MFT_STRING;
	PUGL_MenuItem ugl_menuItem = NULL;
	PUGL_Menu ugl_menu = UGL_MENU(menuHandle);


	ugl_menuItem = _MenuItem_New(menuHandle, type, TRUE, index, uglError);

	Menu_Redraw(ugl_menu);

	return (UGL_Int)ugl_menuItem;
}
//========================================================================================

void StateMenuItem_SetSelection(UGL_Int statefulMenuItemHandle, UGL_Boolean selected, UGL_Error uglError) {
	PUGL_MenuItem ugl_item = UGL_MENUITEM(statefulMenuItemHandle);

	HMENU hMenu = ugl_item->parent_menu->hMenu;
	int item_id = ugl_item->id; //ugl_item->menu_item_info.wID;

#ifdef _WIN32_WCE
	int flags;
	int count;
#else
	MENUITEMINFO info;
	BOOL success;
#endif // #ifdef _WIN32_WCE

	// cache this so we are in sync with the java side
	// the java side sets the selected state before setting
	// it natively
	ugl_item->isSelected = selected;

#ifdef _WIN32_WCE
	count = Display_MenuItemCount(UGL_WIDGET(ugl_item)->display, ugl_item->parent_menu->hwndCB);
	if (count <= 2)
	{
		// there aren't more than 2 menu items, so don't set the checked state yet
		// because both of the menu items are on the menu bar
		return;
	}

	hMenu = getMenuForCheck(ugl_item->parent_menu);

	flags = MF_BYCOMMAND | (selected ? MF_CHECKED : MF_UNCHECKED);
	if (CheckMenuItem(hMenu, item_id, flags) == -1) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Menu item doesn't exist"));
		return;
	}
#else
	info.cbSize = sizeof(MENUITEMINFO);
	info.fMask = MIIM_STATE;
	success = GetMenuItemInfo(hMenu, item_id, FALSE, (LPMENUITEMINFO)&info);
	if (!success) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to get current state menu selection"));
		return;
	}

	info.fState &= ~MFS_CHECKED;
	if (selected) info.fState |= MFS_CHECKED;
	success = SetMenuItemInfo (hMenu, item_id, FALSE, (LPMENUITEMINFO)&info);
	
	if (!success) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to get current state menu selection"));
		return;
	}
#endif // defined _WIN32_WCE

	Menu_Redraw(ugl_item->parent_menu);
}

void StateMenuItem_DisableAutoSelection(UGL_Int statefulMenuItemHandle, UGL_Boolean disableAutoSelection)
{
	PUGL_MenuItem ugl_item = UGL_MENUITEM(statefulMenuItemHandle);
	ugl_item->disableAutoSelection = disableAutoSelection;
}

//========================================================================================

UGL_Int SeparatorItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error uglError) {
	int type = MFT_SEPARATOR;
	
	return (UGL_Int)_MenuItem_New(menuHandle, type, FALSE, index, uglError);
}
//========================================================================================

UGL_Int PushMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error uglError) {
	int type = MFT_STRING;
	
	return (UGL_Int)_MenuItem_New(menuHandle, type, FALSE, index, uglError);
	
}
//========================================================================================

void _LabeledMenuItem_SetText(PUGL_MenuItem ugl_menuItem, UGL_String text, UGL_String accelText, UGL_Error uglError) {
	MENUITEMINFO info;	
	LPTSTR combined = NULL;


		int accelLength;
	
		int pos = FindMenu(ugl_menuItem->parent_menu->hMenu, ugl_menuItem->id);
		if (pos == -1) {
			Win32Error_SetError(uglError, GetLastError(), _T("Unable to find the item"));
			return;
		}

		accelLength = lstrlen(accelText);
		if (accelLength > 0) {
			combined = (LPTSTR)calloc(sizeof(TCHAR), accelLength + lstrlen(text) + 2);
			if (combined == NULL) {
				Win32Error_SetError(uglError, UGL_ERROR_OUT_OF_MEMORY, _T(""));
				return;
			} else {
				lstrcat(combined, text);
				lstrcat(combined, _T("\t"));
				lstrcat(combined, accelText);
			}
		}
	
	
		/*
		* Bug in Windows 2000.  For some reason, when MIIM_TYPE is set
		* on a menu item that also has MIIM_BITMAP, the MIIM_TYPE clears
		* the MIIM_BITMAP style.  The fix is to reset both MIIM_BITMAP.
		* Note, this does not happen on Windows 98.
		*/
		//boolean hasBitmap = false;
		//if (!OS.IsWinCE && (OS.WIN32_MAJOR << 16 | OS.WIN32_MINOR) >= (4 << 16 | 10)) {
		//	info.fMask = OS.MIIM_BITMAP;
		//	OS.GetMenuItemInfo (hMenu, id, false, info);
		//	hasBitmap = info.hbmpItem != 0;
		//}
		
		/* Use the character encoding for the default locale */
		
		info.cbSize = sizeof(MENUITEMINFO);
	#ifndef _WIN32_WCE
		info.fMask = MIIM_STRING;
	#else
		info.fMask = MIIM_TYPE; // This would be nicer as MIIM_STRING but that doesn't exist on WinCE
		info.fType = MFT_STRING;
	#endif
		info.dwTypeData = combined == NULL ? text : combined;

		// NOTE:  The SetMenuItemInfo call must use MF_BYPOSITION since windows will
		// not properly find menu items by id if they have a submenu set.
		if (!SetMenuItemInfo(ugl_menuItem->parent_menu->hMenu, pos, TRUE, &info)) {
			Win32Error_SetError(uglError, GetLastError(), _T("Unable to update the menu item's label"));
		} else {
			Menu_Redraw(ugl_menuItem->parent_menu);
		}

		if (combined != NULL) free(combined);

		/*
		* Restore the bitmap that was removed to work around a problem
		* in GetMenuItemInfo() and menu items that have bitmaps set with
		* MIIM_BITMAP.
		*/
		//if (!OS.IsWinCE && (OS.WIN32_MAJOR << 16 | OS.WIN32_MINOR) >= (4 << 16 | 10)) {
		//	if (hasBitmap) {
		//		info.fMask = OS.MIIM_BITMAP;
		//		info.hbmpItem = OS.HBMMENU_CALLBACK;
		//		success = OS.SetMenuItemInfo (hMenu, id, false, info);
		//	}
		//}
}

//========================================================================================

void LabeledMenuItem_SetText(UGL_Int menuItemHandle, UGL_String text, UGL_String accelText, UGL_Error uglError) {
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(menuItemHandle);

#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	{
		UGL_Display *display = display = UGL_WIDGET(ugl_menuItem)->display;
		HWND hwndCB = ugl_menuItem->parent_menu->hwndCB;
		UGL_String revise;
		UGL_Boolean isRevise = UGL_FALSE;
		UGL_Int i, j, len;
		for(i=0, len=0; *(text+i)!=0; i++) {
			if(*(text+i) != '&')
				len++;
			else
				isRevise = UGL_TRUE;
		}
		if(isRevise) {
			revise = (UGL_String)calloc(len+1, sizeof(UGL_String));
			for(i=0, j=0; *(text+i)!=0; i++) {
				if(*(text+i) != '&') {
					*(revise+j) = *(text+i);
					j++;
				}
			}
		} else {
			revise = text;
		}

		if (hwndCB != 0) {
			// top level "menu"bar
			TBBUTTONINFO buttonInfo;

			memset(&buttonInfo, 0, sizeof(TBBUTTONINFO));
			buttonInfo.cbSize = sizeof(TBBUTTONINFO);
			buttonInfo.dwMask = TBIF_TEXT | TBIF_STATE;
			buttonInfo.pszText = (LPWSTR) revise;
			buttonInfo.fsState &= ~TBSTATE_ENABLED;
			if (ugl_menuItem->enabled){
				buttonInfo.fsState |= TBSTATE_ENABLED;
			}

			// check for softkey support
			if (Platform_GetMajorVersion() >= 5) {

				ugl_menuItem->text = (UGL_String)calloc(lstrlen(revise), sizeof(UGL_String));			
				lstrcpy(ugl_menuItem->text, revise);			

				//if first item, set left softkey text
				if(ugl_menuItem->index == 0){
					
						if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						}
						
						if (!SendMessage(ugl_menuItem->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						}
						

				} else if (ugl_menuItem->index == 1){
					
					//if second item and only two items, set to right softkey
					if(display->menuItemCount == 2 && !ugl_menuItem->hasSubMenu){
						if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						}
						if (!SendMessage(ugl_menuItem->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						}
					} else {
						// else we need to set the text of the softkey's menu Item.					
						MENUITEMINFO info;
						HMENU hMenu = (HMENU) SendMessage (ugl_menuItem->parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						memset(&info, 0, sizeof(MENUITEMINFO));
						info.cbSize = sizeof(MENUITEMINFO);
						info.fMask = MIIM_TYPE;
						info.dwTypeData = revise;

						if(!SetMenuItemInfo(hMenu, ugl_menuItem->index-1, TRUE, &info)){
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						} 
						
					}
				} else {
					// else simply set the text for the menuitem
						MENUITEMINFO info;
						HMENU hMenu = (HMENU) SendMessage (ugl_menuItem->parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						memset(&info, 0, sizeof(MENUITEMINFO));
						info.cbSize = sizeof(MENUITEMINFO);
						info.fMask = MIIM_TYPE;
						info.dwTypeData = revise;

						if(!SetMenuItemInfo(hMenu, ugl_menuItem->index-1, TRUE, &info)){
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
						}
				}
			} else {	
				if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ugl_menuItem->id, (LPARAM) &buttonInfo)) {
					Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item text"));
				}
			}
		} else {
			// submenu
			_LabeledMenuItem_SetText(ugl_menuItem, revise, accelText, uglError);
		}
		if(isRevise)
			free(revise);
	}

#else
	_LabeledMenuItem_SetText(ugl_menuItem, text, accelText, uglError);
#endif
}
//========================================================================================

void LabeledMenuItem_SetAccelerator(UGL_Int menuItemHandle, UGL_Int accelerator, UGL_Error uglError) {
	PUGL_MenuItem menuItem = UGL_MENUITEM(menuItemHandle);
	PUGL_Shell parent= UGL_SHELL(menuItem->parent_menu->parent);
	WORD cmd = (short)menuItem->id;

	if (accelerator != 0) {
        ACCEL entry;
		int modifiers = (accelerator >> 16) & 0x0000FFFF;

		entry.cmd = cmd;

		// Although not documented in MSDN, it appears that TranslateAccelerator will
		// only honor the FALT and FCONTROL constants if FVIRTKEY is used as well.
		entry.fVirt = FVIRTKEY;
		if (IS_FLAG_SET(modifiers, UGL_MODIFIER_MASK_ALT))		entry.fVirt |= FALT;
		if (IS_FLAG_SET(modifiers, UGL_MODIFIER_MASK_CONTROL))	entry.fVirt |= FCONTROL;
		if (IS_FLAG_SET(modifiers, UGL_MODIFIER_MASK_SHIFT))	entry.fVirt |= FSHIFT;

		entry.key = KeyUtils_ConvertToNativeKeycode(accelerator & 0x0000FFFF);

		_Shell_SetAccelerator(parent, &entry, uglError);
	} else {
		_Shell_RemoveAccelerator(parent, cmd);
	}
}
//========================================================================================

void _AbstractMenuItem_SetEnabled(PUGL_MenuItem ugl_menuItem, UGL_Boolean enabled, UGL_Error uglError) {
	int bits;
	HMENU hMenu = ugl_menuItem->parent_menu->hMenu;

	int pos = FindMenu(hMenu, ugl_menuItem->id);
	if (pos == -1) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to find the item"));
		return;
	}

	bits = MF_BYPOSITION | (enabled ? MF_ENABLED : MF_GRAYED);

	if (EnableMenuItem(hMenu, pos, bits) == -1) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("The menu item does not exist"));
		return;
	}

	Menu_Redraw(ugl_menuItem->parent_menu);
}
//========================================================================================

void AbstractMenuItem_SetEnabled(UGL_Int menuItemHandle, UGL_Boolean enabled, UGL_Error uglError) {
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(menuItemHandle);

#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	{
		HWND hwndCB = ugl_menuItem->parent_menu->hwndCB;

		if (hwndCB != NULL) {
			TBBUTTONINFO buttonInfo;
			memset(&buttonInfo, 0, sizeof(TBBUTTONINFO));
			buttonInfo.cbSize = sizeof(TBBUTTONINFO);
			buttonInfo.dwMask = TBIF_STATE;
			
			
			// if has softkey support
			if (Platform_GetMajorVersion() >= 5) {
				
				UGL_Display *display = display = UGL_WIDGET(ugl_menuItem)->display;
				ugl_menuItem->enabled = enabled;
				buttonInfo.fsState &= ~TBSTATE_ENABLED;
				if (enabled){
					buttonInfo.fsState |= TBSTATE_ENABLED;
				}
				if(ugl_menuItem->index == 0){
					
						if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item enabled state"));
						}
						
						if (!SendMessage(ugl_menuItem->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item enabled state"));
						}
						

				} else if (ugl_menuItem->index == 1){
					
					//if second item and only two items, set to right softkey
					if(display->menuItemCount == 2 && !ugl_menuItem->hasSubMenu){
						if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (LPARAM) &buttonInfo)) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item enabled state"));
						}
					} else {
						// else we need to set the state of the softkey's menu Item.
						int bits;
						HMENU hMenu = (HMENU) SendMessage (ugl_menuItem->parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						bits = MF_BYPOSITION | (enabled ? MF_ENABLED : MF_GRAYED);

						if (EnableMenuItem(hMenu,  ugl_menuItem->index-1, bits) == -1) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("The menu item does not exist"));
							return;
						}
					}
				} else {
					// else simply set the state for the menuitem
						int bits;
						HMENU hMenu = (HMENU) SendMessage (ugl_menuItem->parent_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						bits = MF_BYPOSITION | (enabled ? MF_ENABLED : MF_GRAYED);

						if (EnableMenuItem(hMenu,  ugl_menuItem->index-1, bits) == -1) {
							Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("The menu item does not exist"));
							return;
						}
				}

			} else {
				// else no softkey support
				if (!SendMessage(hwndCB, TB_GETBUTTONINFO, ugl_menuItem->id, (LPARAM) &buttonInfo)) {
					Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item enabled state"));
				}
				buttonInfo.fsState &= ~TBSTATE_ENABLED;
				if (enabled) buttonInfo.fsState |= TBSTATE_ENABLED;
				if (!SendMessage(hwndCB, TB_SETBUTTONINFO, ugl_menuItem->id, (LPARAM) &buttonInfo)) {
					Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set menu item enabled state"));
				}
			}
		} else {
			_AbstractMenuItem_SetEnabled(ugl_menuItem, enabled, uglError);
		}
		
	}
#else
	_AbstractMenuItem_SetEnabled(ugl_menuItem, enabled, uglError);
#endif
}
//========================================================================================

void LabeledMenuItem_SetImage(UGL_Int menuItemHandle, UGL_Int imageHandle, UGL_Error uglError) {
	//if ((style & SWT.SEPARATOR) != 0) return;
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(menuItemHandle);
	HMENU hMenu = ugl_menuItem->parent_menu->hMenu;
	int id = ugl_menuItem->id;

#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	TBBUTTONINFO buttonInfo;
	HIMAGELIST imagelist;			
	int imageWidth, imageHeight, index;
	BITMAP bitmap;
	HANDLE hBitmap;
	HBITMAP himg, hmask, hOldBitmap;
	HDC hdc, hdc1, hdc2;

	HWND hwndCB = ugl_menuItem->parent_menu->hwndCB;
	
	ugl_menuItem->ugl_image = UGL_IMAGE(imageHandle);
	GetObject(ugl_menuItem->ugl_image->image_handle, sizeof(BITMAP), &bitmap);
	
	imageWidth = bitmap.bmWidth;
	imageHeight = bitmap.bmHeight;

	if(hwndCB != 0)
	{
		(long) imagelist = SendMessage(hwndCB, TB_GETIMAGELIST, 0, 0);
		if(imagelist == NULL){
			// create imagelist if it's not previously created
			imagelist = ImageList_Create(imageWidth, 20, ILC_COLOR | ILC_MASK, 16, 16);
			SendMessage (hwndCB, TB_SETIMAGELIST, 0, (LPARAM) imagelist);
		} 
		// add image to the imagelist
		
		
		ImageList_SetIconSize(imagelist, imageWidth, imageHeight);
		hBitmap = ugl_menuItem->ugl_image != NULL ? ugl_menuItem->ugl_image->image_handle : NULL;
		
		// prepare HBATMAP handle's for image and the mask
		GetObject(hBitmap, sizeof(BITMAP), &bitmap);
		hdc = GetDC(0);
		hdc1 = CreateCompatibleDC(hdc);
		SelectObject(hdc1, hBitmap);
		hdc2 = CreateCompatibleDC(hdc);
		himg = CreateCompatibleBitmap(hdc, bitmap.bmWidth, bitmap.bmHeight);
		SelectObject(hdc2, himg);
		BitBlt(hdc2, 0, 0, bitmap.bmWidth, bitmap.bmHeight, hdc1, 0, 0, SRCCOPY);

		DeleteDC(hdc1);
		DeleteDC(hdc2);
		ReleaseDC(0, hdc);
		// now we have Hbitmap for image

		GetObject(hBitmap, sizeof(BITMAP), &bitmap);
		hmask = CreateBitmap(bitmap.bmWidth, bitmap.bmHeight, 1, 1, NULL);
		hdc = GetDC(0);
		hdc1 = CreateCompatibleDC(hdc);

		hOldBitmap = SelectObject(hdc1, hmask);
		PatBlt(hdc1, 0, 0, bitmap.bmWidth, bitmap.bmHeight, BLACKNESS);
//		SelectObject(hdc1, hOldBitmap);
		
		ReleaseDC(0, hdc);
		DeleteDC(hdc1);
		// now we have the mask

		index = ImageList_Add(imagelist, himg, hmask);			
		DeleteObject(hmask);
		DeleteObject(himg);

		memset(&buttonInfo, 0, sizeof(TBBUTTONINFO));
		buttonInfo.cbSize = sizeof(TBBUTTONINFO);
		buttonInfo.dwMask = TBIF_TEXT;
		buttonInfo.dwMask |= TBIF_IMAGE;   
		if(index != -1){
			buttonInfo.iImage = index;
		} 
		SendMessage(hwndCB, TB_SETBUTTONINFO, ugl_menuItem->id, (LPARAM) &buttonInfo);
	}
	else
	{
		// the menuItem in PPC doesn't receive the drawchild and messurechild event
			MENUITEMINFO info;
			info.cbSize = sizeof(MENUITEMINFO);
			info.fMask = MIIM_TYPE;
			info.fType = MFT_OWNERDRAW;
			SetMenuItemInfo(hMenu, id, TRUE, &info);
	}
	
#endif

#ifndef _WIN32_WCE
	ugl_menuItem->ugl_image = UGL_IMAGE(imageHandle);
	
	if (ugl_menuItem->parent_menu->isMenuBar == FALSE) {

		// Touch all items in the menu when the image changes on one of them so
		// that they will all layout correctly.  See the _wmMeasureChild method 
		// for more information about why each item should be redrawn

		int itemCount = GetMenuItemCount(hMenu);
		int i;

		for (i=0;i<itemCount;i++) {
			MENUITEMINFO info;

			info.cbSize = sizeof(MENUITEMINFO);
			info.fMask = MIIM_BITMAP;
			info.hbmpItem = HBMMENU_CALLBACK;
			SetMenuItemInfo(hMenu, i, TRUE, &info);
		}
	} else {
		MENUITEMINFO info;
		int pos = FindMenu(hMenu, ugl_menuItem->id);
		if (pos == -1) {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Failed to find menu item"));
			return;
		}

		info.cbSize = sizeof(MENUITEMINFO);
		info.fMask = MIIM_BITMAP;
		info.hbmpItem = imageHandle == 0 ? NULL : HBMMENU_CALLBACK;
		if (!SetMenuItemInfo(hMenu, pos, TRUE, &info)) {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Failed to set image on menu item"));
			return;
		}
	}

	Menu_Redraw(ugl_menuItem->parent_menu);
#endif // ifndef _WIN32_WCE
}
//========================================================================================
PUGL_MenuItem _MenuItem_New_WinCe(PUGL_Menu ugl_menu, PUGL_MenuItem ugl_menuItem, int index, int type) {
	int bits = 0;
	MENUITEMINFO info;

	if (IS_FLAG_SET(type, MFT_SEPARATOR)) bits |= MF_SEPARATOR;
	else if (IS_FLAG_SET(type, MFT_STRING)) bits |= MF_STRING;
	if (!InsertMenu(ugl_menu->hMenu, index, MF_BYPOSITION | bits, (UINT)ugl_menuItem->id, _T(""))) {
		return UGL_MENUITEM(NULL_HANDLE);
	}
	
	memset(&info, 0, sizeof(MENUITEMINFO));
	info.cbSize = sizeof(MENUITEMINFO);
	info.fMask = MIIM_DATA;
	info.dwItemData = ugl_menuItem->id;
	
	SetMenuItemInfo(ugl_menu->hMenu, ugl_menuItem->id, FALSE, &info);

	return ugl_menuItem;
}

//========================================================================================

PUGL_MenuItem _MenuItem_New(UGL_Int menuHandle, int type, BOOLEAN isCascade, int index, UGL_Error uglError) {
	PUGL_MenuItem ugl_menuItem = NULL;
	PUGL_Menu ugl_menu = UGL_MENU(menuHandle);

	ugl_menuItem = UGL_MENUITEM(calloc(sizeof(UGL_MenuItem), 1));
	UGL_WIDGET(ugl_menuItem)->display = Control_GetDisplay(ugl_menu->parent);
	ugl_menuItem->parent_menu = ugl_menu;
	ugl_menuItem->wmDrawChild = _wmDrawChild;
	ugl_menuItem->wmMeasureChild = _wmMeasureChild;
	ugl_menuItem->wmCommandChild = _wmCommandChild;
	ugl_menuItem->enabled = TRUE;
	ugl_menuItem->disableAutoSelection = FALSE;
	ugl_menuItem->isSelected = FALSE;
	UGL_WIDGET(ugl_menuItem)->widget_dispose = _Dispose_MenuItem;

	
	
#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
{
	// if major version > 5, use softkey support code.
	if (Platform_GetMajorVersion() >= 5)
	{

		TBBUTTONINFO info; 
		HMENU hMenu;
		PUGL_MenuItem tempItem;
		_TCHAR msgBuff[MENU_MSG_BUFF_SIZE];
		
		UGL_Display *display;
		int i,j;
		//int count = UGL_WIDGET(ugl_menu)->display->menuItemCount;
		int count; 
		int bits = 0;
		int enable = 0;
		FILE * pFile; 

		if (IS_FLAG_SET(type, MFT_SEPARATOR)) ugl_menuItem->isSeparator = TRUE;
		else ugl_menuItem->isSeparator = FALSE;
		ugl_menuItem->index = index;
		ugl_menuItem->text = _T("");
		ugl_menuItem->hasSubMenu = FALSE;


		memset(&info, 0, sizeof(TBBUTTONINFO));
		display = UGL_WIDGET(ugl_menuItem)->display;	
		count = Display_MenuItemCount(display, ugl_menu->hwndCB);


		/*
		pFile = fopen ("\\myfile.txt","a"); 
		fprintf(pFile, "display->activeShell = %d, ugl_menu->parent = %d;\n", display->activeShell, ugl_menu->parent); 
		fclose(pFile);
		*/
		if (display->activeShell == ugl_menu->parent) {
			display->activeMenuBar = ugl_menu;
		}

		
		// if main menu (on menu Bar)
		if(ugl_menu->hwndCB !=0){
			// if left softkey position is empty put the newly created item directly in it
			

			if(Display_FindMenuItem(display, ugl_menu->hwndCB, 0) == NULL && count > 0) {
				info.cbSize = sizeof(info); 
				info.dwMask = TBIF_TEXT;
				info.pszText = _T("");
				SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	

				ugl_menuItem->index = 0;
				Display_AddMenuItem(ugl_menuItem);
				count = UGL_WIDGET(ugl_menu)->display->menuItemCount;
				//ShowWindow(ugl_menu->hwndMB, SW_SHOW);
				//ShowWindow(ugl_menu->hwndCB, SW_HIDE);
				if (display->activeShell == ugl_menu->parent) {
					ShowWindow(ugl_menu->hwndMB, SW_SHOW);
					ShowWindow(ugl_menu->hwndCB, SW_HIDE);
				} 

				/*
				pFile = fopen ("\\myfile.txt","a"); 
				fprintf(pFile, "=============== _MenuItem_New 1 ========================\n", count); 
				fprintf(pFile, "menuItemCount = %d\n", count); 
				fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
				fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
				for(i=0; i<display->menuItemsLength; i++){	
					if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
					if(display->menuItems[i] != NULL){
						// setting the left softkey text
						if(display->menuItems[i]->parent_menu->hwndCB != 0){
							fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
						}
					}
				}
				
				fprintf(pFile, "========================================================\n"); 
				fclose(pFile);
				*/

			} else {
				Display_AddMenuItem(ugl_menuItem);
				Display_IncMenuItemIndex(display, ugl_menu->hwndCB, ugl_menuItem->index);
				count = Display_MenuItemCount(display, ugl_menu->hwndCB);
				ugl_menuItem->index--;

				switch(Display_MenuItemCount(display, ugl_menu->hwndCB)){
					case 1: 
						// if it's the first item, add to the left softkey
						info.cbSize = sizeof(info); 
						info.dwMask = TBIF_TEXT;
						info.pszText = _T("");
						SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
						//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
						//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

						if (display->activeShell == ugl_menu->parent) {
							ShowWindow(ugl_menu->hwndCB, SW_SHOW);
							ShowWindow(ugl_menu->hwndMB, SW_HIDE);
						} 
	
						/*
						pFile = fopen ("\\myfile.txt","a"); 
						fprintf(pFile, "=============== _MenuItem_New 2 ========================\n", count); 
						fprintf(pFile, "menuItemCount = %d\n", count); 
						fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
						fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
						for(i=0; i<display->menuItemsLength; i++){						
							if(display->menuItems[i] != NULL){
								// setting the left softkey text
								if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
								if(display->menuItems[i]->parent_menu->hwndCB != 0){
									fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
								}
							}
						}
						
						fprintf(pFile, "========================================================\n", count); 
						fclose(pFile);
						*/

						break;
					case 2:
						// if it's the second, index == 0, add to left, may need to move the left text to right
						if(ugl_menuItem->index == 0){
							info.cbSize = sizeof(info); 
							info.dwMask = TBIF_TEXT;
							info.pszText = _T("");
							SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
						
							// if left key originally have text, move to right
							tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 1);
							if(tempItem != NULL && tempItem->text != NULL){												
								info.cbSize = sizeof(TBBUTTONINFO);
								info.dwMask = TBIF_TEXT;
								info.pszText = (LPWSTR) tempItem->text;
								if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info)){
									MessageBox(0, _T("unable to set button info"), _T("new"), 0); 
								}								
							}
							//_Update_Enabled(display, ugl_menu->hwndCB);
							/*if (display->activeShell == ugl_menu->parent) {
								ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								ShowWindow(ugl_menu->hwndCB, SW_SHOW);
								_Update_Enabled(display, ugl_menu->hwndCB);
							} else {
								ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								ShowWindow(ugl_menu->hwndCB, SW_HIDE);
							}*/


							// else if it's the right item
						} else {
							// do nothing
						}

						if (display->activeShell == ugl_menu->parent) {
							ShowWindow(ugl_menu->hwndCB, SW_SHOW);
							ShowWindow(ugl_menu->hwndMB, SW_HIDE);
							_Update_Enabled(display, ugl_menu->hwndCB);
						} 

						//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
						//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

						/*
						pFile = fopen ("\\myfile.txt","a"); 
						fprintf(pFile, "=============== _MenuItem_New 3 ========================\n", count); 
						fprintf(pFile, "menuItemCount = %d\n", count); 
						fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
						fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
						for(i=0; i<display->menuItemsLength; i++){						
							if(display->menuItems[i] != NULL){
								// setting the left softkey text
								if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
								if(display->menuItems[i]->parent_menu->hwndCB != 0){
									fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
								}
							}
						}
						
						fprintf(pFile, "========================================================\n", count); 
						fclose(pFile);
						*/


						break;
					case 3:
							// setting text for Menu
							hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
							info.cbSize = sizeof(info);
							info.dwMask = TBIF_TEXT;
							LoadResourceString(display->hInstance, IDS_U_MENU, msgBuff, MENU_MSG_BUFF_SIZE);
							info.pszText = msgBuff;
							SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);

							// insert menu
							bits = 0; 
							enable = 0;
							tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 0);
							if(tempItem != NULL){
								// setting the left softkey text
								if(tempItem->text != NULL){
									info.pszText = tempItem->text;
									SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
								}
								// setting menuitem text
								tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 1);
								if(tempItem != NULL){
									if(!tempItem->hasSubMenu) {	//no submenu											
										if(tempItem->isSeparator){
											bits |= MF_SEPARATOR;
										}											
										InsertMenu(hMenu, 0, MF_BYPOSITION | bits, tempItem->id, tempItem->text);																												
										
										enable = MF_BYPOSITION | (tempItem->enabled ? MF_ENABLED : MF_GRAYED);
										if (EnableMenuItem(hMenu,  0, enable) == -1) {
											Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("The menu item does not exist, enable failed"));
										}
										if (IS_FLAG_SET(UGL_WIDGET(tempItem)->ugl_style, UGL_STYLE_CHECK)) {
												if (tempItem->isSelected) {
													if (CheckMenuItem(hMenu, tempItem->id, MF_BYCOMMAND | MF_CHECKED) == -1) {
														// disregard the error
														// Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Menu item doesn't exist"));
														return;
													}
												}
										}
									}
								}

								bits = 0; 
								enable = 0;

								tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 2);
								if(tempItem != NULL){
									if(!tempItem->hasSubMenu){										
										if(tempItem->isSeparator){
											bits |= MF_SEPARATOR;
										}											

										InsertMenu(hMenu, 1, MF_BYPOSITION | bits, tempItem->id, tempItem->text);													
										
										enable = MF_BYPOSITION | (tempItem->enabled ? MF_ENABLED : MF_GRAYED);
										if (EnableMenuItem(hMenu, 1, enable) == -1) {
											Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("The menu item does not exist, enable failed"));
										}
										
									}
								}
								
							}
							

						if (display->activeShell == ugl_menu->parent) {
							ShowWindow(ugl_menu->hwndMB, SW_SHOW);
							ShowWindow(ugl_menu->hwndCB, SW_HIDE);
							_Update_Enabled(display, ugl_menu->hwndCB);
						} 

							//ShowWindow(ugl_menu->hwndMB, SW_SHOW);
							//ShowWindow(ugl_menu->hwndCB, SW_HIDE);
						/*
						pFile = fopen ("\\myfile.txt","a"); 
						fprintf(pFile, "=============== _MenuItem_New 4 ========================\n", count); 
						fprintf(pFile, "menuItemCount = %d\n", count); 
						fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
						fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB); 
						for(i=0; i<display->menuItemsLength; i++){						
							if(display->menuItems[i] != NULL){
								// setting the left softkey text
								if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
								if(display->menuItems[i]->parent_menu->hwndCB != 0){
									fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
								}
							}
						}
						
						fprintf(pFile, "========================================================\n", count); 
						fclose(pFile);
						*/


						break;

					default:
						// any item more than 3

							hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
							if (IS_FLAG_SET(type, MFT_SEPARATOR)) {
								bits |= MF_SEPARATOR;
							}
							else if (IS_FLAG_SET(type, MFT_STRING)) bits |= MF_STRING;

							if(ugl_menuItem->index != 0){
								// not to the left softkey, simply add it to the menu on the right
								InsertMenu(hMenu, ugl_menuItem->index-1, MF_BYPOSITION | bits, ugl_menuItem->id, _T(""));

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
								}


								//ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndCB, SW_HIDE);

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _MenuItem_New 6 ========================\n", count); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								fclose(pFile);
								fprintf(pFile, "========================================================\n", count); 

								*/

							} else {

								for(i=0; i<display->menuItemsLength; i++){
									if(display->menuItems[i] != NULL){									
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
									
										// insert it to the left softkey text
										if(display->menuItems[i]->index == 0 && display->menuItems[i]->parent_menu->hwndCB != 0){
											info.cbSize = sizeof(info);
											info.dwMask = TBIF_TEXT;
											info.pszText = _T("");
											SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
										}
										// insert the original left softkey item to the right
										
										
										if(display->menuItems[i]->index == 1 && display->menuItems[i]->parent_menu->hwndCB != 0){
											if(display->menuItems[i]->text != NULL){
												InsertMenu(hMenu, 0, MF_BYPOSITION, display->menuItems[i]->id, display->menuItems[i]->text);							
											} else {
												InsertMenu(hMenu, 0, MF_BYPOSITION, display->menuItems[i]->id, _T(""));							
											}
										}
									}
								}	

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									_Update_Enabled(display, ugl_menu->hwndCB);
								} 

								//ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndCB, SW_HIDE);
								
						/*
						pFile = fopen ("\\myfile.txt","a"); 
						fprintf(pFile, "=============== _MenuItem_New 7 ========================\n", count); 
						fprintf(pFile, "menuItemCount = %d\n", count); 
						fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
						fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
						for(i=0; i<display->menuItemsLength; i++){						
							if(display->menuItems[i] != NULL){
								// setting the left softkey text
								if(display->menuItems[i]->parent_menu->hwndCB != 0){
									fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
								}
							}
						}
						fclose(pFile);
						fprintf(pFile, "========================================================\n", count); 
						*/
						}


						break;
					}

				}
			// else if hWndCB == 0, sub menu

			} else {
				// sub menu
				Display_AddMenuItem(ugl_menuItem);
				//Display_IncMenuItemIndex(display, ugl_menuItem->index);
				if (!_MenuItem_New_WinCe(ugl_menu, ugl_menuItem, index, type)) {
					Display_RemoveMenuItem(ugl_menuItem);
					free(ugl_menuItem);
					Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to create new menu item"));
					return UGL_MENUITEM(NULL_HANDLE);
				}
			}


	}

	// else no softkey support
	else 
	{
		HWND hwndCB = ugl_menu->hwndCB;
		Display_AddMenuItem(ugl_menuItem);
		

		if (hwndCB != 0) {
			// top level "menu"bar
			TBBUTTON lpButton;
			memset(&lpButton, 0, sizeof(TBBUTTON));
			lpButton.idCommand = ugl_menuItem->id;
			lpButton.fsStyle = (BYTE)TBSTYLE_AUTOSIZE;


			if (IS_FLAG_SET(type, MFT_SEPARATOR)) lpButton.fsStyle = TBSTYLE_SEP;//(BYTE)BTNS_SEP;
			else if (isCascade) lpButton.fsStyle |= TBSTYLE_DROPDOWN | 0x80;
			
			lpButton.fsState = (BYTE)TBSTATE_ENABLED;
			lpButton.iBitmap = I_IMAGENONE;
			if (SendMessage (hwndCB, TB_INSERTBUTTON, index, (LPARAM) &lpButton) == FALSE) {
				Display_RemoveMenuItem(ugl_menuItem);
				free(ugl_menuItem);
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to create new menu item"));
				return UGL_MENUITEM(NULL_HANDLE);
			}
		} else {
			// sub menu
			if (!_MenuItem_New_WinCe(ugl_menu, ugl_menuItem, index, type)) {
				Display_RemoveMenuItem(ugl_menuItem);
				free(ugl_menuItem);
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to create new menu item"));
				return UGL_MENUITEM(NULL_HANDLE);
			}
		}
	}

}
#endif

#if _WIN32_WCE && !(WIN32_PLATFORM_PSPC) && !(WIN32_PLATFORM_WFSP) 
	{
		Display_AddMenuItem(ugl_menuItem);
		if (!_MenuItem_New_WinCe(ugl_menu, ugl_menuItem, index, type)) {
			Display_RemoveMenuItem(ugl_menuItem);
			free(ugl_menuItem);
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to create new menu item"));
			return UGL_MENUITEM(NULL_HANDLE);
		}
	}
#endif

#ifndef _WIN32_WCE
	Display_AddMenuItem(ugl_menuItem);
	/*
	* Bug in Windows.  For some reason, when InsertMenuItem()
	* is used to insert an item without text, it is not possible
	* to use SetMenuItemInfo() to set the text at a later time.
	* The fix is to insert the item with some text.
	* 
	* Feature in Windows.  When an empty string is used instead
	* of a space and InsertMenuItem() is used to set a submenu
	* before setting text to a non-empty string, the menu item
	* becomes unexpectedly disabled.  The fix is to insert a
	* space.
	*/
	{
	MENUITEMINFO info;
	info.cbSize = sizeof(MENUITEMINFO);
	info.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_DATA | MIIM_ID;
	info.dwItemData = (UINT)ugl_menuItem->id;
	info.wID = (UINT)ugl_menuItem->id;
	info.fType = type;
	info.dwTypeData = _T(" ");

	if (ugl_menu->isMenuBar == FALSE) {
		// Windows has problems with menus which contain items with and without images
		// the fix is to mark every item has having an image.  See the _wmMeasureChild method 
		// for more information.
		info.fMask |= MIIM_BITMAP;
		info.hbmpItem = HBMMENU_CALLBACK;
	}

	if (!InsertMenuItem(ugl_menu->hMenu, index, TRUE, &info)) {
		Display_RemoveMenuItem(ugl_menuItem);
		free(ugl_menuItem);
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to create new menu item"));
		return UGL_MENUITEM(NULL_HANDLE);
	}
	}
#endif
	
	return ugl_menuItem;
}
//========================================================================================

LRESULT _wmDrawChild(UGL_Int handle, WPARAM wParam, LPARAM lParam) {
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(handle);
	LPDRAWITEMSTRUCT draw_struct = (LPDRAWITEMSTRUCT)lParam;
	PUGL_Image image = ugl_menuItem->ugl_image;

	if (image != NULL) {
		HRGN clip = CreateRectRgn(0, 0, 0, 0);
		PUGL_DC ugl_dc;
		RECT shiftedClip;
		int x;

		int res = GetClipRgn(draw_struct->hDC, clip);
		
#ifndef _WIN32_WCE
		POINT viewportOrg;
		// The incomming clip rect is in logical co-ordinates, but
		// the clip going to new_DC is expected to be in device
		// co-ords.  These next lines will shift it into the correct
		// position
		GetViewportOrgEx(draw_struct->hDC, &viewportOrg);
		shiftedClip.top = draw_struct->rcItem.top + viewportOrg.y;
		shiftedClip.bottom = draw_struct->rcItem.bottom + viewportOrg.y;
		shiftedClip.left = draw_struct->rcItem.left + viewportOrg.x;
		shiftedClip.right = draw_struct->rcItem.right + viewportOrg.x;
#endif

		ugl_dc = _internal_new_DC(draw_struct->hDC, NULL,&(shiftedClip), NULL);
		/*
		* Bug in Windows.  When a bitmap is included in the
		* menu bar, the HDC seems to already include the left
		* coordinate.  The fix is to ignore this value when
		* the item is in a menu bar.
		*/
		x = draw_struct->rcItem.left;

		Graphics_DrawImage((UGL_Int)ugl_dc, (UGL_Int)ugl_menuItem->ugl_image, x, draw_struct->rcItem.top, NULL);

		SelectClipRgn(draw_struct->hDC, res == 1 ? clip : NULL );
		DeleteObject(clip);
	}

	return TRUE;
}
//========================================================================================

LRESULT _wmMeasureChild(UGL_Int handle, WPARAM wParam, LPARAM lParam) {

#ifndef _WIN32_WCE
	LPMEASUREITEMSTRUCT measure_struct = (LPMEASUREITEMSTRUCT)lParam;
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(handle);
	PUGL_Image image = ugl_menuItem->ugl_image;
	int width = 0, height = 0;

	if (image != NULL) {
		_Image_GetSize(image, &width, &height, NULL);
	} else {
		/*
		* Bug in Windows.  If a menu contains items that have
		* images and can be checked, Windows does not include
		* the width of the image and the width of the check when
		* computing the width of the menu.  When the longest item
		* does not have an image, the label and the accelerator
		* text can overlap.  The fix is to use SetMenuItemInfo()
		* to indicate that all items have a bitmap and then include
		* the width of the widest bitmap in WM_MEASURECHILD.
		*/
		HMENU hMenu = ugl_menuItem->parent_menu->hMenu;

		MENUITEMINFO info;
		int i;
		int maxWidth = 0;
		int itemCount = GetMenuItemCount(hMenu);
		info.cbSize = sizeof(MENUITEMINFO);
		info.fMask = MIIM_DATA;

		for (i=0;i<itemCount;i++) {
			PUGL_MenuItem item;
			GetMenuItemInfo(hMenu, i, TRUE, &info);

			item = Display_GetMenuItem(ugl_menuItem->widget.display, info.dwItemData);
			if (item->ugl_image != NULL) {
				_Image_GetSize(item->ugl_image, &width, &height, NULL);
				if (width > maxWidth) maxWidth = width;
			}
		}

		width = maxWidth;
		height = 0;
	}

	if (width != 0 || height != 0) {
		/*
		* Feature in Windows.  On Windows 98, it is necessary
		* to add 4 pixels to the width of the image or the image
		* and text are too close.  On other Windows platforms,
		* this causes the text of the longest item to touch the
		* accelerator text.  The fix is to add only 2 pixels in
		* this case.
		*/
		measure_struct->itemWidth = width + (/*OS.IsWin95 ? 4 :*/ 2);
		measure_struct->itemHeight = height + 4;
	} else {
		measure_struct->itemWidth = width;
		measure_struct->itemHeight = height;
	}

#endif // if defined _WIN32_WCE

	return 0;
}
//========================================================================================

LRESULT _wmCommandChild(UGL_Int handle, WPARAM wParam, LPARAM lParam) {
	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(handle);

	//FILE * pFile; 
	UGL_Display *display = display = UGL_WIDGET(ugl_menuItem)->display;
	int count = display->menuItemCount;
	int i;
	/*
	pFile = fopen ("\\myfile.txt","a"); 
	fprintf(pFile, "=============== wmCommandChild ========================\n", count); 
	fprintf(pFile, "menuItemCount = %d\n", count); 
	fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
	for(i=0; i<display->menuItemsLength; i++){						
		if(display->menuItems[i] != NULL){
			// setting the left softkey text
			if(display->menuItems[i]->parent_menu->hwndCB != 0){
				fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
			}
		}
	}
	fclose(pFile);
	fprintf(pFile, "========================================================\n", count); 
	*/

	if (IS_FLAG_SET(UGL_WIDGET(ugl_menuItem)->ugl_style, UGL_STYLE_CHECK)) {
		if (ugl_menuItem->disableAutoSelection == FALSE) {
			MENUITEMINFO info;
			int id = ugl_menuItem->id;
#ifdef _WIN32_WCE
			HMENU hMenu = getMenuForCheck(ugl_menuItem->parent_menu);
#else
			HMENU hMenu = ugl_menuItem->parent_menu->hMenu;
#endif // #ifdef _WIN32_WCE
			BOOL selected = FALSE;

			info.cbSize = sizeof(MENUITEMINFO);
			info.fMask = MIIM_STATE;
			GetMenuItemInfo(hMenu, id, FALSE, (LPMENUITEMINFO)&info);
			selected = IS_FLAG_SET(info.fState, MFS_CHECKED);

			StateMenuItem_SetSelection((UGL_Int)ugl_menuItem, !selected, NULL);
		}
	}

	MenuSelectionCallback(UGL_WIDGET(ugl_menuItem)->callback_target, 0);
	
	return 0;
}
//========================================================================================

void _Dispose(PUGL_Menu ugl_menu, int id, UGL_Error uglError) {
	HMENU hMenu = ugl_menu->hMenu;
	int pos;
	
#if defined(DEBUG) && !defined(_WIN32_WCE)
	if (IsMenu(hMenu) == FALSE) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Removing menu item from previously disposed menu"));
		return;
	}
#endif
	pos = FindMenu(hMenu, id);

	if (pos == -1) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to find menu item"));
		return;
	}

	if (!DeleteMenu(hMenu, pos, MF_BYPOSITION)) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to delete menu item"));
	} else {
		PUGL_Display ugl_display = UGL_WIDGET(ugl_menu)->display;
		PUGL_MenuItem ugl_menuItem = Display_GetMenuItem(ugl_display, id);

		if (ugl_menuItem != NULL) {
			Display_RemoveMenuItem(ugl_menuItem);
		}		
		Menu_Redraw(ugl_menu);
	}
}
//========================================================================================

// this helper updates the enabled state of the left and right softkey
void _Update_Enabled(PUGL_Display display, HWND hwndCB) {
#ifdef _WIN32_WCE
	PUGL_MenuItem tempItem;
	TBBUTTONINFO buttonInfo;
	memset(&buttonInfo, 0, sizeof(TBBUTTONINFO));
	buttonInfo.cbSize = sizeof(TBBUTTONINFO);
	buttonInfo.dwMask = TBIF_STATE;

// left softkey
	tempItem = Display_FindMenuItem(display, hwndCB, 0);
	if(tempItem != NULL){
		buttonInfo.fsState &= ~TBSTATE_ENABLED;
		if (tempItem->enabled){
				buttonInfo.fsState |= TBSTATE_ENABLED;
			}
		SendMessage(tempItem->parent_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo);
		SendMessage(tempItem->parent_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (LPARAM) &buttonInfo);
			
		
	}

// right softkey
	tempItem = Display_FindMenuItem(display, hwndCB, 1);
	if(tempItem != NULL){
		
		buttonInfo.fsState &= ~TBSTATE_ENABLED;
		if (tempItem->enabled){
				buttonInfo.fsState |= TBSTATE_ENABLED;
			}
		SendMessage(tempItem->parent_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (LPARAM) &buttonInfo);
	}
#endif

}
//========================================================================================
void _Dispose_MenuItem(UGL_Int handle, UGL_Error uglError) {
	// new code, softkey

	PUGL_MenuItem ugl_menuItem = UGL_MENUITEM(handle);
	PUGL_Menu ugl_menu = ugl_menuItem->parent_menu;
	int id = ugl_menuItem->id;
	PUGL_Shell parentShell = UGL_SHELL(ugl_menuItem->parent_menu->parent);
	_Shell_RemoveAccelerator(parentShell, ugl_menuItem->id);
	
	
	
	#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	{
	HIMAGELIST imagelist;
	HWND hwndCB = ugl_menu->hwndCB;
	TBBUTTONINFO info; 
	HMENU hMenu;
	UGL_String buffer;
	UGL_Display *display;
	PUGL_MenuItem tempItem, tempItem2, tempItem3, tempItem4;
	BOOL switchCBMB, found, foundfirst, foundfirstsub;
	BOOL isSubMenu[2];
	int i,j;
	FILE * pFile; 
	
	display = UGL_WIDGET(ugl_menuItem)->display;
	memset(&info, 0, sizeof(TBBUTTONINFO));	
	buffer = (UGL_String)calloc(1, sizeof(UGL_String));


	/*
	pFile = fopen ("\\myfile.txt","a"); 
	fprintf(pFile, "display->activeShell = %d, ugl_menu->parent = %d;\n", display->activeShell, ugl_menu->parent); 
	fclose(pFile);
	*/

	if (display->activeShell == ugl_menu->parent) {
		display->activeMenuBar = ugl_menu;
	}


	// if menuitems of menu bar
	if (hwndCB != 0) {
		// if support softkey
		if (Platform_GetMajorVersion() >= 5) {
			int pos = id - 108;
			int index = ugl_menuItem->index;
			int count = Display_MenuItemCount(display, ugl_menu->hwndCB);


			if (pos == -1) {
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to find menu item"));
				return;	
			}

			
			(long) imagelist = SendMessage (hwndCB, TB_GETIMAGELIST, 0, 0);
			if (imagelist != NULL) {
				ImageList_Destroy(imagelist); 
			}

			// if there's sub menus but not normal menuitem

/*			
				pFile = fopen ("\\myfile.txt","a"); 
				fprintf(pFile, "=============== _Dispose_MenuItem 0 ========================\n", count); 
				fprintf(pFile, "menuItemCount = %d\n", count); 
				fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
				for(i=0; i<display->menuItemsLength; i++){						
					if(display->menuItems[i] != NULL){
						// setting the left softkey text
						if(display->menuItems[i]->parent_menu->hwndCB != 0){
							fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
						}
					}
				}
				fclose(pFile);
				fprintf(pFile, "========================================================\n", count); 
				*/
				


			if(Display_FindMenuItem(display, ugl_menu->hwndCB, 0) == NULL && count > 0){
				// directly remove from the submenu
				/*
				pFile = fopen ("\\myfile.txt","a"); 
				fprintf(pFile, "=============== _Dispose_MenuItem 1 ========================\n", count); 
				fclose(pFile);
				*/


				hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
				RemoveMenu ((HMENU) hMenu, index-1, MF_BYPOSITION);
				Display_RemoveMenuItem(ugl_menuItem);
				Display_DecMenuItemIndex(display, ugl_menu->hwndCB, ugl_menuItem->index);

				/*
				pFile = fopen ("\\myfile.txt","a"); 
				fprintf(pFile, "menuItemCount = %d\n", count); 
				fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
				fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
				for(i=0; i<display->menuItemsLength; i++){						
					if(display->menuItems[i] != NULL){
						// setting the left softkey text
						if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
						if(display->menuItems[i]->parent_menu->hwndCB != 0){
							fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
						}
					}
				}
				fprintf(pFile, "========================================================\n", count); 
				fclose(pFile);

				*/

				if(count == 1){
					info.cbSize = sizeof(info); 
					info.dwMask = TBIF_TEXT;
					info.pszText = _T("");
					SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
					SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);	
					if (display->activeShell == ugl_menu->parent) {
						ShowWindow(ugl_menu->hwndCB, SW_SHOW);
						ShowWindow(ugl_menu->hwndMB, SW_HIDE);
					} 

					/*
					pFile = fopen ("\\myfile.txt","a"); 
					fprintf(pFile, "=============== _Dispose_MenuItem 2 ========================\n", count); 
					fprintf(pFile, "menuItemCount = %d\n", count); 
					fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
					fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
					for(i=0; i<display->menuItemsLength; i++){						
						if(display->menuItems[i] != NULL){
							// setting the left softkey text
							if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
							if(display->menuItems[i]->parent_menu->hwndCB != 0){
								fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
							}
						}
					}
					fprintf(pFile, "========================================================\n", count); 
					fclose(pFile);
					*/
				}
			} else {
				switch(count){
					case 1: // item at left softkey
						info.cbSize = sizeof(info); 
						info.dwMask = TBIF_TEXT;
						info.pszText = _T("");
						SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	

						/*
						pFile = fopen ("\\myfile.txt","a"); 
						fprintf(pFile, "=============== _Dispose_MenuItem 3 ========================\n", count); 
						fprintf(pFile, "menuItemCount = %d\n", count); 
						fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
						fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
						for(i=0; i<display->menuItemsLength; i++){						
							if(display->menuItems[i] != NULL){
								// setting the left softkey text
								if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
								if(display->menuItems[i]->parent_menu->hwndCB != 0){
									fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
								}
							}
						}
						fprintf(pFile, "========================================================\n", count); 
						fclose(pFile);
						*/


						if (display->activeShell == ugl_menu->parent) {
							ShowWindow(ugl_menu->hwndCB, SW_SHOW);
							ShowWindow(ugl_menu->hwndMB, SW_HIDE);
						} 

						break;

					case 2:		
						// this is the softkey with Button/Menu (one of the item has submenu)
						tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 1);
						if((tempItem != NULL && tempItem->hasSubMenu)){						
							// if removing button, simply remove it.						
							if(index == 0){
								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 4 ========================\n", count); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/


								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								} 

								break;



							} else {
								// else remove menu and swtich back to Button/Button menu
								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 5 ========================\n", count); 
								fclose(pFile);
								*/

								hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
								RemoveMenu ((HMENU) hMenu, index-1, MF_BYPOSITION);

								tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 0);
										
								if(tempItem != NULL && tempItem->text != NULL){								
									info.cbSize = sizeof(TBBUTTONINFO);
									info.dwMask = TBIF_TEXT;
									info.pszText = (LPWSTR) tempItem->text;
									if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
									}
								}

								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);
								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_SHOW);
									ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								} 


								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								break;


							}
				
						} else {
							// softkey button/button case						
							// find the item on the rightsoftkey and set its text to the left														
							if(index == 0){
								if(tempItem != NULL && tempItem->text != NULL){								
										info.cbSize = sizeof(TBBUTTONINFO);
										info.dwMask = TBIF_TEXT;
										info.pszText = (LPWSTR) tempItem->text;
										tempItem->index = 0;
										//Set right button to left button
										if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
											Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
										}
									}
								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								//Set empty menuitem to right button
								SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);


								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 6 ========================\n", count); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[index] == NULL) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/
								//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_SHOW);
									ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								} 

								break;

							} else {
								// else simply set the right softkey to ""					
								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);
								//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_SHOW);
									ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								} 

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 7 ========================\n", count); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								break;
								*/

							}
						}
						break;

					case 3:
						/*					<SUB>
											<SUB>
							<NULL>			<SUB>
						*/
						hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 0);
						if (tempItem == NULL) { // remove SUB directly
							/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 8 ========================\n", count); 
							fclose(pFile);
							*/

							RemoveMenu (hMenu, ugl_menuItem->index-1, MF_BYPOSITION);
							Display_DecMenuItemIndex(display, ugl_menu->hwndCB, index);		

							/*	
							pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if(display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

							if (display->activeShell == ugl_menu->parent) {
								ShowWindow(ugl_menu->hwndCB, SW_HIDE);
								ShowWindow(ugl_menu->hwndMB, SW_SHOW);
							} 
							break;


						} else if (index == 0) {
							/* check if the other two menuitems has submenus */
							/* case 1:
												<SUB>
								<item1>			<SUB>
								if <item1> is to be deleted, set the left softkey to ""
							*/
							/* case 2:
												<SUB>
								<item1>			<item2>
							
							    or
												<item2>
								<item1>			<SUB>
								if <item1> is to be deleted, move <item2> to left and change to <button/button>
							*/
							/* case 3:
												<item3>
								<item1>			<item2>
								if <item1> is to be deleted, move <item2> to left and change to <button/button>
							*/
							found = FALSE;
							foundfirst = FALSE;
							tempItem = tempItem2 = tempItem3 = NULL;
							for(i=0; i<display->menuItemsLength; i++){
								if(display->menuItems[i] != NULL){
									if(display->menuItems[i]->parent_menu->hwndCB != 0) {
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if (display->menuItems[i]->index == index) continue;
										if (!display->menuItems[i]->hasSubMenu) {
											found = TRUE; //found item2, doesn't have submenu
											if(!foundfirst) {
												foundfirst = TRUE;
												tempItem = display->menuItems[i]; // this is item2
											} else {
												tempItem3 = display->menuItems[i]; // this is item3
											}
										} else {
											tempItem2 = display->menuItems[i]; // this is SUB
										}

									}
								}
							}
							if (!found) { // case 1
								info.cbSize = sizeof(TBBUTTONINFO);
								info.dwMask = TBIF_TEXT;
								info.pszText = (LPWSTR) _T("");
								if(!SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
									Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 9 ========================\n", count); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								} 

								break;


							} else if (tempItem2 != NULL) { //case 2
							/*					<SUB>
								<item1>			<item2>

								or				
												<item2>
								<item1>			<SUB>

								if the <item1> is to be deleted, need to move the <item2> to left softkey
							*/
								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 10 ========================\n", count); 
								fclose(pFile);
								*/
								
								RemoveMenu ((HMENU) hMenu, tempItem->index-1, MF_BYPOSITION);
								Display_DecMenuItemIndex(display, ugl_menu->hwndCB, tempItem->index);
								tempItem->index = 0;
								info.cbSize = sizeof(TBBUTTONINFO);
								info.dwMask = TBIF_TEXT;
								info.pszText = (LPWSTR) tempItem->text;
								if(!SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								} 


								break;


							} else { // case 3
							/*					<item3>
								<item1>			<item2>
								if <item1> is to be deleted, move <item2> to left and change to <button/button>
							*/

							/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 11 ========================\n", count); 
							fclose(pFile);
							*/

								hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
								Display_DecMenuItemIndex(display, ugl_menu->hwndCB, tempItem->index);
								tempItem->index = 0;
								info.cbSize = sizeof(TBBUTTONINFO);
								info.dwMask = TBIF_TEXT;
								info.pszText = (LPWSTR) tempItem->text;
								if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}
								info.pszText = (LPWSTR) tempItem3->text;
								if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}
								RemoveMenu ((HMENU) hMenu, 1, MF_BYPOSITION);
								RemoveMenu ((HMENU) hMenu, 0, MF_BYPOSITION);

								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
								SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);	

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_SHOW);
									ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								} 


								//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								break;



							}
						} else { //if item1 is not to be deleted
							/* case 1:
												<SUB2>
								<item1>			<SUB1>
								
							*/
							/* case 2:
												<SUB>
								<item1>			<item2>
							
							    or
												<item2>
								<item1>			<SUB>
								
							*/
							/* case 3:
												<item3>
								<item1>			<item2>
								
							*/

							/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 12 ========================\n", count); 
							fclose(pFile);
							*/

							found = FALSE;
							foundfirst = FALSE;
							foundfirstsub = FALSE;
							tempItem = tempItem2 = tempItem3 = NULL;
							for(i=0; i<display->menuItemsLength; i++){
								if(display->menuItems[i] != NULL){
									if(display->menuItems[i]->parent_menu->hwndCB != 0) {
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if (display->menuItems[i]->index == index) continue;
										if (display->menuItems[i]->index == 0) continue; // keep index 0 as left softkey regardlessly
										tempItem = display->menuItems[i]; // the remaining one, might be item or SUB. 
									}
								}
							}

							if (tempItem->hasSubMenu) { //the remaining one is SUB
								RemoveMenu (hMenu, ugl_menuItem->index-1, MF_BYPOSITION);
								Display_DecMenuItemIndex(display, ugl_menu->hwndCB, index);		

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								} 

								break;


							} else { // the remaining one is <item>, need to change to <button/button>
								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "=============== _Dispose_MenuItem 13 ========================\n", count); 
								fclose(pFile);
								*/

								tempItem2 = Display_FindMenuItem(display, ugl_menu->hwndCB, 0);
								Display_DecMenuItemIndex(display, ugl_menu->hwndCB, index);
								info.cbSize = sizeof(TBBUTTONINFO);
								info.dwMask = TBIF_TEXT;
								info.pszText = (LPWSTR) tempItem2->text;
								if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}
								info.pszText = (LPWSTR) tempItem->text;
								if(!SendMessage (ugl_menu->hwndCB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
								}
								RemoveMenu ((HMENU) hMenu, 1, MF_BYPOSITION);
								RemoveMenu ((HMENU) hMenu, 0, MF_BYPOSITION);
								info.cbSize = sizeof(info);
								info.dwMask = TBIF_TEXT;
								info.pszText = _T("");
								SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info);	
								SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY1, (long) &info);	

								//ShowWindow(ugl_menu->hwndCB, SW_SHOW);
								//ShowWindow(ugl_menu->hwndMB, SW_HIDE);

								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_SHOW);
									ShowWindow(ugl_menu->hwndMB, SW_HIDE);
								} 


								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								break;

							}
							
						}

						break;

					default:
						//MessageBox(0, ugl_menuItem->text, _T("new"), 0);
						found = FALSE;
						hMenu = (HMENU) SendMessage (ugl_menu->hwndMB, SHCMBM_GETSUBMENU, 0, ID_SPSOFTKEY1);
						
						//if removing items in a menu, just remove it.
						if (index != 0){
							/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 14 ========================\n", count); 
							fclose(pFile);
							*/

							RemoveMenu (hMenu, ugl_menuItem->index-1, MF_BYPOSITION);
							Display_DecMenuItemIndex(display, ugl_menu->hwndCB, index);

							/*	
							pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								break;


						} else { //index == 0, it's the first item

							// removing the first item, check if the next item has submenu
							tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 1);
							/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 15 ========================\n", count); 
							fclose(pFile);
							*/


							if(tempItem != NULL && tempItem->hasSubMenu){
								for(i = 2; i < display->menuItemsLength; i++){
									tempItem2 = Display_FindMenuItem(display, ugl_menu->hwndCB, i);
									if(tempItem2!= NULL && !tempItem2->hasSubMenu){
										// move the first item that doesn't have submenu into the first spot
										RemoveMenu (hMenu, tempItem2->index-1, MF_BYPOSITION);
										Display_DecMenuItemIndex(display, ugl_menu->hwndCB, tempItem2->index);
										tempItem2->index = 0;
										info.cbSize = sizeof(TBBUTTONINFO);
										info.dwMask = TBIF_TEXT;
										info.pszText = (LPWSTR) tempItem2->text;
										if(!SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
											Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
										}
										found = TRUE;
										break;
									}
								}
								if(!found){
									info.cbSize = sizeof(TBBUTTONINFO);
									info.dwMask = TBIF_TEXT;
									info.pszText = (LPWSTR) _T("");
									if(!SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
									}
								}

								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								break;


							} else { //the first menuitem on the right softkey has no submenu
								// else remove text on the left softkey and pop the first menuitem's text onto the left softkey
								/*
							pFile = fopen ("\\myfile.txt","a"); 
							fprintf(pFile, "=============== _Dispose_MenuItem 16 ========================\n", count); 
							fclose(pFile);
							*/

								tempItem = Display_FindMenuItem(display, ugl_menu->hwndCB, 1);								
								RemoveMenu (hMenu, 0, MF_BYPOSITION);
								Display_DecMenuItemIndex(display, ugl_menu->hwndCB, 1);
								tempItem->index = 0;
								
								if(tempItem != NULL && tempItem->text != NULL){								
									info.cbSize = sizeof(TBBUTTONINFO);
									info.dwMask = TBIF_TEXT;
									info.pszText = (LPWSTR) tempItem->text;
									if(!SendMessage (ugl_menu->hwndMB, TB_SETBUTTONINFO, ID_SPSOFTKEY0, (long) &info)){
										Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to set buttoninfo"));
									}
								}


								/*
								pFile = fopen ("\\myfile.txt","a"); 
								fprintf(pFile, "menuItemCount = %d\n", count); 
								fprintf(pFile, "ugl_menuItem->index = %d, id = %d \n", ugl_menuItem->index, ugl_menuItem->id); 
								fprintf(pFile, "ugl_menu->hwndMB = %d, ugl_menu->hwndCB = %d \n", ugl_menu->hwndMB, ugl_menu->hwndCB);
								for(i=0; i<display->menuItemsLength; i++){						
									if(display->menuItems[i] != NULL){
										// setting the left softkey text
										if (display->menuItems[i]->parent_menu->hwndCB != ugl_menu->hwndCB) continue;
										if(display->menuItems[i]->parent_menu->hwndCB != 0){
											fprintf(pFile, "menuItem[%d] index=%d, id=%d\n", i, display->menuItems[i]->index, display->menuItems[i]->id); 
										}
									}
								}
								
								fprintf(pFile, "========================================================\n", count); 
								fclose(pFile);
								*/

								
							}
								if (display->activeShell == ugl_menu->parent) {
									ShowWindow(ugl_menu->hwndCB, SW_HIDE);
									ShowWindow(ugl_menu->hwndMB, SW_SHOW);
								} 

								break;


							
						}
				
					break;

 		}
				Display_RemoveMenuItem(ugl_menuItem);
				//_Update_Enabled(display);
				free(buffer);
			
			}	
		} else {
			// not supporting softkey
			int pos = SendMessage(hwndCB, TB_COMMANDTOINDEX, id, 0);		
			
			if (pos == -1) {
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to find menu item"));
				return;	
			}
			if (!SendMessage(hwndCB, TB_DELETEBUTTON, pos, 0)) {
				Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to delete menu item"));
				return;
			}
			Display_RemoveMenuItem(ugl_menuItem);
			(long) imagelist = SendMessage (hwndCB, TB_GETIMAGELIST, 0, 0);
			if (imagelist != NULL) {
				ImageList_Destroy(imagelist); 
			}

		}
	// if sub menu
	} else {
		_Dispose(ugl_menu, id, uglError);
	}
	}

#else
	_Dispose(ugl_menu, id, uglError);
#endif


	/*  old code
#if defined(WIN32_PLATFORM_PSPC) || defined(WIN32_PLATFORM_WFSP)
	{
	HIMAGELIST imagelist;
	HWND hwndCB = ugl_menu->hwndCB;

	if (hwndCB != 0) {
		int pos = SendMessage(hwndCB, TB_COMMANDTOINDEX, id, 0);		
		
		if (pos == -1) {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to find menu item"));
			return;	
		}
		if (!SendMessage(hwndCB, TB_DELETEBUTTON, pos, 0)) {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unable to delete menu item"));
			return;
		}
		Display_RemoveMenuItem(ugl_menuItem);
		(long) imagelist = SendMessage (hwndCB, TB_GETIMAGELIST, 0, 0);
		if (imagelist != NULL) {
			ImageList_Destroy(imagelist); 
		}
	} else {
		_Dispose(ugl_menu, id, uglError);
	}
	}
#else
	_Dispose(ugl_menu, id, uglError);
#endif
	*/


}
//========================================================================================
