/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

package org.eclipse.swt.widgets;


import org.eclipse.swt.*;
import org.eclipse.swt.events.*;
import org.eclipse.swt.graphics.Image;

import com.ibm.ugl.eswt.OS;

/**
 * Instances of this class represent a selectable user interface object
 * that issues notification when pressed and released. 
 * <dl>
 * <dt><b>Styles:</b></dt>
 * <dd>CHECK, CASCADE, PUSH, RADIO, SEPARATOR</dd>
 * <dt><b>Events:</b></dt>
 * <dd>Selection</dd>
 * </dl>
 * <p>
 * Note: Only one of the styles CHECK, CASCADE, PUSH, RADIO and SEPARATOR
 * may be specified.
 * </p><p>
 * IMPORTANT: This class is <em>not</em> intended to be subclassed.
 * </p>
 */
public class MenuItem extends Item {
	Menu parent, menu;
	int accelerator;
	boolean enabled=true;
	boolean selected=false;
	boolean disableAutoSelection=false;
	
/**
 * Constructs a new instance of this class given its parent
 * (which must be a <code>Menu</code>) and a style value
 * describing its behavior and appearance. The item is added
 * to the end of the items maintained by its parent.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>SWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>SWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p>
 *
 * @param parent a menu control which will be the parent of the new instance (cannot be null)
 * @param style the style of control to construct
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 *
 * @see SWT#CHECK
 * @see SWT#CASCADE
 * @see SWT#PUSH
 * @see SWT#RADIO
 * @see SWT#SEPARATOR
 * @see Widget#checkSubclass
 * @see Widget#getStyle
 */
public MenuItem(Menu parent, int style) {
	super (parent, checkStyle (style));
	this.parent = parent;
	createWidget(parent.getItemCount());
}
/**
 * Constructs a new instance of this class given its parent
 * (which must be a <code>Menu</code>), a style value
 * describing its behavior and appearance, and the index
 * at which to place it in the items maintained by its parent.
 * <p>
 * The style value is either one of the style constants defined in
 * class <code>SWT</code> which is applicable to instances of this
 * class, or must be built by <em>bitwise OR</em>'ing together 
 * (that is, using the <code>int</code> "|" operator) two or more
 * of those <code>SWT</code> style constants. The class description
 * lists the style constants that are applicable to the class.
 * Style bits are also inherited from superclasses.
 * </p>
 *
 * @param parent a menu control which will be the parent of the new instance (cannot be null)
 * @param style the style of control to construct
 * @param index the index to store the receiver in its parent
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the parent is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the parent</li>
 *    <li>ERROR_INVALID_SUBCLASS - if this class is not an allowed subclass</li>
 * </ul>
 *
 * @see SWT#CHECK
 * @see SWT#CASCADE
 * @see SWT#PUSH
 * @see SWT#RADIO
 * @see SWT#SEPARATOR
 * @see Widget#checkSubclass
 * @see Widget#getStyle
 */
public MenuItem(Menu parent, int style, int index) {
	super(parent, checkStyle(style));
	this.parent = parent;
	createWidget(index);
}

void selectionCallback(int modifiers) {
	// Ignore all events that come in when the item (or parent) is disabled
	// See bug 106888
	if (isEnabled() == false) return; 
	
	if ((internal_style & SWT.CHECK) == SWT.CHECK) {
		if (disableAutoSelection == false) {
			selected = !selected;
		}
	} else if ((internal_style & SWT.RADIO) == SWT.RADIO) {
		if ((parent.getStyle() & SWT.NO_RADIO_GROUP) == SWT.NO_RADIO_GROUP) {
			selected = !selected;
		} else {
			selectRadio();
		}
	}
	Event event = new Event();
	setInputState(event, SWT.Selection, modifiers);
	postEvent(SWT.Selection, event);
}

/**
 * Adds the listener to the collection of listeners who will
 * be notified when the menu item is selected, by sending
 * it one of the messages defined in the <code>SelectionListener</code>
 * interface.
 * <p>
 * When <code>widgetSelected</code> is called, the stateMask field of the event object is valid.
 * <code>widgetDefaultSelected</code> is not called.
 * </p>
 *
 * @param listener the listener which should be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see SelectionListener
 * @see #removeSelectionListener
 * @see SelectionEvent
 */
public void addSelectionListener(SelectionListener listener) {
	checkWidget();
	if (listener == null) error(SWT.ERROR_NULL_ARGUMENT);
	TypedListener typedListener = new TypedListener(listener);
	addListener(SWT.Selection, typedListener);
	addListener(SWT.DefaultSelection, typedListener);
}

static int checkStyle(int style) {
	return checkBits(style, SWT.PUSH, SWT.CHECK, SWT.RADIO, SWT.SEPARATOR, SWT.CASCADE, 0);
}

protected void internal_createHandle(int index) {
	int count = parent.getItemCount();
	if (!(0 <= index && index <= count)) error(SWT.ERROR_INVALID_RANGE);
	
	int parentHandle = parent.internal_handle;
	if ((internal_style & SWT.SEPARATOR) == SWT.SEPARATOR) {
		internal_handle = OS.SeparatorItem_New(parentHandle, index);
	} else 
	if ((internal_style & SWT.RADIO) == SWT.RADIO) {
		internal_handle = OS.StateMenuItem_New(parentHandle,index, OS.STYLE_RADIO);
	} else
	if ((internal_style & SWT.CHECK) == SWT.CHECK) {
		internal_handle = OS.StateMenuItem_New(parentHandle,index, OS.STYLE_CHECK);
	} else
	if ((internal_style & SWT.PUSH) == SWT.PUSH) {
		internal_handle = OS.PushMenuItem_New(parentHandle,index);
	} else		
	if ((internal_style & SWT.CASCADE) == SWT.CASCADE) {
		internal_handle = OS.CascadeMenuItem_New(parentHandle,index);
	}
	if (internal_handle == 0) error(SWT.ERROR_NO_HANDLES);
	parent.insertItem(this, index);
}

public void dispose() {
	Menu parent = this.parent;
	super.dispose();
	if (parent != null) parent.removeItem(this);
}
/**
 * Returns the widget accelerator.  An accelerator is the bit-wise
 * OR of zero or more modifier masks and a key. Examples:
 * <code>SWT.CONTROL | SWT.SHIFT | 'T', SWT.ALT | SWT.F2</code>.
 * The default value is zero, indicating that the menu item does
 * not have an accelerator.
 *
 * @return the accelerator or 0
 *
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public int getAccelerator() {
	checkWidget();
	return accelerator;
}

/**
 * Returns <code>true</code> if the receiver is enabled, and
 * <code>false</code> otherwise. A disabled menu item is typically
 * not selectable from the user interface and draws with an
 * inactive or "grayed" look.
 *
 * @return the receiver's enabled state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @see #isEnabled
 */
public boolean getEnabled() {
	checkWidget();
	return enabled;
}

/**
 * Returns the receiver's cascade menu if it has one or null
 * if it does not. Only <code>CASCADE</code> menu items can have
 * a pull down menu. The sequence of key strokes, button presses 
 * and/or button releases that are used to request a pull down
 * menu is platform specific.
 *
 * @return the receiver's menu
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public Menu getMenu() {
	checkWidget();
	return menu;
}

/**
 * Returns the receiver's parent, which must be a <code>Menu</code>.
 *
 * @return the receiver's parent
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public Menu getParent() {
	checkWidget();
	return parent;
}

/**
 * Returns <code>true</code> if the receiver is selected,
 * and false otherwise.
 * <p>
 * When the receiver is of type <code>CHECK</code> or <code>RADIO</code>,
 * it is selected when it is checked.
 *
 * @return the selection state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public boolean getSelection() {
	checkWidget();
	if ((internal_style & (SWT.CHECK | SWT.RADIO)) == 0) return false;
	return selected;
}

public String getText() {
	checkWidget();
	return text;
}
/**
 * Returns <code>true</code> if the receiver is enabled and all
 * of the receiver's ancestors are enabled, and <code>false</code>
 * otherwise. A disabled menu item is typically not selectable from the
 * user interface and draws with an inactive or "grayed" look.
 *
 * @return the receiver's enabled state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @see #getEnabled
 */
public boolean isEnabled() {
	checkWidget();
	return getEnabled () && parent.isEnabled ();
}

void releaseWidget() {
	if (menu != null && !menu.isDisposed()) menu.releaseResources();
	menu = null;
	super.releaseWidget ();
	accelerator = 0;
	parent = null;
}

/**
 * Removes the listener from the collection of listeners who will
 * be notified when the control is selected.
 *
 * @param listener the listener which should no longer be notified
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the listener is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 *
 * @see SelectionListener
 * @see #addSelectionListener
 */
public void removeSelectionListener(SelectionListener listener) {
	checkWidget();
	if (listener == null) error(SWT.ERROR_NULL_ARGUMENT);
	if (eventTable == null) return;
	eventTable.unhook(SWT.Selection, listener);
	eventTable.unhook(SWT.DefaultSelection,listener);
}
/**
 * Sets the widget accelerator.  An accelerator is the bit-wise
 * OR of zero or more modifier masks and a key. Examples:
 * <code>SWT.MOD1 | SWT.MOD2 | 'T', SWT.MOD3 | SWT.F2</code>.
 * <code>SWT.CONTROL | SWT.SHIFT | 'T', SWT.ALT | SWT.F2</code>.
 * The default value is zero, indicating that the menu item does
 * not have an accelerator.
 *
 * @param accelerator an integer that is the bit-wise OR of masks and a key
 *
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setAccelerator(int accelerator) {
	checkWidget();
	this.accelerator = accelerator;
	if ((internal_style & SWT.SEPARATOR) == 0) {
		int keyMods = 0;
		if ((accelerator & SWT.MOD3) != 0) keyMods |= OS.MODIFIER_MASK_ALT;
		if ((accelerator & SWT.MOD2) != 0) keyMods |= OS.MODIFIER_MASK_SHIFT;
		if ((accelerator & SWT.MOD1) != 0) keyMods |= OS.MODIFIER_MASK_CONTROL;
		int key = accelerator & ~(SWT.MOD3 | SWT.MOD2 | SWT.MOD1);
        key = Display.convertToUglKeyCode(key);
	    int value = (keyMods << 16) | key;
	    OS.LabeledMenuItem_SetAccelerator(internal_handle,value);
	}
}

/**
 * Sets the image the receiver will display to the argument.
 * <p>
 * Note: This operation is a hint and is not supported on
 * platforms that do not have this concept (for example, Windows NT).
 * </p>
 *
 * @param image the image to display
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setImage(Image image){
		super.setImage(image);	
		if ((internal_style & SWT.SEPARATOR) == 0) {
		    int imageHandle = 0;
		    //Avoiding to throw NULL image object to native. Which will cause j9 dump
		    if (image != null) {
		        imageHandle = image.internal_handle;
		    	OS.LabeledMenuItem_SetImage(internal_handle, imageHandle);
		    }
		
		}
}


/**
 * Enables the receiver if the argument is <code>true</code>,
 * and disables it otherwise. A disabled menu item is typically
 * not selectable from the user interface and draws with an
 * inactive or "grayed" look.
 *
 * @param enabled the new enabled state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setEnabled(boolean enabled) {
	checkWidget();
	
	if (enabled == this.enabled) return;
	
	this.enabled = enabled;
	if (parent.getEnabled()) {
		updateEnabledStatus(true);
	} else {
		// The native widgets enabled status does not
		// need to change now, because the parent is disabled
		// and the correct enabled status will be reflected
		// when the parent is enabled again
	}
}

/**
 * Sets the receiver's pull down menu to the argument.
 * Only <code>CASCADE</code> menu items can have a
 * pull down menu. The sequence of key strokes, button presses
 * and/or button releases that are used to request a pull down
 * menu is platform specific.
 *
 * @param menu the new pull down menu
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_MENU_NOT_DROP_DOWN - if the menu is not a drop down menu</li>
 *    <li>ERROR_MENUITEM_NOT_CASCADE - if the menu item is not a <code>CASCADE</code></li>
 *    <li>ERROR_INVALID_ARGUMENT - if the menu has been disposed</li>
 *    <li>ERROR_INVALID_PARENT - if the menu is not in the same widget tree</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setMenu(Menu menu) {
	checkWidget();
	if ((internal_style & SWT.CASCADE) == 0) error(SWT.ERROR_MENUITEM_NOT_CASCADE);
	
	int menuHandle;
	if (menu != null) {
		if (menu.isDisposed()) error(SWT.ERROR_INVALID_ARGUMENT);
		if ((menu.internal_style & SWT.DROP_DOWN) == 0) error(SWT.ERROR_MENU_NOT_DROP_DOWN);
		if (menu.parent != parent.parent) error(SWT.ERROR_INVALID_PARENT);
		menuHandle = menu.internal_handle;
		this.menu = menu;
		this.menu.cascade = this;
	} else {
		menuHandle = 0;
		
		// this.menu can be null if the menu has never been set before
		if (this.menu != null) {
			this.menu.cascade = null; // reset the previous one
			this.menu = null;
		}
	}
	OS.CascadeMenuItem_SetMenu(internal_handle, menuHandle);
}

void selectRadio () {
	int index = 0;
	MenuItem[] items = parent.getItems();
	while (index < items.length && items[index] != this) index++;
	int i = index - 1;
	while (i >= 0 && items[i].setRadioSelection(false)) --i;
	int j = index + 1;
	while (j < items.length && items[j].setRadioSelection (false)) j++;
	setSelection(true);
}
boolean setRadioSelection(boolean value) {
	if ((internal_style & SWT.RADIO) == 0) return false;
	if (getSelection() != value) {
		setSelection(value);
		postEvent(SWT.Selection);
	}
	return true;
}

/**
 * Sets the selection state of the receiver.
 * <p>
 * When the receiver is of type <code>CHECK</code> or <code>RADIO</code>,
 * it is selected when it is checked.
 *
 * @param selected the new selection state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setSelection(boolean selected) {
	checkWidget();
	if ((internal_style & (SWT.CHECK | SWT.RADIO)) == 0) return;
	this.selected = selected; 
	OS.StateMenuItem_SetSelection(internal_handle, selected);
}

/**
 * Sets the disableAutoSelection state of the receiver.
 * <p>
 * When the receiver is of type <code>CHECK</code>,
 * it is selected when it is checked.
 *
 * @param disableAutoSelection the new disableAutoSelection state
 *
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 */
public void setDisableAutoSelection(boolean disableAutoSelection) {
	checkWidget();
	if ((internal_style & (SWT.CHECK)) == 0) return;
	this.disableAutoSelection = disableAutoSelection;
	OS.StateMenuItem_DisableAutoSelection(internal_handle, disableAutoSelection);
}

/**
 * Sets the receiver's text. The string may include
 * the mnemonic character and accelerator text.
 * <p>
 * Mnemonics are indicated by an '&amp' that causes the next
 * character to be the mnemonic.  When the user presses a
 * key sequence that matches the mnemonic, a selection
 * event occurs. On most platforms, the mnemonic appears
 * underlined but may be emphasised in a platform specific
 * manner.  The mnemonic indicator character '&amp' can be
 * escaped by doubling it in the string, causing a single
 *'&amp' to be displayed.
 * </p>
 * <p>
 * Accelerator text is indicated by the '\t' character.
 * On platforms that support accelerator text, the text
 * that follows the '\t' character is displayed to the user,
 * typically indicating the key stroke that will cause
 * the item to become selected.  On most platforms, the
 * accelerator text appears right aligned in the menu.
 * Setting the accelerator text does not install the
 * accelerator key sequence. The accelerator key sequence
 * is installed using #setAccelerator.
 * </p>
 * 
 * @param string the new text
 *
 * @exception IllegalArgumentException <ul>
 *    <li>ERROR_NULL_ARGUMENT - if the text is null</li>
 * </ul>
 * @exception SWTException <ul>
 *    <li>ERROR_WIDGET_DISPOSED - if the receiver has been disposed</li>
 *    <li>ERROR_THREAD_INVALID_ACCESS - if not called from the thread that created the receiver</li>
 * </ul>
 * 
 * @see #setAccelerator
 */
public void setText(String string) {
	super.setText(string);
	if ((internal_style & SWT.SEPARATOR) == 0) {
		String itemText = string;
		String accelText = "";
		
		int tab = string.indexOf('\t');
		if (tab > -1) {
			
			if (tab == 0) {
				itemText = "";
			}
			else {
				itemText = string.substring(0,tab);
			}
			
			int length = string.length();
			if (tab < (length-1)) {
				accelText = string.substring(tab+1,length);
			}
		}
		
		OS.LabeledMenuItem_SetText(internal_handle, itemText, accelText);
	}
}

void updateEnabledStatus(boolean parentEnabled) {
	if (internal_handle != 0) {
		OS.AbstractMenuItem_SetEnabled(internal_handle, parentEnabled && enabled);
	}
}

void fixMenus(Decorations newParent) {
	if (menu != null) menu.fixMenus(newParent);
}

}
