/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Additions/modifications to this source file by Oracle America, Inc. 2011
 *******************************************************************************/

#include "UGL_ErrorImpl.h"
#include "UGL_Error.h"
#include "TypeConverter.h"
#include <jni.h>
#include "Image.h"
#include "JNIHelpers.h"
#include "UGLTypes.h"

#ifdef __cplusplus
#define UNUSED(var) 
#else
#define UNUSED(var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Frees the platform resources associated with the image identified by
 * <code>handle</code>.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1Dispose(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_Dispose\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	Image_Dispose(handle, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Image_Dispose", &error);
	}
}

/**
 * Returns the position and size of the image identified by <code>handle</code>.
 */
JNIEXPORT jintArray JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1GetBounds(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Int returnHeight;
	UGL_Int returnWidth;
	UGL_Int returnX;
	UGL_Int returnY;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_GetBounds\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	Image_GetBounds(handle, &returnX, &returnY, &returnWidth, &returnHeight, &error);
	if (error.error_code == 0) {
		return createJavaRect(env, returnX, returnY, returnWidth, returnHeight);
	} else {
		UGL_ErrorHandler(env, "Image_GetBounds", &error);
		return 0;
	}
}

/**
 * Returns the color depth of the image identified by <code>handle</code>.
 * The supported color depths are 1, 2, 4, 8, 16, 24 and 32 bit. The
 * returned depth must be one of the values supported by UGL. In the case
 * of 3, 5, 6 or 7 bit GIFs the returned depth should reflect the actual
 * "color resolution" as defined in the GIF89a spec.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1GetDepth(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_GetDepth\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	returnValue = Image_GetDepth(handle, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_GetDepth", &error);
		return 0;
	}
}

/**
 * Creates a new image from the supplied pixel data. Each pixel is an
 * index into a color palette. The color palette is supplied in the
 * indexed color arrays. The palette may not use all possible color
 * values . In other words the number of elements may be less than
 * (2^depth). Supported color depths are 1, 2, 4 and 8 bits per pixel.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewFromByteData(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jint width, jint height, jbyteArray imageData, jint bytesPerLine, jint depth, jbyteArray indexedReds, jbyteArray indexedGreens, jbyteArray indexedBlues, jint paletteSize, jint transparentPixel)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	UGL_ByteArray uglimageData;
	UGL_ByteArray uglindexedBlues;
	UGL_ByteArray uglindexedGreens;
	UGL_ByteArray uglindexedReds;
	
	uglimageData = GetByteArrayElements(env, imageData);
	if (ExceptionCheck(env)) return 0;
	uglindexedReds = GetByteArrayElements(env, indexedReds);
	if (ExceptionCheck(env)) return 0;
	uglindexedGreens = GetByteArrayElements(env, indexedGreens);
	if (ExceptionCheck(env)) return 0;
	uglindexedBlues = GetByteArrayElements(env, indexedBlues);
	if (ExceptionCheck(env)) return 0;
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromByteData\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APIARGTRACE(env, "\twidth=%d\n" , width);
	APIARGTRACE(env, "\theight=%d\n" , height);
	APIARGTRACE(env, "\tuglimageData (ByteArray)=%p\n" , uglimageData);
	APIARGTRACE(env, "\tbytesPerLine=%d\n" , bytesPerLine);
	APIARGTRACE(env, "\tdepth=%d\n" , depth);
	APIARGTRACE(env, "\tuglindexedReds (ByteArray)=%p\n" , uglindexedReds);
	APIARGTRACE(env, "\tuglindexedGreens (ByteArray)=%p\n" , uglindexedGreens);
	APIARGTRACE(env, "\tuglindexedBlues (ByteArray)=%p\n" , uglindexedBlues);
	APIARGTRACE(env, "\tpaletteSize=%d\n" , paletteSize);
	APIARGTRACE(env, "\ttransparentPixel=%d\n" , transparentPixel);
	#endif

	returnValue = Image_NewFromByteData(deviceHandle, width, height, uglimageData, bytesPerLine, depth, uglindexedReds, uglindexedGreens, uglindexedBlues, paletteSize, transparentPixel, &error);
	ReleaseByteArrayElements(env, imageData, uglimageData);
	ReleaseByteArrayElements(env, indexedReds, uglindexedReds);
	ReleaseByteArrayElements(env, indexedGreens, uglindexedGreens);
	ReleaseByteArrayElements(env, indexedBlues, uglindexedBlues);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromByteData", &error);
		return 0;
	}
}

/**
 * Creates a new image using the supplied raw data.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewFromFileData(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jbyteArray data, jint length)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	UGL_ByteArray ugldata;
	
	ugldata = GetByteArrayElements(env, data);
	if (ExceptionCheck(env)) return 0;
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromFileData\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APIARGTRACE(env, "\tugldata (ByteArray)=%p\n" , ugldata);
	APIARGTRACE(env, "\tlength=%d\n" , length);
	#endif

	returnValue = Image_NewFromFileData(deviceHandle, ugldata, length, &error);
	ReleaseByteArrayElements(env, data, ugldata);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromFileData", &error);
		return 0;
	}
}

/**
 * Creates a new image from the supplied pixel data. Each pixel is a direct
 * color RGB value. Supported color depths are 16, 24 and 32 bits per pixel.
 * For 16 bit depth the RGB bits are layed out as 5-6-5 bits for red, green
 * and blue respectively, 24 bit depth uses 8-8-8 bits and no alpha channel.
 * 32 bit color depth uses 8-8-8 bits and the most significant byte for the
 * alpha channel for an ARGB pixel layout. For 32 bit pixel data the alpha
 * channel byte is ignored if either one of the
 * <code>transparentColor</code>, <code>globalAlpha</code> or
 * <code>alphaData</code> is specified. Only one of the three transparency/
 * alpha arguments may be specified at a time.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewFromIntData(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jint width, jint height, jintArray imageData, jint depth, jint transparentColor, jint globalAlpha, jbyteArray alphaData)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	UGL_ByteArray uglalphaData;
	UGL_IntArray uglimageData;
	
	uglimageData = GetIntArrayElements(env, imageData);
	if (ExceptionCheck(env)) return 0;
	uglalphaData = GetByteArrayElements(env, alphaData);
	if (ExceptionCheck(env)) return 0;
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromIntData\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APIARGTRACE(env, "\twidth=%d\n" , width);
	APIARGTRACE(env, "\theight=%d\n" , height);
	APIARGTRACE(env, "\tuglimageData (IntArray)=%p\n" , uglimageData);
	APIARGTRACE(env, "\tdepth=%d\n" , depth);
	APIARGTRACE(env, "\ttransparentColor=%d\n" , transparentColor);
	APIARGTRACE(env, "\tglobalAlpha=%d\n" , globalAlpha);
	APIARGTRACE(env, "\tuglalphaData (ByteArray)=%p\n" , uglalphaData);
	#endif

	returnValue = Image_NewFromIntData(deviceHandle, width, height, uglimageData, depth, transparentColor, globalAlpha, uglalphaData, &error);
	ReleaseIntArrayElements(env, imageData, uglimageData);
	ReleaseByteArrayElements(env, alphaData, uglalphaData);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromIntData", &error);
		return 0;
	}
}

/**
 * Creates a new, empty image with the given dimensions.
 * <p>
 * Image will be filled with the provided fill color which
 * is in 0xAARRGGBB format.
 * </p>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewFromSize(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jint width, jint height, jint fillColor)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromSize\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APIARGTRACE(env, "\twidth=%d\n" , width);
	APIARGTRACE(env, "\theight=%d\n" , height);
	APIARGTRACE(env, "\tfillColor=%d\n" , fillColor);
	#endif

	returnValue = Image_NewFromSize(deviceHandle, width, height, fillColor, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromSize", &error);
		return 0;
	}
}

JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewFromSystemResource(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jint identifier)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromSystemResource\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APIARGTRACE(env, "\tidentifier=%d\n" , height);
	APIARGTRACE(env, "\tfillColor=%d\n" , fillColor);
	#endif

	returnValue = Image_NewFromSystemResource(deviceHandle, identifier, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromSystemResource", &error);
		return 0;
	}
}

/**
 * Creates a new graphics context on the image identified by
 * <code>handle</code>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1NewGraphics(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewGraphics\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	returnValue = Image_NewGraphics(handle, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewGraphics", &error);
		return 0;
	}
}


/**
 * Returns the direct color model masks of the image identified by
 * <code>handle</code>.
 * <p>
 * Color masks are used to access individual color components of a
 * direct color pixel. There is one mask for each of the red, green,
 * blue and alpha component of a pixel. The masks for a 32 bit
 * per pixel image for example might be red=0xFF, green=0x00FF,
 * blue=0x0000FF, alpha=0x000000FF.
 * </p>
 * <p>
 * Example: Given a pixel value, <code>int pixel</code>, returned by
 * <code>Image_GetPixelsHighColor()</code> the red color component can be extracted
 * as follows.
 * <pre>
 * int[] masks = Image_GetDirectPaletteMasks(imageHandle);
 * int redMask = masks[OS.PALETTE_MASK_RED];
 * int redShift = getMaskShift(redMask);
 * int red = (pixel & redMask) >>> redShift;
 * </pre>
 * Where getMaskShift is defined separately to return the bit shift
 * necessary to shift the masked bits to the low order bits.
 * <pre>
 * int getMaskShift(int colorMask) {
 * int shiftCount = 0;
 * if (mask == 0) return 0;
 * while (((mask >> shiftCount) & 0x1) == 0) {
 * shiftCount++;
 * }
 * return shiftCount;
 * }
 * </pre>
 * </p>
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - direct palette not supported for this color depth</li>
 * </ul>
 * </p>
 */
JNIEXPORT jintArray JNICALL Java_com_ibm_ugl_eswt_OS_Image_1GetDirectPaletteMasks(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Int returnAlphaMask;
	UGL_Int returnBlueMask;
	UGL_Int returnGreenMask;
	UGL_Int returnRedMask;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_GetDirectPaletteMasks\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	Image_GetDirectPaletteMasks(handle, &returnRedMask, &returnGreenMask, &returnBlueMask, &returnAlphaMask, &error);
	if (error.error_code == 0) {
		return createJavaPaletteMasks(env, returnRedMask, returnGreenMask, returnBlueMask, returnAlphaMask);
	} else {
		UGL_ErrorHandler(env, "Image_GetDirectPaletteMasks", &error);
		return 0;
	}
}

/**
 * Returns the indexed palette of the image identified by <code>handle</code>.
 * The red, green and blue arrays must be big enough to hold all color
 * indices for the image's color depth.
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - indexed palette not supported for this color depth</li>
 * </ul>
 * </p>
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_Image_1GetIndexedPalette(JNIEnv* env, jclass UNUSED(clazz), jint handle, jbyteArray red, jbyteArray green, jbyteArray blue)
{
	UGL_ErrorStruct error;
	UGL_ByteArray uglblue;
	UGL_ByteArray uglgreen;
	UGL_ByteArray uglred;
	
	uglred = GetByteArrayElements(env, red);
	if (ExceptionCheck(env)) return ;
	uglgreen = GetByteArrayElements(env, green);
	if (ExceptionCheck(env)) return ;
	uglblue = GetByteArrayElements(env, blue);
	if (ExceptionCheck(env)) return ;
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_GetIndexedPalette\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APIARGTRACE(env, "\tuglred (ByteArray)=%p\n" , uglred);
	APIARGTRACE(env, "\tuglgreen (ByteArray)=%p\n" , uglgreen);
	APIARGTRACE(env, "\tuglblue (ByteArray)=%p\n" , uglblue);
	#endif

	Image_GetIndexedPalette(handle, uglred, uglgreen, uglblue, &error);
	ReleaseByteArrayElements(env, red, uglred);
	ReleaseByteArrayElements(env, green, uglgreen);
	ReleaseByteArrayElements(env, blue, uglblue);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Image_GetIndexedPalette", &error);
	}
}

/*
 * Class:     com_ibm_ugl_eswt_OS
 * Method:    Image_GetPixelsHighColor
 * Signature: (I)[I
 */
JNIEXPORT jintArray JNICALL Java_com_ibm_ugl_eswt_OS_Image_1GetPixelsHighColor
#ifdef __cplusplus
  (JNIEnv *env, jclass, jint imageHandle)
#else  
  (JNIEnv *env, jclass clazz, jint imageHandle)
#endif
{
	UGL_ErrorStruct error;
  	jint dataSize; 
	jintArray javaArray;
	UGL_Int x, y, width, height;
  	
  	error.error_code = 0;
  	Image_GetBounds(imageHandle, &x, &y, &width, &height, &error);
  	dataSize = width * height;
  	
  	if (error.error_code == 0) {
	  	javaArray = NewIntArray(env, dataSize);
	  	if (ExceptionCheck(env)) return NULL;
	  	
		if (dataSize > 0) {
		  	jint *pixels = GetIntArrayElements(env, javaArray);
			if (ExceptionCheck(env)) return javaArray;		  	
	    	Image_GetPixelsHighColor(imageHandle, pixels, dataSize, &error);
	    	ReleaseIntArrayElements(env, javaArray, pixels);
			if (error.error_code != 0) {
				UGL_ErrorHandler(env, "Image_GetPixelsHighColor", &error);
				return 0;
			}
		}
		return javaArray;
	} else {
		UGL_ErrorHandler(env, "Image_GetPixelsHighColor", &error);
		return 0;	
	}
}
/*
 * Class:     com_ibm_ugl_eswt_OS
 * Method:    Image_GetPixelsLowColor
 * Signature: (I)[B
 */
JNIEXPORT jbyteArray JNICALL Java_com_ibm_ugl_eswt_OS_Image_1GetPixelsLowColor
#ifdef __cplusplus
  (JNIEnv *env, jclass, jint imageHandle)
#else  
  (JNIEnv *env, jclass clazz, jint imageHandle)
#endif
{
	UGL_ErrorStruct error;
  	jint dataSize;
	jbyteArray javaArray; 
	
	error.error_code = 0;
	dataSize = Image_GetSize(imageHandle, &error);
	if (error.error_code == 0) { 
		javaArray = NewByteArray(env, dataSize);
		if (ExceptionCheck(env)) return NULL;
	  	
		if (dataSize > 0) {
		  	jbyte *pixels = GetByteArrayElements(env, javaArray);
			if (ExceptionCheck(env)) return javaArray;
	    	Image_GetPixelsLowColor(imageHandle, pixels, dataSize, &error);
	    	ReleaseByteArrayElements(env, javaArray, pixels);
			if (error.error_code != 0) {
				UGL_ErrorHandler(env, "Image_GetPixelsLowColor", &error);
				return 0;
			}
		}
		return javaArray;
	} else {
		UGL_ErrorHandler(env, "Image_GetPixelsLowColor", &error);
		return 0;
	}
}
/**
 * Returns the transparent color of the image identified by
 * <code>handle</code> or -1 if the image has no transparency.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_Image_1GetTransparentColor(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_GetTransparentColor\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	returnValue = Image_GetTransparentColor(handle, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_GetTransparentColor", &error);
		return 0;
	}
}

/**
 * Returns whether the image identified by <code>handle</code> uses
 * a direct color model or a paletted color model.
 */
JNIEXPORT jboolean JNICALL Java_com_ibm_ugl_eswt_OS_Image_1IsDirect(JNIEnv* env, jclass UNUSED(clazz), jint handle)
{
	UGL_ErrorStruct error;
	UGL_Boolean returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_IsDirect\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	#endif

	returnValue = Image_IsDirect(handle, &error);
	if (error.error_code == 0) {
		return convertToJavaBoolean(env, returnValue);
	} else {
		UGL_ErrorHandler(env, "Image_IsDirect", &error);
		return JNI_FALSE;
	}
}

/**
 * Creates a new image loaded from the specified file.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_IO - I/O error occurred opening or reading the image file</li>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_Image_1NewFromFile(JNIEnv* env, jclass UNUSED(clazz), jint deviceHandle, jstring fileName)
{
	UGL_ErrorStruct error;
	const jchar* fileNameCharData;
	jint fileNameLength;
	UGL_Boolean fileNameShouldFree;
	UGL_Int returnValue;
	UGL_String uglfileName;
	
	fileNameCharData = GetStringChars(env, fileName);
	if (ExceptionCheck(env)) return 0;
	fileNameLength = GetStringLength(env, fileName);
	uglfileName = convertToNativeString(fileNameCharData, fileNameLength, &fileNameShouldFree);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_NewFromFile\n");
	APIARGTRACE(env, "\tdeviceHandle=%d\n" , deviceHandle);
	APITRACE(env, "\tfileNameCharData=");
	{
		int i;
		for (i=0; i<fileNameLength; i++) APIARGTRACE(env, "%c" , fileNameCharData[i]);
	}
	APITRACE(env, "\n");
	#endif

	returnValue = Image_NewFromFile(deviceHandle, uglfileName, &error);
	if (fileNameShouldFree) freeNativeString(uglfileName);
	ReleaseStringChars(env, fileName, fileNameCharData);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Image_NewFromFile", &error);
		return 0;
	}
}

/**
 * Copies the specified area from the image identified by <code>handle</code>
 * into the destination image. The area is copied from location <code>x</code>,
 * <code>y</code> in the source image to location 0, 0 in the destination
 * image.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_p3ml_OS_Image_1CopyArea(JNIEnv* env, jclass UNUSED(clazz), jint handle, jint destImageHandle, jint x, jint y, jint width, jint height)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Image_CopyArea\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APIARGTRACE(env, "\tdestImageHandle=%d\n" , destImageHandle);
	APIARGTRACE(env, "\tx=%d\n" , x);
	APIARGTRACE(env, "\ty=%d\n" , y);
	APIARGTRACE(env, "\twidth=%d\n" , width);
	APIARGTRACE(env, "\theight=%d\n" , height);
	#endif

	Image_CopyArea(handle, destImageHandle, x, y, width, height, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Image_CopyArea", &error);
	}
}


#ifdef __cplusplus
}
#endif

