/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
#define _WIN32_WINNT 0x0500

#include "ugl_win32.h"

#include "Callback.h"
#include "UGL_Win32_DC.h"
#include "UGL_Win32_Display.h"
#include "UGL_Win32_Widget.h"
#include "UGL_Win32_Menu.h"
#include "UGL_Win32_Tray.h"
#include "KeyUtils.h"

#include "Shell.h"
#include "Color.h"
#include "Button.h"
#include "Label.h"
#include "TextArea.h"
#include "Composite.h"
#include "List.h"
#include "Combo.h"
#include "Device.h"

#include "Display.h"
#include <stdio.h>
#include <shlwapi.h>

static LRESULT CALLBACK	creationProc(HWND, UINT, WPARAM, LPARAM);

static VOID CALLBACK _Timer_Proc(HWND, UINT, UINT_PTR, DWORD);
void _Dispose_Display(UGL_Int handle, UGL_Error uglError);

#define MENU_ITEMS_INCREMENT 4

/**
 * Strange behavior happens when menu IDs less than 108 are used.  The
 * PocketPC will display some system menus when certian low IDs are used.
 * This offset is the difference between the menu item index in the global
 * list and the value of the id.
 */
#define MENU_ITEM_ID_OFFSET 108

HINSTANCE g_hInstance = NULL;
BOOL WINAPI DllMain(HANDLE hInstDLL, DWORD dwReason, LPVOID lpvReserved)
{
	if (dwReason == DLL_PROCESS_ATTACH) {
		if (g_hInstance == NULL) g_hInstance = hInstDLL;
	}
	return TRUE;
}

//========================================================================================

CallbackLookup* Display_GetCallbacks(UGL_Int displayHandle) {
	PUGL_Display display = UGL_DISPLAY(displayHandle);
	return display->callback_lookup;
}
//========================================================================================

void Display_SetCallbacks(UGL_Int displayHandle, CallbackLookup* callbacks) {
	PUGL_Display display = UGL_DISPLAY(displayHandle);
	display->callback_lookup = callbacks;
}

//========================================================================================

void Display_StartTimer(UGL_Int handle, UGL_Int milliseconds, UGL_Int timerId, UGL_Error uglError) {
	PUGL_Display ugl_display = UGL_DISPLAY(handle);

	// Although the MSDN documentation does not mention this, SetTimer will fail on 
	// PocketPC when a timerId of 0 is used.  Bumping the timerId by 1 will avoid the 
	// failure, but it is important that the 1 be subtracted from the id before being 
	// sent back to Java.  See _Timer_Proc to see the id being bumped down.
	int newTimer = SetTimer(ugl_display->creation_hWnd, timerId+1, milliseconds, _Timer_Proc);
	if (newTimer == 0) {
		Win32Error_SetError(uglError, GetLastError(), _T("Failed to create timer"));
	}
}

//========================================================================================

UGL_Int Display_New(UGL_Error uglError) {
	PUGL_Device ugl_device;
	PUGL_Display ugl_display = NULL;

	WNDCLASS staticClass;
	WNDCLASS wndClass;
	HWND creation_hWnd = NULL;

	if (GetClassInfo(g_hInstance, _T("STATIC"), &staticClass) == FALSE) {
		Win32Error_SetError(uglError, GetLastError(), _T("Unable get \"STATIC\" class data"));
		return 0;
	}

	memset(&wndClass, 0, sizeof(WNDCLASS));

	wndClass.hInstance			= g_hInstance;
	wndClass.lpfnWndProc		= (WNDPROC)DefWindowProc; 
#ifdef _WIN32_WCE
	wndClass.style				= CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
#else
	wndClass.style				= CS_BYTEALIGNWINDOW | CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
#endif
	wndClass.hCursor			= staticClass.hCursor;
	wndClass.lpszClassName		= WINDOW_CLASS_NAME;
	wndClass.hbrBackground		= staticClass.hbrBackground;
	RegisterClass (&wndClass);


	ugl_display = UGL_DISPLAY(calloc(sizeof(UGL_Display), 1));

	ugl_device = UGL_DEVICE(ugl_display);
	ugl_device->device_dispose = _Dispose_Display;

	// message only window that will be called to create UI on the display thread
	creation_hWnd = CreateWindowEx(0, WINDOW_CLASS_NAME, NULL, WS_OVERLAPPED, 0, 0, 0, 0, 0, NULL, g_hInstance, NULL);
	if (!creation_hWnd) {
		free(ugl_display);
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to create message/creation window"));
		return NULL_HANDLE;
	}

	SetWindowLong(creation_hWnd, GWL_WNDPROC, (LONG)creationProc);
	SetWindowLong(creation_hWnd, GWL_USERDATA, (LONG)ugl_display);

	ugl_display->callback_lookup = NULL;
	ugl_display->creation_hWnd = creation_hWnd;
	ugl_display->processID = GetCurrentProcessId();
	ugl_display->threadID = GetCurrentThreadId();
	ugl_display->hInstance = g_hInstance;

	ugl_display->menuItemsLength = 0;
	ugl_display->menuItemCount = 0;
	ugl_display->menuItems = NULL;

	return (UGL_Int)ugl_display;
}
void _Display_AddMenu(PUGL_Display display, UGL_Int menuHandle, HMENU hMenu, UGL_Error uglError) {
	PMenuList menuList, endMenuList;
	if (display->menus == NULL) {
		menuList =  (PMenuList)calloc(sizeof(MenuList), 1);
		display->menus = menuList;
	} else {
		menuList = display->menus;
		while (menuList != NULL) {
			endMenuList = menuList;
			menuList = menuList->next;
		}
		menuList = (PMenuList)calloc(sizeof(MenuList), 1);
		endMenuList->next = menuList;
	}
	menuList->menuHandle = menuHandle;
	menuList->hMenu = hMenu;
}
void _Display_RemoveMenu(PUGL_Display display, UGL_Int menuHandle, UGL_Error uglError) {
	PMenuList menuList = display->menus;
	PMenuList prevMenu;
	if (menuList->menuHandle == menuHandle) {
		display->menus = menuList->next;
	} else {
		prevMenu = menuList;
		menuList = menuList->next;
		while (menuList != NULL) {
			if (menuList->menuHandle == menuHandle) break;
			prevMenu = menuList;
			menuList = menuList->next;
		}
		if (menuList == NULL) return;
		prevMenu->next = menuList->next;
	}
	free(menuList);
}
UGL_Int _Display_FindMenu(PUGL_Display display, HMENU hMenu) {
	PMenuList menuList = display->menus;
	while (menuList != NULL) {
		if (menuList->hMenu == hMenu) return menuList->menuHandle;
		menuList = menuList->next;
	}
	return 0;
}
PUGL_Shell _Display_FindClosestShell(HWND child) {
	while (child != NULL) {
		LONG style = GetWindowLong(child, GWL_STYLE);
		if ((style & WS_CHILD) != WS_CHILD) {
			PUGL_Shell shell = (PUGL_Shell)GetWindowLong(child, GWL_USERDATA);
			return shell;
		}
		child = GetParent(child);
	} 
	return NULL;
}
int _Display_GetComCtrlMajorVersion(PUGL_Display display, UGL_Error uglError) {
	HINSTANCE library;
    DLLVERSIONINFO dvi;

	if (display->comCtrl32Major != 0) return display->comCtrl32Major;
	library = LoadLibrary(_T("comctl32.dll"));
	if (library == NULL) {
		Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Failed to load comctl32.dll"));
		return 0;
	} else {
		DLLGETVERSIONPROC pDllGetVersion = (DLLGETVERSIONPROC)GetProcAddress(library, "DllGetVersion");
		if (pDllGetVersion != NULL) {
            HRESULT result;
            dvi.cbSize = sizeof(dvi);
            result = pDllGetVersion(&dvi);
            if(result == S_OK) {
				display->comCtrl32Major = dvi.dwMajorVersion;
				display->comCtrl32Minor = dvi.dwMinorVersion;
				FreeLibrary(library);
				return dvi.dwMajorVersion;
			} 
		} 
		Win32Error_SetError(uglError, GetLastError(), _T("Unable to get comctrl version"));
        FreeLibrary(library);
    }
	return 0;
}

BOOL _Display_FilterMessage(PUGL_Display display, LPMSG msg) {
	/*
	 * Although the MSDN documentation for TranslateAccelerator says that
	 * only WM_KEYDOWN/KEYUP/SYSKEYDOWN/SYSKEYUP messages are translated,
	 * accelerators were not dispatched correctly unless ALL key messages were
	 * handled
	 */
	if ((msg->message >= WM_KEYFIRST) && (msg->message <= WM_KEYLAST)) {
		PUGL_Shell shell = _Display_FindClosestShell(msg->hwnd);
		if (shell != NULL) {
			HACCEL table = _Shell_GetAcceleratorTable(shell);
			if (table != 0) {
				if (TranslateAccelerator(shell->scrollable.control.hWnd, table, msg) != 0) {
					return TRUE;
				}
			}
		}
	}
	return FALSE;
}

//========================================================================================

UGL_Boolean Display_ReadAndDispatch(UGL_Int handle, UGL_Error uglError) {
	PUGL_Display display = UGL_DISPLAY(handle);
	MSG msg;

	if (PeekMessage (&msg, 0, 0, 0, PM_REMOVE)) {
		if (_Display_FilterMessage(display, &msg) == TRUE) return TRUE;
		TranslateMessage (&msg);
		DispatchMessage (&msg);
		return TRUE;
	}
    return FALSE;
}
//========================================================================================

void Display_Sleep(UGL_Int handle, UGL_Error uglError) {
#ifdef _WIN32_WCE
	MsgWaitForMultipleObjectsEx (0, 0, INFINITE, QS_ALLINPUT, MWMO_INPUTAVAILABLE);
#else
	MSG msg;

	// According to MSDN, WaitMessage will not return immediately if
	// there is a message already on the queue.  In PersonalProfile that is not a 
	// problem because the event loop is not controled by the user, but in eSWT 
	// it is possible to write an incorrect event loop.
	//
	// If event loop sleeps after every event is read and dispatched then some
	// events can be queued up and only processed after a new event comes in.  
	//
	// Checking for a pre-existing message will fix this problem
	if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) != 0) return;
	WaitMessage();
#endif
}
//========================================================================================

void Display_Wake(UGL_Int handle, UGL_Error uglError) {
	PUGL_Display ugl_display = UGL_DISPLAY(handle);
#ifdef _WIN32_WCE
	PostMessage (ugl_display->creation_hWnd, WM_NULL, 0, 0); 
#else
	PostThreadMessage(ugl_display->threadID, WM_NULL, 0, 0);
#endif
}
//========================================================================================

UGL_Boolean Display_PostKeyEvent(UGL_Int handle, UGL_Int keyEventType, UGL_Int keyCode, UGL_Char character, UGL_Error uglError) {

	/*
	 * This function doesn't work quite like one would imagine. For example, if I send a 
	 * shift down event followed by a key down event with the 's' key, I don't get the 'S'
	 * char in the generated event, but another unrecognized character. This is the exact same
	 * behavior as SWT though.
	 * Also note that whether the upper- or lower-case keycode for a character is passed,
	 * the keybd_event() function will only take the keycode of the upper-case character, and 
	 * the generated event will always have the keycode of the lower-case character.
	 */
	
	if (keyCode > 96 && keyCode < 123) keyCode = keyCode - 32;
	
	else {
		keyCode = KeyUtils_ConvertToNativeKeycode(keyCode);
	}

	if (keyEventType == UGL_EVENT_KEY_DOWN) {
		keybd_event( keyCode, 0, KEYEVENTF_EXTENDEDKEY | 0, 0 ); 
	}
	else {
		keybd_event( keyCode, 0, KEYEVENTF_EXTENDEDKEY | KEYEVENTF_KEYUP, 0 );
	}
	
	return TRUE;


}
//========================================================================================

LRESULT CALLBACK creationProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam) {
	if (message >= WM_USER) {

		switch (message) {
			case UI_NOTIFY_TRAY: {
				if(lParam == WM_LBUTTONUP) {
					PUGL_Display ugl_display = UGL_DISPLAY(GetWindowLong(hWnd, GWL_USERDATA));
					PUGL_TrayItem ugl_trayitem = Tray_GetTrayItem(ugl_display, (UINT)wParam);
#ifndef _WIN32_WCE
					POINT point;
					GetCursorPos(&point);
					TrayItemCallback(UGL_WIDGET(ugl_trayitem)->callback_target, TRAYITEM_SELECTION, point.x, point.y);
#else
					TrayItemCallback(UGL_WIDGET(ugl_trayitem)->callback_target, TRAYITEM_SELECTION, 0, 0);
#endif
				} else if(lParam == WM_LBUTTONDBLCLK) {
					PUGL_Display ugl_display = UGL_DISPLAY(GetWindowLong(hWnd, GWL_USERDATA));
					PUGL_TrayItem ugl_trayitem = Tray_GetTrayItem(ugl_display, (UINT)wParam);
#ifndef _WIN32_WCE
					POINT point;
					GetCursorPos(&point);
					TrayItemCallback(UGL_WIDGET(ugl_trayitem)->callback_target, TRAYITEM_DEFAULT_SELECTION, point.x, point.y);
#else
					TrayItemCallback(UGL_WIDGET(ugl_trayitem)->callback_target, TRAYITEM_DEFAULT_SELECTION, 0, 0);
#endif
#ifndef _WIN32_WCE
				} else if(lParam == WM_RBUTTONUP) {
					PUGL_Display ugl_display = UGL_DISPLAY(GetWindowLong(hWnd, GWL_USERDATA));
					PUGL_TrayItem ugl_trayitem = Tray_GetTrayItem(ugl_display, (UINT)wParam);
					POINT point;
					GetCursorPos(&point);
					TrayItemCallback(UGL_WIDGET(ugl_trayitem)->callback_target, TRAYITEM_MENU_DETECT, point.x, point.y);
#endif
				}
			} break;
			case UI_CREATE_MENUBAR: {
				UGL_Error* error = (UGL_Error*)lParam;
                return (LRESULT)_MenuBar_New((PUGL_Control)wParam, *error);
			}
			case UI_CREATE_POPUP: {
				UGL_Error* error = (UGL_Error*)lParam;
                return  _PopupMenu_New(wParam, *error);
			}

			case UI_DESTROY_WINDOW: {
				return DestroyWindow((HWND)wParam);
			}
	// ----------------------------------------------------
			// we only have two possible messages defined. Create either a shell, or create
			// a control.
			case UI_CREATE_SHELL: {
				PUGL_Display ugl_display = UGL_DISPLAY(GetWindowLong(hWnd, GWL_USERDATA));
				UGL_Int *args = (UGL_Int *)lParam;
				UGL_Int parent_handle = args[0];
				UGL_Error uglError = (UGL_Error)args[1];
				UGL_Int ugl_style = (UGL_Int)wParam;
				
				UGL_Int shell_handle = 0;
				if (parent_handle == 0) shell_handle = Shell_New((UGL_Int)ugl_display, ugl_style, uglError);
				else shell_handle = Shell_NewChild(parent_handle, ugl_style, uglError);

#ifdef _WIN32_WCE
				return (LRESULT)shell_handle;
#else
				ReplyMessage(shell_handle);
#endif

						  } break;

			// create a control
			case UI_CREATE_CONTROL: {
				UGL_Int *args = (UGL_Int *)lParam;
				UGL_Int parentHandle = (UGL_Int)wParam;
				UGL_String control_class = (UGL_String)args[0];
				UGL_Int struct_size = args[1];
				WNDPROC Control_Proc = (WNDPROC)args[2];
				UGL_Int ugl_style = args[3];
				UGL_Int style = args[4];
				UGL_Int extended_style = args[5];
				UGL_Error uglError = (UGL_Error)args[6];

				PUGL_Control ugl_handle = _Control_New(control_class, struct_size, Control_Proc, parentHandle, ugl_style, style, extended_style, uglError);
#ifdef _WIN32_WCE
				return (LRESULT)ugl_handle;
#else
				ReplyMessage((LRESULT)ugl_handle);
#endif
			} break;

			case UI_SETCALLBACKTARGET: {
				// Please see Widget_SetCallbackTarget for
				// a detailed description of why this
				// must be done in the display thread
				PUGL_Widget widget = UGL_WIDGET(wParam);
				CallbackTarget* newTarget = (CallbackTarget*)lParam;
				CallbackTarget* old_target = widget->callback_target;
				widget->callback_target = newTarget;
				return (LRESULT)old_target;
			}

		}
		return 1;
	} else {
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
}
//========================================================================================

UGL_Boolean Display_PostPointerEvent(UGL_Int handle, UGL_Int type, UGL_Int button, UGL_Error uglError) {
	INPUT inputs;
	UINT result;
	ZeroMemory(&inputs, sizeof(INPUT));
	inputs.type = INPUT_MOUSE;
	
	switch (button) {
		case UGL_POINTER_MASK_BUTTON1: {
			if (type == UGL_EVENT_POINTER_DOWN) inputs.mi.dwFlags = MOUSEEVENTF_LEFTDOWN;
			else inputs.mi.dwFlags = MOUSEEVENTF_LEFTUP;
		} break;
		case UGL_POINTER_MASK_BUTTON2: {
			if (type == UGL_EVENT_POINTER_DOWN) inputs.mi.dwFlags = MOUSEEVENTF_MIDDLEDOWN;
			else inputs.mi.dwFlags = MOUSEEVENTF_MIDDLEUP;
		} break;
		case UGL_POINTER_MASK_BUTTON3: {
			if (type == UGL_EVENT_POINTER_DOWN) inputs.mi.dwFlags = MOUSEEVENTF_RIGHTDOWN;
			else inputs.mi.dwFlags = MOUSEEVENTF_RIGHTUP;
		} break;
		default: return FALSE;
	}
	result = SendInput(1, &inputs, sizeof(INPUT));
	return result != 0;
}
//========================================================================================

UGL_Boolean Display_PostPointerMoveEvent(UGL_Int handle, UGL_Int x, UGL_Int y, UGL_Error uglError) {
	INPUT inputs;
	UINT result;
	int screenWidth = GetSystemMetrics(SM_CXSCREEN) - 1;
	int screenHeight = GetSystemMetrics(SM_CYSCREEN) - 1;

	ZeroMemory(&inputs, sizeof(INPUT));
	inputs.type = INPUT_MOUSE;
	inputs.mi.dwFlags = MOUSEEVENTF_MOVE | MOUSEEVENTF_ABSOLUTE; 
	inputs.mi.dx = (x * 65535) / screenWidth;
	inputs.mi.dy = (y * 65535) / screenHeight;

	result = SendInput(1, &inputs, sizeof(INPUT));
	return result != 0;
}


//========================================================================================
void Display_Beep(UGL_Int handle, UGL_Error uglError) {
	MessageBeep(MB_OK);
}

//========================================================================================

UGL_Int Display_GetSystemColor(UGL_Int handle, UGL_Int id, UGL_Error uglError) {
	int pixel = 0x02000000;
	UGL_Int ugl_color = 0;

	switch (id) {	
		case UGL_COLOR_CONTROL_SHADOW_DARK:			pixel = GetSysColor (COLOR_3DDKSHADOW);	break;		
		case UGL_COLOR_CONTROL_SHADOW_NORMAL:		pixel = GetSysColor(COLOR_3DSHADOW); 	break;
		case UGL_COLOR_CONTROL_HIGHLIGHT_NORMAL:	pixel = GetSysColor(COLOR_3DLIGHT);		break;
		case UGL_COLOR_CONTROL_BACKGROUND: 			pixel = GetSysColor(COLOR_3DFACE);  	break;
		case UGL_COLOR_CONTROL_FOREGROUND:			pixel = GetSysColor(COLOR_WINDOWTEXT);	break;


		case UGL_COLOR_TEXT_HIGHLIGHT_BACKGROUND:	pixel = GetSysColor(COLOR_HIGHLIGHT);	break;
		case UGL_COLOR_TEXT_HIGHLIGHT_FOREGROUND:	pixel = GetSysColor(COLOR_HIGHLIGHTTEXT); break;

		case UGL_COLOR_TITLE_BACKGROUND:			pixel = GetSysColor(COLOR_ACTIVECAPTION); break;
		case UGL_COLOR_TITLE_FOREGROUND:			pixel = GetSysColor(COLOR_CAPTIONTEXT); break;

		case UGL_COLOR_INFO_FOREGROUND:				pixel = GetSysColor(COLOR_INFOTEXT);	break;
		case UGL_COLOR_INFO_BACKGROUND:				pixel = GetSysColor(COLOR_INFOBK);		break;

		case UGL_COLOR_INACTIVE_CAPTION_FOREGROUND: 	pixel = GetSysColor(COLOR_INACTIVECAPTIONTEXT);	break;
		case UGL_COLOR_INACTIVE_CAPTION_BACKGROUND:		pixel = GetSysColor(COLOR_INACTIVECAPTION);		break;
		default: {
			Win32Error_SetError(uglError, UGL_ERROR_OTHER, _T("Unrecognized color ID"));
			return NULL_HANDLE;
				 } break;
			
	}

	ugl_color = Color_New(handle, GetRValue(pixel), GetGValue(pixel), GetBValue(pixel), uglError);
	return ugl_color;
}

//========================================================================================

/**
 * Adds a control to the list of controls.
 */
BOOL ControlList_AddControl(PControlList* list, struct UGL_Control* control) {
	// It is important that new controls be added to the front of the list so that
	// the topmost control is always in the front of the list.  
	PControlList element;
	for (element = *list; element != NULL; element = element->next) {
		// It is possible that the control is already in the list and it
		// should not be added twice
		if (element->control == control) return TRUE;
	}

	element = (PControlList)calloc(sizeof(ControlList), 1);
	if (element == NULL) return FALSE;
	element->control = control;
	element->next = *list;

	*list = element;
	return TRUE;
}

BOOL ControlList_RemoveControl(PControlList* list, struct UGL_Control* control) {
	if (*list == NULL) {
		return FALSE;  // The control is not in the empty list
	}

	if ((*list)->control == control) {
		PControlList nextList = (*list)->next;
		free(*list);
		*list = nextList;
		return TRUE;
	} else {
		return ControlList_RemoveControl(&((*list)->next), control);
	}
}

BOOL IsChildOfShell(struct UGL_Control* shell, struct UGL_Control* child) {
	// If there is a modal dialog currently displayed then
	// we need to ignore all mouse events on windows that are 
	// not owned by that dialog
	struct UGL_Control* test = child;

	// In PPro it is possible for a modal dialog to open other
	// windows which should still be usable.  The solution it to
	// walk the whole list of shells until the top-most shell checking
	// if any are the modal dialog
	while (test != shell) {
		if (test == NULL) return FALSE; // None of the parents were the modal dialog
		test = test->parent;
	};

	return TRUE;
}

/**
 * Run through the list of shell's that want to be enable and adjust their
 * enabled status based on the current modal dialog.
 */
void UpdateEnabledShells(PUGL_Display display) {
	struct UGL_Control* topmostModalShell = display->modalShells == NULL ? NULL : display->modalShells->control;
	
	PControlList element;
	for (element = display->enabledShells; element != NULL; element = element->next) {
		BOOL enabled;
		struct UGL_Control* shell = element->control;

		// Shells that are children of the modal shell do not need to be 
		// disabled when the modal shell is shown.
		if ((topmostModalShell == NULL) || IsChildOfShell(topmostModalShell, shell)) {
			enabled = TRUE;
		} else {
			enabled = FALSE;
		}

		// It is important that we don't call Control_SetEnabled() to do the work because
		// that will result in the shell being removed from the list of shells that would
		// prefer to be enabled
		EnableWindow(shell->hWnd, enabled);
	}
}

/**
 * When a modal shell is added to the display all other visible shell's are disabled. When 
 * the last modal shell is removed from the display all shells that want to be enabled
 * will be re-enabled
 */
void Display_UpdateModalShellVisibility(PUGL_Display display, struct UGL_Control* modalShell, BOOL visible, UGL_Error error) {
	if (visible) {
		ControlList_AddControl(&(display->modalShells), modalShell);
	} else {
		ControlList_RemoveControl(&(display->modalShells), modalShell);
	}
	UpdateEnabledShells(display);
}

/**
 * This method will track the new shell and manage it's enabled state
 * based on the current modal shell.
 */
void Display_AddShell(PUGL_Display display, struct UGL_Control* shell, UGL_Error error) {
	// Only shells that are enabled need to be tracked,  if the shell 
	// become enabled in the future it will go through Display_EnableShell
	// and be added to the list then.
	if (IsWindowEnabled(shell->hWnd)) {
		if (ControlList_AddControl(&(display->enabledShells), shell)) {
			// If there is a current modal shell and the new shell is not a child of 
			// that shell then it needs to be disabled to reflect the modal status.
			if ((display->modalShells != NULL) && !IsChildOfShell(display->modalShells->control, shell)) {
				// It is important that we don't call Control_SetEnabled() to do the work because
				// that will result in the shell being removed from the list of shells that would
				// prefer to be enabled
				EnableWindow(shell->hWnd, FALSE);
			}
		} else {
			Win32Error_SetError(error, UGL_ERROR_OTHER, _T("Could not add control to control list"));
		}
	}
}

/**
 * Stops tracking the specified shell.  This should only be called 
 * when a shell is being disposed.
 */
void Display_RemoveShell(PUGL_Display display, struct UGL_Control* shell) {
	ControlList_RemoveControl(&(display->enabledShells), shell);
}

/**
 * Returns TRUE if the shell needs to be updated with the new
 * enabled status.  It will return FALSE if the shell already has the 
 * style specified.  For example, If the this method is called to enable a shell while
 * a modal shell is visible it will return FALSE, since the shell must remain 
 * disabled while the dialog is visible, but when the modal shell is hidden the desired
 * enabled state will be remembered and updated.
 */
BOOL Display_EnableShell(PUGL_Display display, struct UGL_Control* shell, BOOL enable) {
	if (enable) {
		ControlList_AddControl(&(display->enabledShells), shell);
	} else {
		ControlList_RemoveControl(&(display->enabledShells), shell);
	}

	if ((display->modalShells != NULL) && !IsChildOfShell(display->modalShells->control, shell)) {
		// Since there is either no modal shell or the specified shell is a child of the modal shell
		// it can do whatever it would like.
		return TRUE;
	} else {
		return FALSE;
	}
}


void _Dispose_Display(UGL_Int handle, UGL_Error uglError) {
	PUGL_Display ugl_display = UGL_DISPLAY(handle);

	// destroy creation window

	if (ugl_display->defaultControlFont != NULL) {
		Font_Dispose((UGL_Int)ugl_display->defaultControlFont, uglError);
	}

	DestroyWindow(ugl_display->creation_hWnd);

	if (!UnregisterClass(WINDOW_CLASS_NAME, ugl_display->hInstance)) {
		free(ugl_display);
//		Win32Error_SetError(uglError, GetLastError(), _T("Unable to unregister class for device"));
		return;
	}

	ugl_display->callback_lookup = NULL;

	if (ugl_display->menuItems != NULL) {
		free(ugl_display->menuItems);
	}

	free(ugl_display);
}



//========================================================================================

void Display_GetIconSize(UGL_Int handle, UGL_Int* returnX, UGL_Int* returnY, UGL_Error uglError) {	
	*returnX = GetSystemMetrics(SM_CXICON);
	*returnY = GetSystemMetrics(SM_CYICON);
}
//========================================================================================

UGL_Int Display_GetDoubleClickTime(UGL_Int handle, UGL_Error uglError) {
	return GetDoubleClickTime();
}
//========================================================================================

void Display_SetAppName(UGL_String name, UGL_Error uglError) {
	// Nothing to do on Windows
}

//========================================================================================

void CALLBACK _Timer_Proc(HWND hwnd, UINT msg, UINT_PTR timer_ID, DWORD dwTime) {
	PUGL_Display ugl_display = UGL_DISPLAY(GetWindowLong(hwnd, GWL_USERDATA));
	
	// Because of PocketPC not liking timerIds with a value of 0, the id that
	// Java is expecting will be 1 less then the one comming into this method.
	// See Display_StartTimer() for me details.
	DisplayTimerCallback((UGL_Int)ugl_display, (UGL_Int)timer_ID-1);
}

void Display_AddMenuItem(struct UGL_MenuItem *menuItem) {
	PUGL_Display display = UGL_WIDGET(menuItem)->display;
	struct UGL_MenuItem **menuItems;
	int index;	
#ifdef _WIN32_WCE
	if (menuItem->parent_menu->hwndCB != 0) {
		display->menuItemCount+=1;
	}
#endif
	for (index = 0; index < display->menuItemsLength; index++) {
		if (display->menuItems[index] == NULL) {
			display->menuItems[index] = menuItem;
			menuItem->id = index + MENU_ITEM_ID_OFFSET;
			return;
		}
	}
	menuItems = display->menuItems;
	display->menuItems = calloc(display->menuItemsLength + MENU_ITEMS_INCREMENT, sizeof(PUGL_MenuItem));
	memcpy(display->menuItems, menuItems, display->menuItemsLength * sizeof(PUGL_MenuItem));
	if (menuItems != NULL) {
		free(menuItems);
	}
	display->menuItems[display->menuItemsLength] = menuItem;
	menuItem->id = display->menuItemsLength + MENU_ITEM_ID_OFFSET;
	display->menuItemsLength += MENU_ITEMS_INCREMENT;
}

PUGL_MenuItem Display_GetMenuItem(PUGL_Display display, int id) {
	PUGL_MenuItem menuItem = NULL;
	int index = id - MENU_ITEM_ID_OFFSET;

	if (index >= 0 && index < display->menuItemsLength) {
		menuItem = display->menuItems[index];
	}
	return menuItem;
}

void Display_RemoveMenuItem(struct UGL_MenuItem *menuItem) {
	PUGL_Display display = UGL_WIDGET(menuItem)->display;
	int menuItemIndex = menuItem->id - MENU_ITEM_ID_OFFSET;

	if (menuItemIndex >= 0 && menuItemIndex < display->menuItemsLength) {
		display->menuItems[menuItemIndex] = NULL;
		menuItemIndex = -1;
	}
#ifdef _WIN32_WCE
	if (menuItem->parent_menu->hwndCB != 0) {
		display->menuItemCount -= 1;
	}
#endif
}

#ifdef _WIN32_WCE
// this function finds the menuitem in the main menu (top level menu) at index (index)
/*
PUGL_MenuItem Display_FindMenuItem(PUGL_Display display, int index) {
	int i;
	
	for(i=0; i<display->menuItemsLength; i++){
		if(display->menuItems[i] != NULL){
			if(display->menuItems[i]->index == index){
				if(display->menuItems[i]->parent_menu->hwndCB != 0) {
					return display->menuItems[i];
				}
			}
		}
	}

	return NULL;
}
*/

PUGL_MenuItem Display_FindMenuItem(PUGL_Display display, HWND hwndCB, int index) {
	int i;
	for(i=0; i<display->menuItemsLength; i++){
		if(display->menuItems[i] != NULL){
			if (display->menuItems[i]->parent_menu->hwndCB != hwndCB) continue;
			if(display->menuItems[i]->index == index){
				if(display->menuItems[i]->parent_menu->hwndCB != 0) {
					return display->menuItems[i];
				}
			}
		}
	}
	return NULL;
}


// this function decrements the index of menuitems in the main menu (top level menu) starting at index (start)
// for example, when we remove menuItem index 3 we would wish to decrement the index of all menuitems greater than 3 by one
/*
void Display_DecMenuItemIndex(PUGL_Display display, HWND hwndCB, int start){
	// new for softkey, maintain the menuitem order
	int index;
		for (index = 0; index < display->menuItemsLength; index++) {
			
			if (display->menuItems[index] != NULL) {
				if(display->menuItems[index]->parent_menu->hwndCB != 0){
					//Fix duplicate menuItem problem.
					if(start == display->menuItems[index]->index) {
					 display->menuItems[index]->index = -1 ;
					}
					if(start < display->menuItems[index]->index){
						display->menuItems[index]->index--;			
					}
				}
			}
		}
}
*/


void Display_DecMenuItemIndex(PUGL_Display display, HWND hwndCB, int start){
	// new for softkey, maintain the menuitem order
	int index;
		for (index = 0; index < display->menuItemsLength; index++) {
			if (display->menuItems[index] != NULL) {
				if (display->menuItems[index]->parent_menu->hwndCB != hwndCB) continue;
				if(display->menuItems[index]->parent_menu->hwndCB != 0){
					//Fix duplicate menuItem problem.
					if(start == display->menuItems[index]->index) {
					 display->menuItems[index]->index = -1 ;
					}
					if(start < display->menuItems[index]->index){
						display->menuItems[index]->index--;			
					}
				}
			}
		}
}

// this function increments the index of menuitems in the main menu (top level menu) starting at index (start)
// for example, when we add menuItem index 3 we would wish to increment the index of all menuitems greater than or equal to 3 by one
void Display_IncMenuItemIndex(PUGL_Display display, HWND hwndCB, int start){
	// new for softkey, maintain the menuitem order
	int index;
	FILE * pFile; 
		for (index = 0; index < display->menuItemsLength; index++) {
			if (display->menuItems[index] != NULL) {
				if (display->menuItems[index]->parent_menu->hwndCB != hwndCB) continue;
				if(display->menuItems[index]->parent_menu->hwndCB != 0){
					if(start <= display->menuItems[index]->index){
						display->menuItems[index]->index++;			
					}
				}
			}
		}
}

// this function returns the menuitem count associated with specific menu hwndCB
int Display_MenuItemCount(PUGL_Display display, HWND hwndCB){
	int index;
	int count=0;
		for (index = 0; index < display->menuItemsLength; index++) {
			if (display->menuItems[index] == NULL) continue;
			if (display->menuItems[index]->parent_menu->hwndCB != hwndCB) continue;
			count ++;
		}
	return count;
}

#endif

//========================================================================================

/**
 * Returns the insets of the display identified by <code>handle</code>.
 * This is the portion of the screen that is unusable for drawing,
 * such as a task bar.
 *
 * @ingroup Display
 *
 * @param handle the display handle
 * @param returnLeft (out) the left inset
 * @param returnTop (out) the top inset
 * @param returnRight (out) the right inset
 * @param returnBottom (out) the bottom inset
 * @param error the error status of the call
 */
void Display_GetInsets(UGL_Int handle, UGL_Int* returnLeft, UGL_Int* returnTop, UGL_Int* returnRight, UGL_Int* returnBottom, UGL_Error error) {
	int left = 0; 
	int top = 0;
	int right = 0; 
	int bottom = 0; 

	RECT workarea;
	int boundsX, boundsY, boundsWidth, boundsHeight;

	SystemParametersInfo (SPI_GETWORKAREA, 0, &workarea, 0);	
	Device_GetBounds(handle, &boundsX, &boundsY, &boundsWidth, &boundsHeight, error);
	
	*returnLeft = workarea.left - boundsX;
	*returnTop = workarea.top - boundsY;
	*returnRight = boundsX + boundsWidth - workarea.right;
	*returnBottom = boundsY + boundsHeight - workarea.bottom;
}

