/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

#ifndef __UGL_IMAGE_H__
#define __UGL_IMAGE_H__

#include "UGL_Error.h"
#include "UGLTypes.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Frees the platform resources associated with the image identified by
 * <code>handle</code>.
 *
 * @ingroup Image
 *
 * @param handle image handle
 * @param error the error status of the call
 */
void Image_Dispose(UGL_Int handle, UGL_Error error);

/**
 * Returns the position and size of the image identified by <code>handle</code>.
 *
 * @ingroup Image
 *
 * @param handle image handle
 * @param returnX (out) the x component of the rect
 * @param returnY (out) the y component of the rect
 * @param returnWidth (out) the width component of the rect
 * @param returnHeight (out) the height component of the rect
 * @param error the error status of the call
 */
void Image_GetBounds(UGL_Int handle, UGL_Int* returnX, UGL_Int* returnY, UGL_Int* returnWidth, UGL_Int* returnHeight, UGL_Error error);

/**
 * Returns the color depth of the image identified by <code>handle</code>.
 * The supported color depths are 1, 2, 4, 8, 16, 24 and 32 bit. The
 * returned depth must be one of the values supported by UGL. In the case
 * of 3, 5, 6 or 7 bit GIFs the returned depth should reflect the actual
 * "color resolution" as defined in the GIF89a spec.
 *
 * @ingroup Image
 *
 * @param handle image handle
 * @param error the error status of the call
 *
 * @return color depth in bits per pixel
 */
UGL_Int Image_GetDepth(UGL_Int handle, UGL_Error error);

/**
 * Creates a new image from the supplied pixel data. Each pixel is an
 * index into a color palette. The color palette is supplied in the
 * indexed color arrays. The palette may not use all possible color
 * values . In other words the number of elements may be less than
 * (2^depth). Supported color depths are 1, 2, 4 and 8 bits per pixel.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @see Image_GetDepth
 *
 * @param deviceHandle the device handle
 * @param width width of the new image in pixels
 * @param height height of the new image
 * @param imageData the pixel data. Each byte holds one or more pixels. An
 *                  8 bit per pixel image has one pixel per byte, a 4 bpp image has exactly
 *                  2 pixels per byte etc. The most significant bit of each byte corresponds
 *                  to the leftmost pixel on a scanline. A scanline may have unused bits if
 *                  the color depth is less than 8 bpp.
 * @param bytesPerLine length in bytes of a scan line. The number of bytes
 *                     per scan line may be greater than required for the actual number of
 *                     pixels. Excess bytes are unused.
 * @param depth color depth of the image data, in number of bits per pixel.
 *              One of 1, 2, 4, 8. It is acceptable for an implementation to create an
 *              image with a different depth if platform limitations do not allow the
 *              original color depth. The depth has to remain the same throughout the
 *              life of the image.
 * @param indexedReds red components of indexed color palette.
 * @param indexedGreens green components of indexed color palette.
 * @param indexedBlues blue components of indexed color palette.
 * @param paletteSize number of entries in the color palette arrays. This
 *                    may be less than the maximum number allowed by the color depth.
 * @param transparentPixel palette index of the transparent pixel. Value of
 *                         <code>-1</code> means no transparency.
 * @param error the error status of the call
 *
 * @return handle to the new image. 0 if no image could be created using the
 *         supplied parameters.
 */
UGL_Int Image_NewFromByteData(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_ByteArray imageData, UGL_Int bytesPerLine, UGL_Int depth, UGL_ByteArray indexedReds, UGL_ByteArray indexedGreens, UGL_ByteArray indexedBlues, UGL_Int paletteSize, UGL_Int transparentPixel, UGL_Error error);

/**
 * Creates a new image using the supplied raw data.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @param deviceHandle the device/display handle
 * @param data the raw image data, including image header
 * @param length number of bytes in the data array
 * @param error the error status of the call
 *
 * @return handle to the new image or 0 if no image could be created
 *         from the supplied data.
 */
UGL_Int Image_NewFromFileData(UGL_Int deviceHandle, UGL_ByteArray data, UGL_Int length, UGL_Error error);

/**
 * Creates a new image from the supplied pixel data. Each pixel is a direct
 * color RGB value. Supported color depths are 16, 24 and 32 bits per pixel.
 * For 16 bit depth the RGB bits are layed out as 5-6-5 bits for red, green
 * and blue respectively, 24 bit depth uses 8-8-8 bits and no alpha channel.
 * 32 bit color depth uses 8-8-8 bits and the most significant byte for the
 * alpha channel for an ARGB pixel layout. For 32 bit pixel data the alpha
 * channel byte is ignored if either one of the
 * <code>transparentColor</code>, <code>globalAlpha</code> or
 * <code>alphaData</code> is specified. Only one of the three transparency/
 * alpha arguments may be specified at a time.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @param deviceHandle the device handle
 * @param width width of the new image in pixels
 * @param height height of the new image
 * @param imageData pixel data. Each int contains one pixel regardless of
 *                  color depth. E.g., a 16 bit pixel is packed in the two least
 *                  significant bytes of each int.
 * @param depth color depth of the image data, in number of bits per pixel.
 *              One of 16, 24, 32. It is acceptable for an implementation to create an
 *              image with a different depth if platform limitations do not allow the
 *              original color depth. The depth has to remain the same throughout the
 *              life of the image.
 * @param transparentColor value of the transparent color. The transparent
 *                         color is specified in the same bit format as the image data. Value of
 *                         <code>-1</code> means no transparency.
 * @param globalAlpha global alpha value. Each pixel will be blended using
 *                    this alpha value. Value of <code>-1</code> means no global alpha value.
 * @param alphaData alpha data of the image. Each pixel may have an alpha
 *                  value ranging from 0 to 255. 0 means fully transparent, 255 means fully
 *                  opaque. May be <code>null</code>.
 * @param error the error status of the call
 *
 * @return handle to the new image. 0 if no image could be created using the
 *         supplied parameters.
 */
UGL_Int Image_NewFromIntData(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_IntArray imageData, UGL_Int depth, UGL_Int transparentColor, UGL_Int globalAlpha, UGL_ByteArray alphaData, UGL_Error error);

/**
 * Creates a new, empty image with the given dimensions.
 * <p>
 * Image will be filled with the provided fill color which
 * is in 0xAARRGGBB format.
 * </p>
 *
 * @ingroup Image
 *
 * @param deviceHandle the device/display handle
 * @param width width of the new image
 * @param height height of the new image
 * @param fillColor value with which to fill the image in 0xAARRGGBB format.
 * @param error the error status of the call
 *
 * @return handle to the new image
 */
UGL_Int Image_NewFromSize(UGL_Int deviceHandle, UGL_Int width, UGL_Int height, UGL_Int fillColor, UGL_Error error);

/**
 * Creates a new graphics context on the image identified by
 * <code>handle</code>
 *
 * @ingroup Image
 *
 * @param handle image handle
 * @param error the error status of the call
 *
 * @return handle to the new graphics context
 */
UGL_Int Image_NewGraphics(UGL_Int handle, UGL_Error error);



/**
 * Returns the direct color model masks of the image identified by
 * <code>handle</code>.
 * <p>
 * Color masks are used to access individual color components of a
 * direct color pixel. There is one mask for each of the red, green,
 * blue and alpha component of a pixel. The masks for a 32 bit
 * per pixel image for example might be red=0xFF, green=0x00FF,
 * blue=0x0000FF, alpha=0x000000FF.
 * </p>
 * <p>
 * Example: Given a pixel value, <code>int pixel</code>, returned by
 * <code>Image_GetPixelsHighColor()</code> the red color component can be extracted
 * as follows.
 * <pre>
 * int[] masks = Image_GetDirectPaletteMasks(imageHandle);
 * int redMask = masks[OS.PALETTE_MASK_RED];
 * int redShift = getMaskShift(redMask);
 * int red = (pixel & redMask) >>> redShift;
 * </pre>
 * Where getMaskShift is defined separately to return the bit shift
 * necessary to shift the masked bits to the low order bits.
 * <pre>
 * int getMaskShift(int colorMask) {
 * int shiftCount = 0;
 * if (mask == 0) return 0;
 * while (((mask >> shiftCount) & 0x1) == 0) {
 * shiftCount++;
 * }
 * return shiftCount;
 * }
 * </pre>
 * </p>
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - direct palette not supported for this color depth</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @see Image_IsDirect
 *
 * @param handle image handle
 * @param returnRedMask (out) the red component of the palette mask
 * @param returnGreenMask (out) the green component of the palette mask
 * @param returnBlueMask (out) the blue component of the palette mask
 * @param returnAlphaMask (out) the alpha component of the palette mask
 * @param error the error status of the call
 */
void Image_GetDirectPaletteMasks(UGL_Int handle, UGL_Int* returnRedMask, UGL_Int* returnGreenMask, UGL_Int* returnBlueMask, UGL_Int* returnAlphaMask, UGL_Error error);

/**
 * Returns the indexed palette of the image identified by <code>handle</code>.
 * The red, green and blue arrays must be big enough to hold all color
 * indices for the image's color depth.
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - indexed palette not supported for this color depth</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @see Image_IsDirect
 *
 * @param handle image handle
 * @param red red component of palette colors
 * @param green green component of palette colors
 * @param blue blue component of palette colors
 * @param error the error status of the call
 */
void Image_GetIndexedPalette(UGL_Int handle, UGL_ByteArray red, UGL_ByteArray green, UGL_ByteArray blue, UGL_Error error);

/**
 * Returns pixel data for images with more than 8 bit per pixel color depth.
 * Each int array element holds exactly one pixel, compatible with the image's 
 * color model. The color components of each pixel are layed out to match the
 * direct palette masks returned by <code>Image_GetDirectPaletteMasks()</code>.
 * <p>
 * There is no scanline pad. Therefore, the size of the returned array equals
 * the image width multiplied with the image height.
 * </p>
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * </ul>  
 * </p>
 * 
 * @param handle image handle
 * @param pixels pixel data of entire image
 * @param length size, in pixels, of the <code>pixels</code> argument
 * @param uglError the error status of the call
 * @see #Image_GetDirectPaletteMasks()
 */
void Image_GetPixelsHighColor(UGL_Int handle, UGL_IntArray pixels, UGL_Int length, UGL_Error uglError);

/**
 * Returns pixel data for images with 8 bit per pixel color depth or less.
 * Each byte array element holds one or more pixels but it never holds 
 * pixels of two different scanlines. An 8 bpp image has one pixel per byte, 
 * a 4 bpp image has up to 2 pixels per byte etc. Each pixel is an index 
 * into the image's indexed color palette (Image_GetIndexedPalette).
 * <p> 
 * The most significant bit of each scanline in the returned buffer 
 * corresponds to the leftmost pixel on the displayed scanline. Each 
 * scanline in the returned buffer may have unused bits, depending on color 
 * depth and image width. For example in a 3x3 image with 4 bpp color depth 
 * the 4 least significant bits of the last byte on each scanline will be 
 * unused. In addition, each scanline may be padded with unused bytes. The 
 * number of bytes per scanline is determined by dividing the image size 
 * (the size of the returned array) by the number of scanlines (height  
 * returned by Image_GetBounds()).
 * </p>
 * <p>
 * May result in the following error:
 * <ul>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * </ul>
 * </p>
 * 
 * @param handle image handle
 * @param pixels pixel data of entire image
 * @param length size, in bytes, of the <code>pixels</code> argument
 * @param uglError the error status of the call
 * @see #Image_GetIndexedPalette()
 * @see #Image_GetSize()
 */
void Image_GetPixelsLowColor(UGL_Int handle, UGL_ByteArray pixels, UGL_Int length, UGL_Error uglError);

/**
 * Returns the size in bytes needed to store all pixels of the image 
 * identified by <code>handle</code>. This should be the equivalent of
 * bytes per scanline * image height. The returned value must match the 
 * size of the byte array returned by <code>Image_GetPixelsLowColor</code> 
 * and the size of the int array returned by <code>Image_GetPixelsHighColor</code>.
 * <p>
 * Note that a scanline may be padded with unused bytes as described in
 * <code>Image_GetPixelsLowColor()</code>. 
 * </p>
 * 
 * @param handle image handle
 * @param error the error status of the call
 * @return the size in bytes needed to store all pixels of the image.
 * @see #Image_GetPixelsLowColor()
 * @see #Image_GetPixelsHighColor()
 */
UGL_Int Image_GetSize(UGL_Int handle, UGL_Error error);

/**
 * Returns the transparent color of the image identified by
 * <code>handle</code> or -1 if the image has no transparency.
 *
 * @ingroup Image
 *
 * @see Image_GetDirectPaletteMasks
 * @see Image_GetIndexedPalette
 * @see Image_IsDirect
 *
 * @param handle image handle
 * @param error the error status of the call
 *
 * @return the transparent color of the image. This is a palette index
 *         in the case of paletted images or an RGB color value for direct
 *         color images (Image_IsDirect). For RGB values the format matches
 *         that returned by <code>Image_GetDirectPaletteMasks()</code>
 */
UGL_Int Image_GetTransparentColor(UGL_Int handle, UGL_Error error);

/**
 * Returns whether the image identified by <code>handle</code> uses
 * a direct color model or a paletted color model.
 *
 * @ingroup Image
 *
 * @see Image_GetDirectPaletteMasks
 * @see Image_GetIndexedPalette
 *
 * @param handle image handle
 * @param error the error status of the call
 *
 * @return true - the image uses a direct color model. Pixel values are
 *         RGB values (Red, Green, Blue) with optional alpha channel.
 *         false - the image uses a paletted color mode. Pixel values are indices
 *         into a color palette.
 */
UGL_Boolean Image_IsDirect(UGL_Int handle, UGL_Error error);

/**
 * Creates a new image loaded from the specified file.
 * <p>
 * May result in the following errors:
 * <ul>
 * <li>ERROR_IO - I/O error occurred opening or reading the image file</li>
 * <li>ERROR_UNSUPPORTED_DEPTH - the color depth is not supported</li>
 * <li>ERROR_INVALID_IMAGE - an error has occurred while creating the image</li>
 * </ul>
 * </p>
 *
 * @ingroup Image
 *
 * @param deviceHandle the device/display handle
 * @param fileName image file to load
 * @param error the error status of the call
 *
 * @return handle to the new image loaded from the specified file.
 *         0 if the file could not be decoded as an image.
 */
UGL_Int Image_NewFromFile(UGL_Int deviceHandle, UGL_String fileName, UGL_Error error);





/**
 * Copies the specified area from the image identified by <code>handle</code>
 * into the destination image. The area is copied from location <code>x</code>,
 * <code>y</code> in the source image to location 0, 0 in the destination
 * image.
 *
 * @ingroup Image
 *
 * @param handle handle to the image to copy
 * @param destImageHandle handle to the destination image
 * @param x x coordinate of the area to copy to the destination image.
 *          Must be within the image's bounds.
 * @param y y coordinate of the area to copy to the destination image.
 *          Must be within the image's bounds.
 * @param width width of the area to copy to the destination image.
 *              Must be within the image's bounds.
 * @param height height of the area to copy to the destination image.
 *               Must be within the image's bounds.
 * @param error the error status of the call
 */
void Image_CopyArea(UGL_Int handle, UGL_Int destImageHandle, UGL_Int x, UGL_Int y, UGL_Int width, UGL_Int height, UGL_Error error);


#ifdef __cplusplus
}
#endif

#endif /* __UGL_IMAGE_H__ */
