/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

#ifndef __UGL_MENUITEM_H__
#define __UGL_MENUITEM_H__

#include "UGL_Error.h"
#include "UGLTypes.h"

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Enables or disables a menu item.
 *
 * @ingroup AbstractMenuItem
 *
 * @param menuItemHandle the menu item handle
 * @param enabled the enablement state of the menu item
 * @param error the error status of the call
 */
void AbstractMenuItem_SetEnabled(UGL_Int menuItemHandle, UGL_Boolean enabled, UGL_Error error);

/**
 * Creates a new cascade menu item, and places it at the given
 * index of the menu.  If the index value is -1, the menu should
 * append the new item.
 *
 * @ingroup CascadeMenuItem
 *
 * @param menuHandle the menu handle which contains this new item
 * @param index the index inside menu where this new item is placed
 * @param error the error status of the call
 *
 * @return handle of new widget
 */
UGL_Int CascadeMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error error);

/**
 * Sets the submenu of a cascade menu item.
 *
 * @ingroup CascadeMenuItem
 *
 * @param cascadeMenuItemHandle the cascade menu item
 * @param dropDownMenuHandle the submenu, or 0 if the current menu
 *                           should be removed.
 * @param error the error status of the call
 */
void CascadeMenuItem_SetMenu(UGL_Int cascadeMenuItemHandle, UGL_Int dropDownMenuHandle, UGL_Error error);

/**
 * Sets the accelerator for a LabeledMenuItem.  The accelerator is a bitwise
 * OR of MODIFIER_MASK constants and keycode values.  The MODIFIER_MASK constants
 * are placed in the upper 16 bits of the value.  The keycode is in the lower
 * 16 bits.  Valid keycodes are either KEYCODE constants or 16-bit char values.
 *
 * @ingroup LabeledMenuItem
 *
 * @param menuItemHandle the menu item handle
 * @param accelerator the accelerator
 * @param error the error status of the call
 */
void LabeledMenuItem_SetAccelerator(UGL_Int menuItemHandle, UGL_Int accelerator, UGL_Error error);

/**
 * Sets the descriptive text and the accelerator text of a LabeledMenuItem.
 * The descriptive text may contain mnemonic characters, which are preceded
 * by the '&amp;' character.  In order to add a '&amp;' character to the text,
 * it should be escaped with another '&amp;' character.
 * The accelerator text is displayed next to the descriptive text in the item,
 * and usually describes the accelerator, if one is being used.  Note that the
 * accelerator text does not actually set the accelerator.  That is done by
 * LabeledMenuItem_SetAccelerator.
 *
 * @ingroup LabeledMenuItem
 *
 * @param menuItemHandle the menu item handle
 * @param text the descriptive text
 * @param accelText the accelerator text
 * @param error the error status of the call
 */
void LabeledMenuItem_SetText(UGL_Int menuItemHandle, UGL_String text, UGL_String accelText, UGL_Error error);

/**
 * Creates a new push menu item.
 *
 * @ingroup PushMenuItem
 *
 * @param menuHandle the menu handle which contains this new item
 * @param index the index inside menu where this new item is placed
 * @param error the error status of the call
 *
 * @return handle of new widget
 */
UGL_Int PushMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error error);

/**
 * Creates a new separator menu item.
 *
 * @ingroup SeparatorItem
 *
 * @param menuHandle the menu handle which contains this new item
 * @param index the index inside menu where this new item is placed
 * @param error the error status of the call
 *
 * @return handle of new widget
 */
UGL_Int SeparatorItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Error error);

/**
 * Creates a new menu item which has a selection state, similar to a checkbox.
 *
 * @ingroup StateMenuItem
 *
 * @param menuHandle the menu handle which contains this new item
 * @param index the index inside menu where this new item is placed
 * @param style either STYLE_RADIO or STYLE_CHECK
 * @param error the error status of the call
 *
 * @return handle of new widget
 */
UGL_Int StateMenuItem_New(UGL_Int menuHandle, UGL_Int index, UGL_Int style, UGL_Error error);

/**
 * Sets the selection state of a state menu item.
 *
 * @ingroup StateMenuItem
 *
 * @param statefulMenuItemHandle the menu item handle
 * @param selected the new selection state
 * @param error the error status of the call
 */
void StateMenuItem_SetSelection(UGL_Int statefulMenuItemHandle, UGL_Boolean selected, UGL_Error error);



/**
 * Sets the image which is displayed on a LabeledMenuItem.  If the
 * imageHandle is zero, remove the current image.
 *
 * @ingroup LabeledMenuItem
 *
 * @param menuItemHandle the menu item handle
 * @param imageHandle the image handle
 * @param error the error status of the call
 */
void LabeledMenuItem_SetImage(UGL_Int menuItemHandle, UGL_Int imageHandle, UGL_Error error);



#ifdef __cplusplus
}
#endif

#endif /* __UGL_MENUITEM_H__ */
