/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

#include "UGL_ErrorImpl.h"
#include "Button.h"
#include "UGL_Error.h"
#include "TypeConverter.h"
#include <jni.h>
#include "JNIHelpers.h"
#include "UGLTypes.h"

#ifdef __cplusplus
#define UNUSED(var) 
#else
#define UNUSED(var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Sets the label of a button.  Please note that a button may not display
 * its image and its label simultaneously.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_Button_1SetLabel(JNIEnv* env, jclass UNUSED(clazz), jint handle, jstring string)
{
	UGL_ErrorStruct error;
	const jchar* stringCharData;
	jint stringLength;
	UGL_Boolean stringShouldFree;
	UGL_String uglstring;
	
	stringCharData = GetStringChars(env, string);
	if (ExceptionCheck(env)) return ;
	stringLength = GetStringLength(env, string);
	uglstring = convertToNativeString(stringCharData, stringLength, &stringShouldFree);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Button_SetLabel\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APITRACE(env, "\tstringCharData=");
	{
		int i;
		for (i=0; i<stringLength; i++) APIARGTRACE(env, "%c" , stringCharData[i]);
	}
	APITRACE(env, "\n");
	#endif

	Button_SetLabel(handle, uglstring, &error);
	if (stringShouldFree) freeNativeString(uglstring);
	ReleaseStringChars(env, string, stringCharData);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Button_SetLabel", &error);
	}
}

/**
 * Set the selection state of a button.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_Button_1SetSelection(JNIEnv* env, jclass UNUSED(clazz), jint handle, jboolean selection)
{
	UGL_ErrorStruct error;
	UGL_Boolean nativeselection;
	
	nativeselection = convertToNativeBoolean(env, selection);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Button_SetSelection\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APIARGTRACE(env, "\tnativeselection=%d\n" , nativeselection);
	#endif

	Button_SetSelection(handle, nativeselection, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Button_SetSelection", &error);
	}
}

/**
 * Creates a new Checkbox widget
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_Checkbox_1New(JNIEnv* env, jclass UNUSED(clazz), jint parentHandle, jint style)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Checkbox_New\n");
	APIARGTRACE(env, "\tparentHandle=%d\n" , parentHandle);
	APIARGTRACE(env, "\tstyle=%d\n" , style);
	#endif

	returnValue = Checkbox_New(parentHandle, style, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "Checkbox_New", &error);
		return 0;
	}
}

/**
 * Create a new PushButton widget
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_PushButton_1New(JNIEnv* env, jclass UNUSED(clazz), jint parentHandle, jint style)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - PushButton_New\n");
	APIARGTRACE(env, "\tparentHandle=%d\n" , parentHandle);
	APIARGTRACE(env, "\tstyle=%d\n" , style);
	#endif

	returnValue = PushButton_New(parentHandle, style, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "PushButton_New", &error);
		return 0;
	}
}

/**
 * Create a new RadioButton widget
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_RadioButton_1New(JNIEnv* env, jclass UNUSED(clazz), jint parentHandle, jint style)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - RadioButton_New\n");
	APIARGTRACE(env, "\tparentHandle=%d\n" , parentHandle);
	APIARGTRACE(env, "\tstyle=%d\n" , style);
	#endif

	returnValue = RadioButton_New(parentHandle, style, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "RadioButton_New", &error);
		return 0;
	}
}



/**
 * Sets the position of the label or image on the button.
 * On left to right locales, STYLE_LEAD is left, and STYLE_TRAIL
 * is right.  Right to left locales reverse this notation.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_Button_1SetAlignment(JNIEnv* env, jclass UNUSED(clazz), jint handle, jint alignment)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Button_SetAlignment\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APIARGTRACE(env, "\talignment=%d\n" , alignment);
	#endif

	Button_SetAlignment(handle, alignment, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Button_SetAlignment", &error);
	}
}

/**
 * Sets the image of a button.  Please note that a button may not display
 * its image and its label simultaneously.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_Button_1SetImage(JNIEnv* env, jclass UNUSED(clazz), jint handle, jint imageHandle)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - Button_SetImage\n");
	APIARGTRACE(env, "\thandle=%d\n" , handle);
	APIARGTRACE(env, "\timageHandle=%d\n" , imageHandle);
	#endif

	Button_SetImage(handle, imageHandle, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "Button_SetImage", &error);
	}
}

/**
 * Create a new ToggleButton widget
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_ToggleButton_1New(JNIEnv* env, jclass UNUSED(clazz), jint parentHandle, jint style)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - ToggleButton_New\n");
	APIARGTRACE(env, "\tparentHandle=%d\n" , parentHandle);
	APIARGTRACE(env, "\tstyle=%d\n" , style);
	#endif

	returnValue = ToggleButton_New(parentHandle, style, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "ToggleButton_New", &error);
		return 0;
	}
}


#ifdef __cplusplus
}
#endif

