/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

#include "UGL_ErrorImpl.h"
#include "UGL_Error.h"
#include "TypeConverter.h"
#include "MenuItem.h"
#include <jni.h>
#include "JNIHelpers.h"
#include "UGLTypes.h"

#ifdef __cplusplus
#define UNUSED(var) 
#else
#define UNUSED(var) var
#endif

#ifdef __cplusplus
extern "C" {
#endif


/**
 * Enables or disables a menu item.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_AbstractMenuItem_1SetEnabled(JNIEnv* env, jclass UNUSED(clazz), jint menuItemHandle, jboolean enabled)
{
	UGL_ErrorStruct error;
	UGL_Boolean nativeenabled;
	
	nativeenabled = convertToNativeBoolean(env, enabled);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - AbstractMenuItem_SetEnabled\n");
	APIARGTRACE(env, "\tmenuItemHandle=%d\n" , menuItemHandle);
	APIARGTRACE(env, "\tnativeenabled=%d\n" , nativeenabled);
	#endif

	AbstractMenuItem_SetEnabled(menuItemHandle, nativeenabled, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "AbstractMenuItem_SetEnabled", &error);
	}
}

/**
 * Creates a new cascade menu item, and places it at the given
 * index of the menu.  If the index value is -1, the menu should
 * append the new item.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_CascadeMenuItem_1New(JNIEnv* env, jclass UNUSED(clazz), jint menuHandle, jint index)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - CascadeMenuItem_New\n");
	APIARGTRACE(env, "\tmenuHandle=%d\n" , menuHandle);
	APIARGTRACE(env, "\tindex=%d\n" , index);
	#endif

	returnValue = CascadeMenuItem_New(menuHandle, index, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "CascadeMenuItem_New", &error);
		return 0;
	}
}

/**
 * Sets the submenu of a cascade menu item.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_CascadeMenuItem_1SetMenu(JNIEnv* env, jclass UNUSED(clazz), jint cascadeMenuItemHandle, jint dropDownMenuHandle)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - CascadeMenuItem_SetMenu\n");
	APIARGTRACE(env, "\tcascadeMenuItemHandle=%d\n" , cascadeMenuItemHandle);
	APIARGTRACE(env, "\tdropDownMenuHandle=%d\n" , dropDownMenuHandle);
	#endif

	CascadeMenuItem_SetMenu(cascadeMenuItemHandle, dropDownMenuHandle, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "CascadeMenuItem_SetMenu", &error);
	}
}

/**
 * Sets the accelerator for a LabeledMenuItem.  The accelerator is a bitwise
 * OR of MODIFIER_MASK constants and keycode values.  The MODIFIER_MASK constants
 * are placed in the upper 16 bits of the value.  The keycode is in the lower
 * 16 bits.  Valid keycodes are either KEYCODE constants or 16-bit char values.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_LabeledMenuItem_1SetAccelerator(JNIEnv* env, jclass UNUSED(clazz), jint menuItemHandle, jint accelerator)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - LabeledMenuItem_SetAccelerator\n");
	APIARGTRACE(env, "\tmenuItemHandle=%d\n" , menuItemHandle);
	APIARGTRACE(env, "\taccelerator=%d\n" , accelerator);
	#endif

	LabeledMenuItem_SetAccelerator(menuItemHandle, accelerator, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "LabeledMenuItem_SetAccelerator", &error);
	}
}

/**
 * Sets the descriptive text and the accelerator text of a LabeledMenuItem.
 * The descriptive text may contain mnemonic characters, which are preceded
 * by the '&amp;' character.  In order to add a '&amp;' character to the text,
 * it should be escaped with another '&amp;' character.
 * The accelerator text is displayed next to the descriptive text in the item,
 * and usually describes the accelerator, if one is being used.  Note that the
 * accelerator text does not actually set the accelerator.  That is done by
 * LabeledMenuItem_SetAccelerator.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_LabeledMenuItem_1SetText(JNIEnv* env, jclass UNUSED(clazz), jint menuItemHandle, jstring text, jstring accelText)
{
	const jchar* accelTextCharData;
	jint accelTextLength;
	UGL_Boolean accelTextShouldFree;
	UGL_ErrorStruct error;
	const jchar* textCharData;
	jint textLength;
	UGL_Boolean textShouldFree;
	UGL_String uglaccelText;
	UGL_String ugltext;
	
	textCharData = GetStringChars(env, text);
	if (ExceptionCheck(env)) return ;
	textLength = GetStringLength(env, text);
	ugltext = convertToNativeString(textCharData, textLength, &textShouldFree);
	accelTextCharData = GetStringChars(env, accelText);
	if (ExceptionCheck(env)) return ;
	accelTextLength = GetStringLength(env, accelText);
	uglaccelText = convertToNativeString(accelTextCharData, accelTextLength, &accelTextShouldFree);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - LabeledMenuItem_SetText\n");
	APIARGTRACE(env, "\tmenuItemHandle=%d\n" , menuItemHandle);
	APITRACE(env, "\ttextCharData=");
	{
		int i;
		for (i=0; i<textLength; i++) APIARGTRACE(env, "%c" , textCharData[i]);
	}
	APITRACE(env, "\n");
	APITRACE(env, "\taccelTextCharData=");
	{
		int i;
		for (i=0; i<accelTextLength; i++) APIARGTRACE(env, "%c" , accelTextCharData[i]);
	}
	APITRACE(env, "\n");
	#endif

	LabeledMenuItem_SetText(menuItemHandle, ugltext, uglaccelText, &error);
	if (textShouldFree) freeNativeString(ugltext);
	ReleaseStringChars(env, text, textCharData);
	if (accelTextShouldFree) freeNativeString(uglaccelText);
	ReleaseStringChars(env, accelText, accelTextCharData);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "LabeledMenuItem_SetText", &error);
	}
}

/**
 * Creates a new push menu item.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_PushMenuItem_1New(JNIEnv* env, jclass UNUSED(clazz), jint menuHandle, jint index)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - PushMenuItem_New\n");
	APIARGTRACE(env, "\tmenuHandle=%d\n" , menuHandle);
	APIARGTRACE(env, "\tindex=%d\n" , index);
	#endif

	returnValue = PushMenuItem_New(menuHandle, index, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "PushMenuItem_New", &error);
		return 0;
	}
}

/**
 * Creates a new separator menu item.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_SeparatorItem_1New(JNIEnv* env, jclass UNUSED(clazz), jint menuHandle, jint index)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - SeparatorItem_New\n");
	APIARGTRACE(env, "\tmenuHandle=%d\n" , menuHandle);
	APIARGTRACE(env, "\tindex=%d\n" , index);
	#endif

	returnValue = SeparatorItem_New(menuHandle, index, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "SeparatorItem_New", &error);
		return 0;
	}
}

/**
 * Creates a new menu item which has a selection state, similar to a checkbox.
 */
JNIEXPORT jint JNICALL Java_com_ibm_ugl_eswt_OS_StateMenuItem_1New(JNIEnv* env, jclass UNUSED(clazz), jint menuHandle, jint index, jint style)
{
	UGL_ErrorStruct error;
	UGL_Int returnValue;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - StateMenuItem_New\n");
	APIARGTRACE(env, "\tmenuHandle=%d\n" , menuHandle);
	APIARGTRACE(env, "\tindex=%d\n" , index);
	APIARGTRACE(env, "\tstyle=%d\n" , style);
	#endif

	returnValue = StateMenuItem_New(menuHandle, index, style, &error);
	if (error.error_code == 0) {
		return returnValue;
	} else {
		UGL_ErrorHandler(env, "StateMenuItem_New", &error);
		return 0;
	}
}

/**
 * Sets the selection state of a state menu item.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_StateMenuItem_1SetSelection(JNIEnv* env, jclass UNUSED(clazz), jint statefulMenuItemHandle, jboolean selected)
{
	UGL_ErrorStruct error;
	UGL_Boolean nativeselected;
	
	nativeselected = convertToNativeBoolean(env, selected);
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - StateMenuItem_SetSelection\n");
	APIARGTRACE(env, "\tstatefulMenuItemHandle=%d\n" , statefulMenuItemHandle);
	APIARGTRACE(env, "\tnativeselected=%d\n" , nativeselected);
	#endif

	StateMenuItem_SetSelection(statefulMenuItemHandle, nativeselected, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "StateMenuItem_SetSelection", &error);
	}
}



/**
 * Sets the image which is displayed on a LabeledMenuItem.  If the
 * imageHandle is zero, remove the current image.
 */
JNIEXPORT void JNICALL Java_com_ibm_ugl_eswt_OS_LabeledMenuItem_1SetImage(JNIEnv* env, jclass UNUSED(clazz), jint menuItemHandle, jint imageHandle)
{
	UGL_ErrorStruct error;
	
	error.error_code = 0;
	
	#ifdef debug_print
	APITRACE(env, "Entering - LabeledMenuItem_SetImage\n");
	APIARGTRACE(env, "\tmenuItemHandle=%d\n" , menuItemHandle);
	APIARGTRACE(env, "\timageHandle=%d\n" , imageHandle);
	#endif

	LabeledMenuItem_SetImage(menuItemHandle, imageHandle, &error);
	if (error.error_code == 0) {
		return;
	} else {
		UGL_ErrorHandler(env, "LabeledMenuItem_SetImage", &error);
	}
}



#ifdef __cplusplus
}
#endif

